/*
 * Fabric Management Agent
 *
 * One runs on each Myrinet node and communicates with the fabric
 * management server(s).
 */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>

#include "libfma.h"
#include "lf_scheduler.h"
#include "lf_channel.h"

#include "fma.h"
#include "fma_fms.h"

/*
 * All global variables live in here
 */
struct fma_vars A;

/*
 * local prototypes
 */
static void fma_init_vars(void);
static int parse_options(int argc, char **argv);
static int fma_load_environment(void);
static int fma_init_process(void);
static void fma_write_pid_file(void);

static void
usage()
{
  fprintf(stderr, "Usage: fma\n");
  fprintf(stderr, "  -h - print this help message\n");
  fprintf(stderr, "  -i - fabric consists of xbars with IDs only\n");
  fprintf(stderr, "  -x - fabric consists of ID-less xbars only\n");
  fprintf(stderr, "  -D - increment debug level\n");
  fprintf(stderr, "  -s <fms_server> - specify fms server\n");
  fprintf(stderr, "  -R <fms_run> - run directory (for pid and log files)\n");
  fprintf(stderr, "  -V - print version\n");
  exit(1);
}

int
main(int argc,
     char **argv)
{
  int ms;
  int rc;

  /* initialize error handler */
  lf_set_error_handler(fma_save_error);

  /* initialize all variables to default values */
  fma_init_vars();

  /* first load environmental settings, then command line may override */
  if (fma_load_environment() != 0) FMA_ERROR("loading environment");
  if (parse_options(argc, argv) != 0) FMA_ERROR("parsing command line");

  /* Initialize process stuff, signals, stdio, etc. */
  if (fma_init_process() == -1) FMA_ERROR("Error initializing process");

  /* initialize channels subsystem */
  if (lf_init_channels() == -1) FMA_ERROR("Initializing channels subsystem");

  /* Initialize Myricom setup */
  sleep(5);			/* give multi-NIC time to come up XXX */
  if (fma_init_myri() == -1) FMA_ERROR("Error initializing Myricom");

  /* initialize FMS subsystem */
  if (fma_init_fms() == -1) FMA_ERROR("Initializing FMS subsystem");

  /* Initialize standalone stuff */
  fma_init_standalone();

  /* if we made it this far, we are up and running!  write pid file */
  fma_write_pid_file();

  /*
   * Just loop handling events forever
   */
  while (!A.done) {
    ms = lf_check_events();	/* look for any scheduled events */

    rc = lf_poll_channels(ms);	/* look for any pending I/O requests */
    if (rc == -1) FMA_ERROR("polling channels");
  }

  fma_exit(0);

 except:
  fma_perror();
  fma_exit(1);
  return 0;
}

/*
 * Initialize the FMA global data structure
 */
static void
fma_init_vars()
{
  /* initialize FMA variables */
  A.done = FALSE;

  A.xbar_types = FMA_XT_ANY;		/* default is any type of xbar */
  A.fma_run = LF_DFLT_FMS_RUN;

  if (gethostname(A.hostname, sizeof(A.hostname)) == -1) {
    FMA_ERROR("Getting local hostname");
  }

  /* Initialize settings vars */
  fma_init_settings_vars();

  /* vars for the Myricom subsystem */
  if (fma_init_myri_vars() != 0) FMA_ERROR("Initializing Myricom interfaces");

  /* vars for the FMS subsystem */
  if (fma_init_fms_vars() != 0) FMA_ERROR("Initializing FMS subsystem");

  /* In case we need to run standalone */
  fma_init_standalone_vars();

  fma_init_tunnel_vars();	/* initialize message tunnel variables */

  return;

 except:
  fma_perror_exit(1);
}

/*
 * Load environment variables
 */
static int
fma_load_environment()
{
  char *s;

  s = getenv(LF_ENV_FMS_SERVER);
  if (s != NULL) {
    if (fma_set_fms_list(s) != 0) FMA_ERROR("Error getting fms server list from env");
  }

  s = getenv(LF_ENV_FMS_RUN);
  if (s != NULL) {
    A.fma_run = s;
  }
  return 0;

 except:
  return -1;
}

/*
 * parse command line args
 */
static int
parse_options(
  int argc,
  char **argv)
{
  extern char *optarg;
  int c;

  while ((c = getopt(argc, argv, "hDR:s:ixV")) != EOF) switch (c) {
  case 'h':
    usage();
    break;
  case 'R':
    A.fma_run = optarg;
    break;
  case 's':
    if (fma_set_fms_list(optarg) != 0) {
      FMA_ERROR("Error loading host list from cmd line");
    }
    break;
  case 'D':
    ++A.debug;		/* increase debug level */
    break;
  case 'i':
    printf("ID-less xbar support disabled\n");
    A.xbar_types = FMA_XT_ID_ONLY;
    break;
  case 'x':
    printf("xbar ID support disabled\n");
    A.xbar_types = FMA_XT_NO_IDS;
    break;
  case 'V':
    printf("FMS version is %s\n", Lf_version);
    exit(0);
    break;
  }

  return 0;

 except:
  return -1;
}


/*
 * Print the error stack
 */
void
fma_perror()
{
  int i;

  for (i=0; i<A.error_index; ++i) {
    fprintf(stderr, "%s:%d %s", A.error_file[i], A.error_line[i],
            A.error_str[i]);
    if (A.error_errno != 0) {
      fprintf(stderr, ": %s", strerror(A.error_errno[i]));
    }
    fprintf(stderr, "\n");
  }

  fma_reset_error();
}

void
fma_reset_error()
{
  A.error_index = 0;
}

/*
 * perform any cleanup and exit
 */
void
fma_exit(
  int code)
{
  exit(code);
}

/* combo */
void
fma_perror_exit(
  int code)
{
  fma_perror();
  abort();		/* XXX */
  fma_exit(code);
}

/*
 * save away an error message
 */
void
fma_save_error(
  char *file,
  int lineno,
  int err_no,
  char *errstr)
{
  if (A.error_index >= (FMA_MAX_ERROR_DEPTH-1)) {
    strncpy(A.error_str[FMA_MAX_ERROR_DEPTH-1], "Error depth overflow",
	FMA_MAX_ERROR_LEN);
    A.error_index = FMA_MAX_ERROR_DEPTH;
  } else {
    A.error_file[A.error_index] = file;
    A.error_line[A.error_index] = lineno;
    A.error_errno[A.error_index] = err_no;

    /* error_str[] is declared with FMA_MAX_ERROR_LEN+1 to force null-term */
    strncpy(A.error_str[A.error_index++], errstr, FMA_MAX_ERROR_LEN);
  }					
}

/*
 * Initialize stdio/stderr
 */
static int
fma_init_process()
{
  setlinebuf(stdout);
  setlinebuf(stderr);

  return 0;
}

/*
 * Write PID file
 */
static void
fma_write_pid_file()
{
  FILE *fp;
  lf_string_t path;

  /* save a PID file */
  sprintf(path, "%s/fma_pid", A.fma_run);
  fp = fopen(path, "w");
  if (fp == NULL) {
    perror(path);
    exit(1);
  }

  /* write our PID to a file */
  fprintf(fp, "%d\n", getpid());
  fclose(fp);
}
