/*
  indent fma_dfs_route.c -bli0 -bfda -kr -psl
*/


#include "assert.h"
#include "libfma.h"
#include "lf_fabric.h"
#include "lf_fms_comm.h"
#include "lf_topo_map.h"
#include "libmyri.h"

#include "fma.h"
#include "fma_myri.h"
#include "fma_map.h"
#include "fma_fabric.h"

#include "fma_dfs_route.h"

#define FMA_DF_A_FEW_TIMES 10
#define MAX_XBAR_PORTS 64	/*buy some time */
enum { UP, DOWN, FLAT };

static void
fma_dfs_clear_seen(struct lf_fabric *fp)
{
  int i, j;
  assert(fp);

  for (i = 0; i < fp->num_xbars; i++) {
    FMA_XBAR(fp->xbars[i])->seen = 0;
  }

  for (i = 0; i < FMA_FABRIC(fp)->num_nics; i++) {
    for (j = 0; j < LF_MAX_NIC_PORTS; j++) {
      FMA_NIC(FMA_FABRIC(fp)->nics[i])->seen[j] = 0;
    }
  }
}

/*get clos direction from 'from' to 'to'*/

static int
fma_dfs_get_direction(struct lf_xbar *from,
		      struct lf_xbar *to)
{
  assert(from && to);

  if (to->clos_level < from->clos_level)
    return DOWN;
  else if (to->clos_level > from->clos_level)
    return UP;
  else
    return FLAT;
}

static void
fma_dfs_clear_load(struct lf_fabric *fp)
{
  int i, j;
  assert(fp);

  for (i = 0; i < fp->num_xbars; i++) {
    for (j = 0; j < fp->xbars[i]->num_ports; j++)
      FMA_XBAR(fp->xbars[i])->link_load[j] = 0;
  }
}

/*
  sets the distance field of each nic to the shortest
  path distance. this is used to prune the dfs.

  0 on error, which would be if the up/down numbering prevents
  the graph from being fully searchable.
*/

static int
fma_dfs_compute_shortest_distances(struct lf_fabric *fp,
				   struct fma_nic_info *nip,
				   int port)
{
  struct lf_xbar *x;
  struct lf_xbar *head;
  struct lf_xbar *tail;
  struct lf_nic *nic;
  int num_nics = 0;
  int num_xbars = 0;
  int i;

  assert(fp && nip && port >= 0);

  nic = FMA_FABRIC(fp)->nics[nip->map_index];

  x = LF_XBAR(nic->topo_ports[port]);
  if (x == NULL) {
    return -1;
  }

  fma_dfs_clear_seen(fp);

  FMA_XBAR(x)->in_port = nic->topo_rports[port];
  FMA_XBAR(x)->distance = 0;
  FMA_XBAR(x)->direction = UP;
  FMA_XBAR(x)->bf_next = 0;
  FMA_XBAR(x)->seen = 1;

  head = tail = x;

  while (head) {
    x = head;
    head = FMA_XBAR(head)->bf_next;
    num_xbars++;

    for (i = 0; i < x->num_ports; i++) {
      struct lf_node *rmt_node = x->topo_ports[i];
      int rmt_port = x->topo_rports[i];

      if (!rmt_node || FMA_XBAR(x)->link_down[i])
	continue;

      if (rmt_node->type == LF_NODE_NIC && !FMA_NIC(rmt_node)->seen[rmt_port]) {
	FMA_NIC(rmt_node)->seen[rmt_port] = 1;
	FMA_NIC(rmt_node)->distance[rmt_port] = FMA_XBAR(x)->distance + 1;
	num_nics++;
      } else if (rmt_node->type == LF_NODE_XBAR && !FMA_XBAR(rmt_node)->seen) {
	int new_dir =
	  fma_dfs_get_direction(x, (struct lf_xbar *) rmt_node);
	if (FMA_XBAR(x)->direction != UP && new_dir != DOWN)
	  continue;

	FMA_XBAR(rmt_node)->distance = FMA_XBAR(x)->distance + 1;
	FMA_XBAR(rmt_node)->direction = new_dir;
	FMA_XBAR(rmt_node)->bf_next = 0;
	FMA_XBAR(rmt_node)->seen = 1;

	if (head) {
	  assert(tail);
	  tail = FMA_XBAR(tail)->bf_next = (struct lf_xbar *) rmt_node;
	} else
	  head = tail = (struct lf_xbar *) rmt_node;
      }
    }
  }
  return 0;
#if 0
  if (num_nics == FMA_FABRIC(fp)->num_nics) {
    return 0;
  } else {
    return -1;
  }
#endif
}

/*
  puts a port ordering in ports corresponding to x's ports, sorted by link_load

*/
static void
fma_dfs_sort_ports(struct lf_xbar *x,
		   int *ports)
{
  int i, j;

  assert(x && ports && FMA_XBAR(x)->link_load);

  /*get the numbers */
  for (i = 0; i < x->num_ports; i++)
    ports[i] = i;

  /*shuffle future ties */
  for (i = 0; i < FMA_DF_A_FEW_TIMES; i++) {
    int a = random() % x->num_ports;
    int b = random() % x->num_ports;
    int t = ports[a];
    ports[a] = ports[b];
    ports[b] = t;
  }

  /*sort like a child */
  for (i = 0; i < x->num_ports; i++) {
    int lowp = -1;
    int lowc = FMA_XBAR(x)->link_load[ports[i]];
    int t;

    for (j = i+1; j < x->num_ports; j++) {

      if (FMA_XBAR(x)->link_load[ports[j]] < lowc) {
	lowp = j;
	lowc = FMA_XBAR(x)->link_load[ports[j]];
      }
    }
    if (lowp >= 0) {
      t = ports[i];
      ports[i] = ports[lowp];
      ports[lowp] = t;
    }
  }
}

/*
  does a dfs from 'x' stopping after 'sought' hosts are found.
  sets routes at each of these hosts.
  returns number of hosts found.
  the dfs is limited by up/down. 
  port order is sorted by least use.
  only hosts found beyond depth are ignored.
*/
static int
fma_dfs(
  struct fma_nic_info *nip,
  int port,
  struct lf_xbar *x,
  int depth,
  int direction,
  int sought,
  int route_num)
{
  int i;
  int found = 0;
  int ports[MAX_XBAR_PORTS];

  assert(nip && x && sought && ports);
  assert(x->num_ports <= MAX_XBAR_PORTS);

  fma_dfs_sort_ports(x, ports);

  for (i = 0; i < x->num_ports && found < sought; i++) {
    struct lf_node *rmt_node;   /* node that port i of 'x' points to */
    int rmt_port;               /* port on 'rmt_node' that port i of 'x' points to */

    assert(ports[i] >= 0 && ports[i] < x->num_ports);
    rmt_node = x->topo_ports[ports[i]];
    rmt_port = x->topo_rports[ports[i]];

    if (!rmt_node || FMA_XBAR(x)->link_down[ports[i]])
      continue;

    if (rmt_node->type == LF_NODE_NIC) {
      if (!FMA_NIC(rmt_node)->seen[rmt_port] && depth <= FMA_NIC(rmt_node)->distance[rmt_port] + 1) {
	/*found a new host. save route */

	int rindex = FMA_ROUTE_INDEX(nip->nic_index, port, rmt_port, route_num);
	unsigned char *route =
	  FMA_NIC(rmt_node)->route_buf + FMA_ROUTE_OFFSET(rindex);
	memcpy(route, FMA_XBAR(x)->route, FMA_XBAR(x)->route_len);
	route[FMA_XBAR(x)->route_len] =
	  LF_DELTA_TO_ROUTE(ports[i] - FMA_XBAR(x)->in_port);
	FMA_NIC(rmt_node)->route_len[rindex] = FMA_XBAR(x)->route_len + 1;
	FMA_NIC(rmt_node)->seen[rmt_port] = 1;
	found++;
      }

    } else if (rmt_node->type == LF_NODE_XBAR && !FMA_XBAR(rmt_node)->seen) {
      int count;
      int new_dir;

      /*obey the law */
      new_dir = fma_dfs_get_direction(x, (struct lf_xbar *) rmt_node);
      if (direction != UP && new_dir != DOWN)
	continue;

      /*record route to this xbar */
      memcpy(FMA_XBAR(rmt_node)->route, FMA_XBAR(x)->route,
	     FMA_XBAR(x)->route_len);
      FMA_XBAR(rmt_node)->route[FMA_XBAR(x)->route_len] =
	LF_DELTA_TO_ROUTE(ports[i] - FMA_XBAR(x)->in_port);
      FMA_XBAR(rmt_node)->route_len = FMA_XBAR(x)->route_len + 1;
      FMA_XBAR(rmt_node)->in_port = rmt_port;
      FMA_XBAR(rmt_node)->direction = new_dir;
      FMA_XBAR(rmt_node)->seen = 1;

      /* search down him */
      count = fma_dfs(nip, port, (struct lf_xbar *) rmt_node, depth + 1,
        new_dir, sought - found, route_num);
      found += count;

      /* record load and erase our tracks */
      FMA_XBAR(x)->link_load[ports[i]] += count;
      FMA_XBAR(rmt_node)->seen = 0;
    }
  }
  return found;
}

static int
fma_dfs_calc_routes(struct lf_fabric *fp,
		    struct fma_nic_info *nip,
		    int port, int route_num, int num_nic_ports)
{
  struct lf_xbar *x;
  struct lf_nic *nic;
  int found = 0;
  int how_many = 1;

  nic = FMA_FABRIC(fp)->nics[nip->map_index];
  x = LF_XBAR(nic->topo_ports[port]);

  /* If disconnected, nothing to do! */
  if (x == NULL) {
    return 0;
  }

  /*for multiple routes we'll end up doing this each time. but we have
    to, because we may be interleaving calculations from multiple nics
    on this host, and per nic, we need a different set of shortest
    paths */

  if (fma_dfs_compute_shortest_distances(fp, nip, port) == -1) {
    FMA_ERROR("couldn't compute shortest distances. not a clos?");
  }

  fma_dfs_clear_seen(fp);

  FMA_XBAR(x)->route_len = 0;
  FMA_XBAR(x)->in_port = nic->topo_rports[port];

  while (found < num_nic_ports) {
    int t;

    /*if we ever fail to find someone, well, maybe it means
      num_nics is spread across disjoiint graphs.
    */
    FMA_XBAR(x)->seen = 1;

    t = fma_dfs(nip, port, x, 0, UP, how_many, route_num);

    if (t == 0) {
      break;
    }

    found += t;
  }
  return 0;

 except:
  return -1;
}

/*
  calcualte a set of routes for a nic
  clear_loads should be 1 the 1st time this is called, and 0 for subsequent
  calls if computing multiple routes.
*/

int
fma_dfs_calc_nic_routes(struct lf_fabric *fp,
			struct fma_nic_info *nip,
			int clear_loads, int route_num)
{
  int i;
  int num_nic_ports = 0;
  struct lf_nic *nicp;


  assert(fp && nip);

  /* count number of interfaces */
  for (i = 0; i < FMA_FABRIC(fp)->num_nics; i++) {
     nicp = FMA_FABRIC(fp)->nics[i];
     assert(nicp);
     num_nic_ports += LF_NIC(nicp)->num_ports;
  }

  if (clear_loads)
    fma_dfs_clear_load(fp);

  for (i = 0; i < num_nic_ports; i++) {
    if (fma_dfs_calc_routes(fp, nip, i, route_num, num_nic_ports) != 0) {
      return -1;
    }
  }
  return 0;
}


/*
  calculate a set of routes for all nics on this host
  clear_loads should be 1 the 1st time this is called, and 0 for subsequent
  calls if computing multiple routes.
*/

int
fma_dfs_calc_all_routes(
  int clear_loads,
  int route_num)
{
  int i, j;
  struct lf_nic *nicp;
  struct lf_fabric *fp;
  int num_nic_ports;
  
  num_nic_ports = 0;
  fp = A.fabric;

  if (clear_loads) {
    fma_dfs_clear_load(fp);
  }

  /* count number of interfaces */
  for (i = 0; i < FMA_FABRIC(fp)->num_nics; i++) {
     nicp = FMA_FABRIC(fp)->nics[i];
     num_nic_ports += LF_NIC(nicp)->num_ports;
  }

  for (i = 0; i < A.myri->num_nics; i++) {
    struct fma_nic_info *nip;

    nip = A.myri->nic_info[i];
    assert(nip);

    for (j = 0; j < nip->myri_info.num_ports; j++)
      if (fma_dfs_calc_routes(fp, nip, j, route_num, num_nic_ports) != 0) {
	return -1;
      }
  }

  return 0;
}
