#ifndef _fma_fabric_h_
#define _fma_fabric_h_

/*
 * This is the FMA-private data for the fabric
 */
struct fma_fabric {
  uint32_t map_id;		/* unique ID for this map */

  int num_nics;
  struct lf_nic **nics;
  int max_nic_ports;		/* most ports on any NIC */

  /* all the links, link indices correspond to map from FMS */
  int num_links;
  struct lf_node **link_end_node;
  char *link_end_port;
};
#define FMA_FABRIC(F) ((struct fma_fabric *)((F)->user))

/*
 * FMA-private struct for xbars
 */
struct fma_xbar {
  struct lf_xbar *parent;	/* useful link back to parent */

  uint32_t node_id;
  int map_index;		/* index in map array */

  int *link_down;		/* true or false for each port */
  int *link_load;               /*for routing algorithms*/
  
  /* Persistant info about best way to reach this xbar */
  int host_dist;			/* closest NIC port on this host */
  struct fma_nic_info *host_nic;	/* which NIC is closest */
  int host_nic_port;			/* which port */
  unsigned char host_route[FMA_IFC_ROUTE_LEN];	/* the route */
  int host_in_port;			/* inbound port on this xbar */

  /*
   * private data for mapping the fabric - must persist across scheduler calls
   */
  struct fma_mf_info *mf_info;		/* per-port info */
  int mf_xbar_resolved;			/* uniquely identified? */
  struct lf_event *mf_compare_timer;	/* xbar compare timer */
  int mf_compare_retries;		/* # of compare retries remaining */
  int mf_compare_port;			/* the port on this xbar */
  int mf_now_comparing;			/* true if compare in progress */
  int mf_ref_count;		/* used when referenced by send callback */

  struct fma_nic_info *mf_nic;		/* how to reach this xbar */
  int mf_nic_port;
  int mf_route_len;
  unsigned char mf_route[FMA_IFC_ROUTE_LEN];
  int mf_in_port;

  struct fma_xbar *mf_next;		/* for linking them */
  struct fma_xbar *mf_prev;

  /*
   * non-persistant information - once you return to scheduler, these may 
   * change!
   */

  /* for breadth-first searches */
  int bf_serial;			/* integer serial number */
  void *bf_serialp;			/* pointer serial number */
  struct lf_xbar *bf_next;
  
  int in_port;
  int route_len;
  unsigned char route[FMA_IFC_ROUTE_LEN];

  int seen;		/*for dfs*/
  int distance;		/*for dfs*/
  int direction;	/*for dfs because that work list stuff confuses me*/
};
#define FMA_XBAR(X) ((struct fma_xbar *)(((struct lf_xbar *)(X))->user))

/*
 * FMA-private struct for NICs
 */
struct fma_nic {
  struct lf_nic *parent;	/* useful link back to parent */
  struct fma_nic_info *local_nip; /* link to myri nic info, NULL if non-local */

  uint32_t node_id;
  int map_index;		/* index in map array */

  int *link_down;		/* true or false for each port */

  unsigned char *route_buf;	/* buffer for holding all routes */
  unsigned char *route_len;  	/* route lengths */

  int seen[LF_MAX_NIC_PORTS];   /* for dfs*/
  int distance[LF_MAX_NIC_PORTS]; /* for dfs*/

  /*
   * for mapping
   */
  struct fma_mf_info *mf_info;
  int mf_host_index;			/* for determining unique owner */
  struct fma_nic *mf_next;		/* for linking NICs during mapping */
  struct fma_nic *mf_prev;
};
#define FMA_NIC(N) ((struct fma_nic *)(((struct lf_nic *)(N))->user))

/*
 * FMA-private struct for hosts
 */
struct fma_host {
  lf_mac_addr_t	max_mac_addr;	/* MAX MAC address on this host */

  int map_send_id;		/* msg ID of topo map being sent */

  int host_id;			/* a global sequential numbering of the hosts */

  /* a quick route to this host */
  uint8_t *quick_route;
  int quick_route_len;
  struct fma_nic_info *quick_nic;	/* NIC/port for this route */
  int quick_port;
};
#define FMA_HOST(H) ((struct fma_host *)(((struct lf_host *)(H))->user))


struct lf_xbar *fma_create_xbar(int num_ports, int xbar_id, int quad_dis);
struct lf_nic *fma_create_nic(int num_ports, lf_mac_addr_t mac_addr);
struct lf_host *fma_create_host(void);
void fma_free_xbar(struct lf_xbar *xp);
void fma_free_nic(struct lf_nic *nicp);
void fma_free_host(struct lf_host *hp);
void fma_calc_xbar_dists(void);

#endif /* _fma_fabric_h_ */
