/*
 * Deal with the FMS
 */
#include <unistd.h>
#include <netinet/in.h>

#include "libfma.h"
#include "lf_fms_comm.h"
#include "lf_channel.h"
#include "lf_scheduler.h"
#include "lf_fabric.h"
#include "lf_topo_map.h"
#include "lf_fms.h"
#include "libmyri.h"

#include "fma.h"
#include "fma_fms.h"
#include "fma_myri.h"
#include "fma_fabric.h"
#include "fma_map.h"
#include "fma_resolve.h"
#include "fma_verify.h"
#include "fma_standalone.h"
#include "fma_settings.h"

/*
 * local functions
 */
static void try_fms_connection(void *);
static void lost_fms_conn(struct lf_channel *);
static void start_fms_message(struct lf_channel *);
static void finish_fms_message(struct lf_channel *);
static int fma_send_fma_version(int sock);

/*
 * Initialize FMS-related variables
 */
int
fma_init_fms_vars()
{
  LF_CALLOC(A.fms, struct fma_fms, 1);
  A.fms->socket = -1;
  return 0;

 except:
  return -1;
}

/*
 * Take a comma seperated list of hostnames
 * and build FMS array
 */
int
fma_set_fms_list(
  char *list)
{
  char *w[LF_STRING_LEN];
  int wc;

  line2words(list, w, ",", 0, &wc);
  if (wc < 1) {
    FMA_ERROR("Parsing fms host list");
  }

  /* free old hostname list */
  if (A.fms->num_hosts > 0) {
    lf_free_string_array(A.fms->host_list, A.fms->num_hosts);
  }

  /* create new hostname list */
  if (wc > 0) {
    A.fms->host_list = lf_dup_string_array(w, wc);
    if (A.fms->host_list == NULL) FMA_ERROR("Allocating FMS hostname list");
  }

  A.fms->num_hosts = wc;

  return 0;

 except:
  return -1;
}

/*
 * Try to connect to an FMS.
 * As long as we are disconnected, we will retry connecting to an FMS
 * via a socket.  It is possible, after we have established communication
 * to other FMAs, that we may be able to speak to an FMS through our
 * Myrinet connection, but keep trying for the socket connection anyhow, it
 * is preferable.
 *
 * The retry on the socket communication is accomplished by scheduling an event
 * in the future every time we fail to connect or lose connection.
 */
int
fma_init_fms()
{
  struct fma_fms *fms;
  struct lf_channel *chp;

  fms = A.fms;

  /* initiate a channel for this connection */
  LF_CALLOC(chp, struct lf_channel, 1);
  fms->chp = chp;
  chp->hangup_rtn = lost_fms_conn;
  chp->context = fms;

  /* initiate FMS connection attemps if server(s) specified */
  if (fms->num_hosts > 0) {
    try_fms_connection(NULL);
  }

  return 0;

 except:
  return -1;
}

/*
 * Try to make a connection to an FMS
 */
static void
try_fms_connection(
  void *x)
{
  struct lf_channel *chp;
  struct fma_fms *fms;
  uint8_t conn_type;
  int rc;

  fms = A.fms;

  /* Wait until connection queries are complete */
  if (!A.myri_ready) {
    printf("Waiting for Myrinet ready\n");
    goto retry_no_rotate;
  }

printf("trying connection to %s\n", fms->host_list[fms->host_index]); /* XXX */

  fms->socket = lf_connect_to_host(fms->host_list[fms->host_index],
				   FMS_SERVICE_PORT);

  /* if no connection, try again later */
  if (fms->socket == -1) goto retry;

  /* fill in the channel */
  chp = fms->chp;
  chp->fd = fms->socket;

  if (lf_add_channel(chp) != 0) LF_ERROR(("Adding channel"));

  /* set up to receive messages from FMS */
  lf_channel_receive(chp, &fms->header, sizeof(fms->header),
                     start_fms_message);

  /* let the FMS know this is an FMA connecting */
  conn_type = FMS_CONN_FMA;
  rc = write(fms->socket, &conn_type, sizeof(conn_type));
  if (rc != sizeof(conn_type)) goto retry;

  /* send startup messages */
  if (fma_send_fma_version(fms->socket) == -1) goto retry;
  if (fma_send_myri_info(fms->socket) == -1) goto retry;

  /* all set to go! */
printf("connection established\n"); /* XXX */
  return;

  /*
   * could not connect to FMS - reset error codes, advance to next FMS in the
   * list, and schedule a retry
   */
 retry:
  /* next FMS in list */
  ++fms->host_index;
  if (fms->host_index >= fms->num_hosts) {
    fms->host_index = 0;
  }

 retry_no_rotate:
  fma_reset_errors();

  /* make sure the socket is closed */
  if (fms->socket != -1) {
    close(fms->socket);
    fms->socket = -1;
  }

  /* schedule a retry */
  if (lf_schedule_event(try_fms_connection, NULL, FMA_FMS_RETRY_TIME) == NULL) {
    FMA_ERROR("Scheduling fms connect event");
  }
  return;

 except:
  fma_perror();
  fma_exit(1);
}

/*
 * Called when we lose connection to the FMS server
 */
static void
lost_fms_conn(
  struct lf_channel *chp)
{
  /* Don't free things more than once if FMS is gone */
  if (A.fms->socket == -1) {
    return;
  }

  /* remove the channel from service and retry the connection */
  lf_remove_channel(chp);

  close(A.fms->socket);
  A.fms->socket = -1;

  /* next FMS in list */
  ++A.fms->host_index;
  if (A.fms->host_index >= A.fms->num_hosts) {
    A.fms->host_index = 0;
  }

  /* schedule a retry */
  if (lf_schedule_event(try_fms_connection, NULL, FMA_FMS_RETRY_TIME) == NULL) {
    FMA_ERROR("Scheduling fms connect event");
  }
  return;

 except:
  fma_perror();
  fma_exit(1);
}

/*
 * Tell the FMS what version of software we are running
 */
static int
fma_send_fma_version(
  int sock)
{
  struct fma_fms_ident_msg msg;
  struct fms_msg_header hdr;
  int rc;

  /* fill in the message header */
  hdr.length_32 = htonl(sizeof(msg));
  hdr.msg_type_32 = htonl(FMA_FMS_IDENT_MSG);
  
  /* fill in FMA version number */
  msg.fma_version_32 = htonl(LF_VERSION);

  /* fill in host info */
  strcpy(msg.hostname, A.hostname);

  /* send the messages */
  rc = lf_write(sock, &hdr, sizeof(hdr));
  if (rc != sizeof(hdr)) return -1;
  rc = lf_write(sock, &msg, sizeof(msg));
  if (rc != sizeof(msg)) return -1;

  return 0;
}

/*
 * Tell the FMS that the map i have received is invalid
 */
void
fma_fms_map_is_invalid()
{
  int rc;

  /* send this header-only message to FMS */
  rc = lf_send_fms_header(A.fms->socket, FMA_FMS_MAP_IS_INVALID, 0);
  if (rc != 0) {
    fma_reset_error();
    lost_fms_conn(A.fms->chp);
  }
  return;
}

/*
 * Called when we get a message header
 */
static void
start_fms_message(
  struct lf_channel *chp)
{
  struct fma_fms *fms;
  int len;
  void *p;

  fms = chp->context;

  /* make sure message buffer is big enough to hold incoming message */
  len = ntohl(fms->header.length_32);

  /* if message length is 0, just handle it */
  if (len == 0) {
    
    finish_fms_message(chp);

  /* non-zero-length message, get the rest */
  } else {

    if (len > fms->msgbuf_size) {
      p = realloc(fms->msgbuf, len);
      if (p == NULL) LF_ERROR(("Error allocating space for incoming message"));

      fms->msgbuf = (union lf_fma_message *)p;
      fms->msgbuf_size = len;
    }

    /* get the rest of the message */
    lf_channel_receive(chp, fms->msgbuf, len, finish_fms_message);
  }
  return;

 except:
  fma_perror_exit(1);
}

/*
 * Called when a complete message has arrived
 */
static void
finish_fms_message(
  struct lf_channel *chp)
{
  struct fma_fms *fms;
  int length;
  int type;

  fms = chp->context;
  type = ntohl(fms->header.msg_type_32);
  length = ntohl(fms->header.length_32);

  /* process the message */
  fma_handle_message(type, length, fms->msgbuf);

  /* set up to receive next message from FMS */
  lf_channel_receive(chp, &fms->header, sizeof(fms->header),
                     start_fms_message);
  return;
}

/*
 * Handle an incoming message
 */
void
fma_handle_message(
  int type,
  int length,
  union lf_fma_message *msg)
{
  int rc;

  /* switch on message type */
  switch (type) {
  case LF_FMA_TERMINATE:
    {
      char buf[2*LF_STRING_LEN];
      sprintf(buf, "Exiting at request of FMS, reason:\n\t%s",
	      msg->terminate.reason);
      FMA_ERROR(buf);
    }
    break;

  case LF_FMA_TOPO_MAP:
    if (length > 0) {

      fma_load_map((struct lf_topo_map *)msg);
      rc = fma_validate_map();

      /* map is OK, process it */
      if (rc == 0) {
	rc = fma_process_topo_map();

	/* print any errors */
	if (rc != 0) {
	  fma_perror();
	}

	/* if standalone, indicate new map arrived */
	if (A.standalone) {
	  fma_standalone_got_map();
	}

      /* map is not valid */
      } else {
	printf("sending \"map invalid\"\n");
	fma_map_is_invalid();
      }
    } else {
      printf("Ignoring zero-length map\n");
    }
    break;

  case LF_FMA_SEND_INVALID_ROUTE:
    if (length > 0) {
      uint32_t xbar_index;

      xbar_index = ntohl(*(uint32_t *)msg);
      rc = fma_send_inv_rts(xbar_index);

      if (rc != 0) {
        printf("invalid route send failed for xbar %d\n", xbar_index);
      }
    } else {
      fma_perror();
    }
    break;

  case LF_FMA_TOPO_LINK_STATE:
    fma_load_link_state((lf_topo_link_state_t *)msg);

    /* calculate the routes */
    rc = fma_process_topo_map();
    if (rc != 0) {
      fma_perror();
    } else {
      printf("routes re-set\n");
    }
    break;

  case LF_FMA_CLEAR_PROBES:
    fma_clear_verify_probes();
    break;

  case LF_FMA_DEFINE_PROBE:
    fma_define_verify_probe_from_msg(&msg->defprobe);
    break;

  case LF_FMA_SETTINGS:
    fma_fms_set_settings(&msg->settings);
    break;

  case LF_FMA_MAP_FABRIC:
    printf("got request to map fabric\n");

    if (A.standalone) {
      fma_start_mapping_fabric(fma_standalone_map_done, TRUE);
    } else {
      /* start mapping, FALSE means not standalone */
      fma_start_mapping_fabric(fma_mf_send_map_to_fms, FALSE);
    }
    break;

  default:
    FMA_ERROR("Unknown message type from FMS");
    break;
  }
  return;

 except:
  fma_perror_exit(1);
}

/*
 * Report a NIC error
 */
void
fma_fms_report_nic_error(
  int nic_id,
  enum myri_error_type error)
{
  struct fma_fms_nic_error_msg msg;
  int sock;
  int rc;

  sock = A.fms->socket;

  /* Make sure there is someone to talk to */
  if (sock == -1) {
    LF_ERROR(("No FMS to which to report NIC error"));
  }

  /* record the NIC ID */
  msg.nic_id_32 = htonl(nic_id);
  msg.error_32 = htonl(error);

  /* send this message header to FMS */
  rc = lf_send_fms_header(sock, FMA_FMS_NIC_ERROR, sizeof(msg));
  if (rc != 0) goto error;

  /* send the message itself */
  rc = lf_write(sock, &msg, sizeof(msg));
  if (rc != sizeof(msg)) goto error;
  return;

 error:
  LF_ERROR(("Error reporting NIC error to FMS"));
 except:
  fma_perror();
}

/*
 * badcrc threshold exceeded
 */
void
fma_fms_nic_badcrc(
  int nic_id,
  int port,
  int badcrcs)
{
  struct fma_fms_nic_badcrc_msg msg;
  int sock;
  int rc;

  sock = A.fms->socket;

  /* Make sure there is someone to talk to */
  if (sock == -1) {
    LF_ERROR(("No FMS to which to report badcrc alert"));
  }

  /* record the NIC ID */
  msg.nic_id_32 = htonl(nic_id);
  msg.port_32 = htonl(port);
  msg.badcrcs_32 = htonl(badcrcs);

  /* send this message header to FMS */
  rc = lf_send_fms_header(sock, FMA_FMS_NIC_BADCRC, sizeof(msg));
  if (rc != 0) goto error;

  /* send the message itself */
  rc = lf_write(sock, &msg, sizeof(msg));
  if (rc != sizeof(msg)) goto error;
  return;

 error:
  LF_ERROR(("Error reporting NIC badcrcs to FMS"));
 except:
  fma_perror();
}
