/*
 * Routines for dealing with the map
 */
#include <sys/types.h>
#include <netinet/in.h>

#include "libfma.h"
#include "lf_fabric.h"
#include "lf_topo_map.h"
#include "lf_fms_comm.h"
#include "libmyri.h"

#include "fma.h"
#include "fma_fms.h"
#include "fma_fabric.h"
#include "fma_myri.h"
#include "fma_map.h"
#include "fma_dfs_route.h"
#include "fma_standalone.h"
#include "fma_probe.h"
#include "fma_verify.h"

/*
 * Local prototypes
 */
static void fma_free_fabric_contents(struct lf_fabric *fp);
static void fma_alloc_route_buffers(struct lf_fabric *fp);

/*
 * We got a new map from the FMS, translate it into a fabric struct
 */
void
fma_load_map(
  struct lf_topo_map *topo_map)
{
  struct lf_fabric *fp;
  int rc;

  /* get pointer to map holder */
  fp = A.fabric;
  if (fp == NULL) {			/* allocate holder if none */
    LF_CALLOC(fp, struct lf_fabric, 1);
    LF_CALLOC(fp->user, struct fma_fabric, 1);
    A.fabric = fp;
  }

  fma_free_fabric_contents(fp);		/* free any previous map */

  printf("  %d hosts, %d nics, %d xbars, %d links\n",
    ntohl(topo_map->num_hosts_32), ntohl(topo_map->num_nics_32),
    ntohl(topo_map->num_xbars_32), ntohl(topo_map->num_links_32));

  /* copy over counts for xbars and NICs */
  fp->num_hosts = ntohl(topo_map->num_hosts_32);
  fp->num_xbars = ntohl(topo_map->num_xbars_32);
  FMA_FABRIC(fp)->num_nics = ntohl(topo_map->num_nics_32);
  FMA_FABRIC(fp)->num_links = ntohl(topo_map->num_links_32);

  /* allocate space for xbars and NICs */
  LF_CALLOC(fp->hosts, struct lf_host *, fp->num_hosts);
  LF_CALLOC(fp->xbars, struct lf_xbar *, fp->num_xbars);
  LF_CALLOC(FMA_FABRIC(fp)->nics, struct lf_nic *, FMA_FABRIC(fp)->num_nics);
  LF_CALLOC(FMA_FABRIC(fp)->link_end_node, struct lf_node *,
            FMA_FABRIC(fp)->num_links);
  LF_CALLOC(FMA_FABRIC(fp)->link_end_port, char, FMA_FABRIC(fp)->num_links);

  /* Allocate all the hosts */
  {
    struct lf_host *hp;
    int h;

    for (h=0; h<fp->num_hosts; ++h) {
      hp = fma_create_host();
      if (hp == NULL) LF_ERROR(("Error allocating host"));
      fp->hosts[h] = hp;

      FMA_HOST(hp)->host_id = h;	/* assign host index */
    }
  }

  /* copy over all the xbars */
  {
    struct lf_topo_xbar *topo_xbar;
    int x;

    topo_xbar = LF_TOPO_XBAR_ARRAY(topo_map);

    for (x=0; x<fp->num_xbars; ++x) {
      struct lf_xbar *xp;

      xp = fma_create_xbar(ntohs(topo_xbar->num_ports_16),
	                   ntohl(topo_xbar->xbar_id_32),
			   topo_xbar->quadrant_disable_8);
      if (xp == NULL) FMA_ERROR("Creating xbar");

      fp->xbars[x] = xp;

      /* fill in stuff from map */
      xp->clos_level = topo_xbar->clos_level_8;
      FMA_XBAR(xp)->node_id = ntohl(topo_xbar->node_id_32);
      FMA_XBAR(xp)->map_index = x;

      ++topo_xbar;
    }
  }
  
  /* copy over all the NICs */
  {
    struct lf_topo_nic *topo_nic;
    int nic_ports;
    int n;
    int h;

    topo_nic = LF_TOPO_NIC_ARRAY(topo_map);
    nic_ports = 0;		/* remember most ports on any NIC */

    for (n=0; n<FMA_FABRIC(fp)->num_nics; ++n) {
      struct lf_nic *nicp;
      struct lf_host *hp;

      nicp = fma_create_nic(ntohs(topo_nic->num_ports_16),
	                    topo_nic->mac_addr);
      if (nicp == NULL) FMA_ERROR("Creating nic");

      FMA_FABRIC(fp)->nics[n] = nicp;
      FMA_NIC(nicp)->map_index = n;
      FMA_NIC(nicp)->node_id = ntohl(topo_nic->node_id_32);

      /* use node ID to get index of host */
      h = LF_TOPO_NIC_HOST_ID(FMA_NIC(nicp)->node_id);
      hp = fp->hosts[h];
      rc = lf_add_existing_nic_to_host(nicp, hp);
      if (rc != 0) LF_ERROR(("Error adding NIC to host"));
      nicp->host_nic_id = LF_TOPO_NIC_HOST_NIC_ID(FMA_NIC(nicp)->node_id);

      /* Make sure "max_mac_addr" is up-to-date for owning host */
      if (LF_MAC_CMP(nicp->mac_addr, FMA_HOST(hp)->max_mac_addr) > 0) {
	LF_MAC_COPY( FMA_HOST(hp)->max_mac_addr, nicp->mac_addr);

	/* Look for us in the map */
	if (LF_MAC_CMP(A.stand->my_max_mac, nicp->mac_addr) == 0) {
	  A.my_host = hp;
	}
      }

      if (nicp->num_ports > nic_ports) {
	nic_ports = nicp->num_ports;
      }

      ++topo_nic;
    }

    FMA_FABRIC(fp)->max_nic_ports = nic_ports;	/* for allocating route space */
  }


  /* allocate space for routes */
  fma_alloc_route_buffers(fp);

  /* Now, make all the links */
  {
    struct lf_topo_link *topo_link;
    lf_topo_link_state_t *topo_link_state;
    int l;

    topo_link = LF_TOPO_LINK_ARRAY(topo_map);
    topo_link_state = LF_TOPO_LINK_STATE_ARRAY(topo_map);

    for (l=0; l<FMA_FABRIC(fp)->num_links; ++l) {
      struct lf_node *np[2];
      int port[2];
      int i;

      /* get pointers to the two nodes */
      for  (i=0; i<2; ++i) {
	int node_index;

	node_index = ntohl(topo_link->node_index_32[i]);

	if (LF_TOPO_NI_XBAR(node_index)) {
	  np[i] = LF_NODE(fp->xbars[LF_TOPO_NI_INDEX(node_index)]);
	} else {
	  np[i] = LF_NODE(FMA_FABRIC(fp)->nics[LF_TOPO_NI_INDEX(node_index)]);
	}
	port[i] = LF_TOPO_NI_PORT(node_index);
      }

      /* link them to each other */
      for  (i=0; i<2; ++i) {
	if (np[i]->type == LF_NODE_XBAR) {
	  struct lf_xbar *xp;
	  xp = LF_XBAR(np[i]);

	  xp->topo_ports[port[i]] = np[1-i];
	  xp->topo_rports[port[i]] = port[1-i];
	} else {
	  struct lf_nic *nicp;
	  nicp = LF_NIC(np[i]);

	  nicp->topo_ports[port[i]] = np[1-i];
	  nicp->topo_rports[port[i]] = port[1-i];
	}
      }

      /* set the link state */
      FMA_FABRIC(fp)->link_end_node[l] = np[0];
      FMA_FABRIC(fp)->link_end_port[l] = port[0];
      fma_set_link_state(np[0], port[0], *topo_link_state);

      ++topo_link;
      ++topo_link_state;
    }
  }

  fma_find_nic_indices();	/* update NIC indices */
  fma_calc_xbar_dists();	/* calc distance from each xbar to this host */

  return;

 except:
  fma_perror_exit(1);
}

/*
 * We got a new link state from the FMS, update our map with the new states
 */
void
fma_load_link_state(
  lf_topo_link_state_t *topo_link_state)
{
  struct lf_fabric *fp;
  int l;

  /* get pointer to fabric */
  fp = A.fabric;

  /* set the link state for each link */
  for (l=0; l<FMA_FABRIC(fp)->num_links; ++l) {

    fma_set_link_state(FMA_FABRIC(fp)->link_end_node[l],
		       FMA_FABRIC(fp)->link_end_port[l],
		       topo_link_state[l]);
  }

  /* xbar distances may have changed */
  fma_calc_xbar_dists();	/* calc distance from each xbar to this host */
}

/*
 * set the state on a link
 */
void
fma_set_link_state(
  struct lf_node *np,
  int port,
  int state)
{
  lf_node_t *onp;
  int oport;
  int down;

  /* if state is anything other than up, it is down for our purposes */
  down = (state != LF_LINK_STATE_UP);

  /* mark this end of the link */
  if (np->type == LF_NODE_XBAR) {
    struct lf_xbar *xp;

    xp = LF_XBAR(np);
if (FMA_XBAR(xp)->link_down[port] != down) printf("%x:p%d is %s\n", FMA_XBAR(xp)->node_id, port, down? "down" : "up");
    FMA_XBAR(xp)->link_down[port] = down;
    onp = xp->topo_ports[port];
    oport = xp->topo_rports[port];
  } else {
    struct lf_nic *nicp;

    nicp = LF_NIC(np);
    FMA_NIC(nicp)->link_down[port] = down;
    onp = nicp->topo_ports[port];
    oport = nicp->topo_rports[port];
  }

  /* now mark other end of link */
  if (onp->type == LF_NODE_XBAR) {
if (FMA_XBAR(onp)->link_down[oport] != down) printf("%x:p%d is %s\n", FMA_XBAR(onp)->node_id, oport, down? "down" : "up");
    FMA_XBAR(onp)->link_down[oport] = down;
  } else {
    FMA_NIC(onp)->link_down[oport] = down;
  }
}


/*
 * Free all the components in a map
 */
static void
fma_free_fabric_contents(
  struct lf_fabric *fp)
{
  int i;

  /* free all the xbars */
  for (i=0; i<fp->num_xbars; ++i) {
    struct lf_xbar *xp;

    xp = fp->xbars[i];
    fma_free_xbar(xp);
  }
  LF_FREE(fp->xbars);
  fp->num_xbars = 0;

  /* free all the hosts and nics */
  for (i=0; i<fp->num_hosts; ++i) {
    struct lf_host *hp;

    hp = fp->hosts[i];
    fma_free_host(hp);	/* also frees NIC */
  }
  LF_FREE(FMA_FABRIC(fp)->nics);
  LF_FREE(fp->hosts);
  FMA_FABRIC(fp)->num_nics = 0;
  fp->num_hosts = 0;

  /* No host struct for us anymore */
  A.my_host = NULL;
}

/*
 * Allocate all route buffers
 */
static void
fma_alloc_route_buffers(
  struct lf_fabric *fp)
{
  int n;
  int size;
  int num_lens;

  num_lens =  FMA_FABRIC(fp)->max_nic_ports * FMA_FABRIC(fp)->max_nic_ports * A.myri->num_nics *
    A.myri->nic_info[0]->myri_info.num_routes;      
  size = num_lens * FMA_IFC_ROUTE_LEN;

  for (n=0; n<FMA_FABRIC(fp)->num_nics; ++n) {
    struct lf_nic *nicp;

    nicp = FMA_FABRIC(fp)->nics[n];
    LF_CALLOC(FMA_NIC(nicp)->route_buf, unsigned char, size);
    LF_CALLOC(FMA_NIC(nicp)->route_len, unsigned char, num_lens);
  }
  return;

 except:
  fma_perror_exit(1);
}


/*
 * Find a NIC index given mac address
 */
int
fma_mac_to_nic_index(
  lf_mac_addr_t mac)
{
  struct lf_fabric *fp;
  int n;

  fp = A.fabric;
  for (n=0; n<FMA_FABRIC(fp)->num_nics; ++n) {
    if (memcmp(mac, FMA_FABRIC(fp)->nics[n]->mac_addr, sizeof(lf_mac_addr_t)) == 0) {
      return n;
    }
  }
  return -1;
}

/*
 * Find our NICs in this map
 */
void
fma_find_nic_indices()
{
  struct fma_myri *mp;
  struct lf_fabric *fp;
  int n;

  fp = A.fabric;
  mp = A.myri;

  /* find each NIC in map */
  for (n=0; n<mp->num_nics; ++n) {
    struct fma_nic_info *nip;
    
    nip = mp->nic_info[n];
    nip->map_index = fma_mac_to_nic_index(nip->myri_info.mac_addr);

    /* mark NIC as local to this host */
    FMA_NIC(FMA_FABRIC(fp)->nics[nip->map_index])->local_nip = nip;

    if (nip->map_index == -1) {
      fprintf(stderr, "Cannot find NIC %d (" LF_MAC_FORMAT ") in map!\n",
	nip->nic_id, LF_MAC_ARGS(nip->myri_info.mac_addr));
      nip->nic_ptr = NULL;
    } else {
      fprintf(stderr, "Found NIC %d at index %d!\n",
	      nip->nic_id, nip->map_index);
      nip->nic_ptr = FMA_FABRIC(A.fabric)->nics[nip->map_index];
    }
  }
}

/*
 * Generate routes for the current topo map
 */
int
fma_process_topo_map()
{
  int rc;
  int i;
  int clear;

  rc = 0;

#if 0
  fma_calc_all_quick_routes();
#else

  /* calculate the routes */
  for (i=0; i<A.myri->nic_info[0]->myri_info.num_routes; ++i) {
    clear = (i == 0);
    if (fma_dfs_calc_all_routes(clear, i) != 0) {
      rc = 1;
    }
  }
#endif

  /* if successful, set the routes */
  if (rc == 0) {
    fma_set_all_routes();
    printf("routes set\n");
  }

  return rc;
}

/*
 * validate map
 */
int
fma_validate_map()
{
  struct fma_myri *fmp;
  struct lf_fabric *fp;
  int n;

  fp = A.fabric;

  fmp = A.myri;
  for (n=0; n<fmp->num_nics; ++n) {
    struct fma_nic_info *nip;
    struct lf_nic *nicp;
    int p;

    nip = fmp->nic_info[n];

    /* Is this NIC even in the map? */
    if (nip->map_index == -1) {

      /* not in map - if connected, map is invalid */
      if (nip->xbar_id[0] != -1) {
	printf("map does not see my connected NIC\n");
	return -1;
      }

    /* found NIC in map - compare connectedness */
    } else {

      nicp = FMA_FABRIC(fp)->nics[nip->map_index];
      for (p=0; p<nicp->num_ports; ++p) {
	struct lf_xbar *xp;
	int xport;

	xp = LF_XBAR(nicp->topo_ports[p]);
	xport = nicp->topo_rports[p];

	/* if we don't believe we are disconnected but map does,
	 * then call it invalid
	 */
	if (nip->xbar_id[p] != -1 && xp == NULL) {
	  printf("map shows disconnected NIC %d, port %d\n", n, p);
	  return -1;
	}

	/* make sure we have a running verify on direct xbar */
	/* XXX - should be checking for link coming back if down */
	if (xp != NULL && xp->n.type == LF_NODE_XBAR) {
	  fma_define_direct_xbar_verify_probe(nip, p);
	}
      }
    }
  }

  printf("map seems OK\n");
  return 0;
}

struct lf_host *
fma_find_host_by_max_mac_addr(
  lf_mac_addr_t max_mac_addr)
{
  struct lf_fabric *fp;
  int h;

  fp = A.fabric;

  /* if no map, then definitely no host to be found */
  if (fp == NULL) return NULL;

  /* look through all hosts for a match */
  for (h=0; h<fp->num_hosts; ++h) {
    struct lf_host *hp;

    hp = fp->hosts[h];
    if (LF_MAC_CMP(FMA_HOST(hp)->max_mac_addr, max_mac_addr) == 0) {
      return hp;
    }
  }
  return NULL;
}

/*
 * Map is invalid - call the right routine
 */
void
fma_map_is_invalid()
{
  if (A.standalone) {
    fma_standalone_map_is_invalid();
  } else {
    fma_fms_map_is_invalid();
  }
}
