/*
 * routines to map a myrinet fabric
 */
#include <stdio.h>
#include <netinet/in.h>
#include <sys/types.h>

#include "libfma.h"
#include "lf_fabric.h"
#include "lf_xbar32.h"
#include "lf_myri_packet.h"
#include "lf_scheduler.h"
#include "lf_topo_map.h"
#include "lf_fms_comm.h"
#include "libmyri.h"

#include "fma.h"
#include "fma_myri.h"
#include "fma_fms.h"
#include "fma_fabric.h"
#include "fma_probe.h"
#include "fma_map_fabric.h"
#include "fma_standalone.h"

/*
 * local prototypes
 */
static void fma_start_mapping_next_nic_port();
static void fma_mf_check_for_more_probes();
static void fma_mf_scout_frontier(struct fma_nic_info *nip, int p,
                                  struct lf_node *np, int port);
static void fma_mf_init_xbar_private(struct lf_xbar *xp,
    struct fma_nic_info *nip, int nic_port, unsigned char *route,
    int route_len, int in_port);
static void fma_mf_init_nic_private(struct lf_nic *nicp,
    lf_mac_addr_t max_mac_addr);
static void fma_mf_add_nic_to_map(struct lf_nic *nicp);
static void fma_mf_add_xbar_to_map(struct lf_xbar *xp);
static void fma_mf_add_unresolved_xbar(struct lf_xbar *xp);
static void fma_mf_unlink_probe(struct fma_probe_desc *pdp);
static int fma_mf_scout_list_empty(struct lf_node *np, int port);
static struct fma_mf_info *fma_mf_get_mf_info(struct lf_node *np, int port);
static void fma_mf_scout_timeout(struct fma_probe_desc *pdp);
static void fma_mf_node_port_resolved(struct lf_node *np, int port);
static void fma_mf_got_nic_resp(struct fma_probe_desc *pdp,
                                struct lf_myri_packet *pkt, int len);
static void fma_mf_clear_probes(struct lf_node *np, int port, int type);
static void fma_mf_got_xbar_resp(struct fma_probe_desc *pdp,
                                 struct lf_myri_packet *pkt, int len);
static void fma_mf_got_x32_resp(struct fma_probe_desc *pdp,
    struct lf_myri_packet *pkt, int len);
static void fma_mf_connect_nodes(struct lf_node *np1, int port1,
                                 struct lf_node *np2, int port2, int resolved);
static void fma_mf_explore_xbar(struct lf_xbar *xp);
static void fma_mf_set_compare_timer(void *vxp);
static void fma_mf_start_xbar_compares(struct lf_xbar *xp, int port);
static int fma_mf_send_xbar_compares(struct lf_xbar *xp, int port);
static void fma_mf_start_next_xbar_compares(void);
static void fma_mf_compare_xbar_port(struct lf_xbar *xp, int port,
                                     struct lf_xbar *txp, int tport);
static void fma_mf_compare_timeout(void *vxp);
static void fma_mf_xbar_is_unique(struct lf_xbar *xp);
static void fma_mf_destroy_xbar(struct lf_xbar *xp);
static void fma_mf_dereference_xbar(struct lf_xbar *xp);
static struct lf_xbar *fma_mf_find_xbar_by_xbar_id(int xbar_id);
static struct lf_xbar *fma_mf_find_xbar_by_node_id(int node_id);
static void fma_mf_merge_xbars(struct lf_xbar *cxp, int cport,
                                       struct lf_xbar *nxp, int nport);
static void fma_mf_unlink_xbar_from_mf_list(struct lf_xbar *xp);
static void fma_mf_stop_compares(struct lf_xbar *xp);
static void fma_mf_mapping_complete(void);
static void fma_mf_make_topo_map(void);
static void fma_mf_free_mapping(void);
static struct lf_nic *fma_mf_find_nic_by_mac(struct fma_map_progress *mpp,
                                             lf_mac_addr_t mac_addr);
static int fma_mf_get_host_index(lf_mac_addr_t mac, int *nic_id);

#define mf_debug (A.debug>1)

static void dump_map(void);

char *route_str(
  unsigned char *route,
  int route_len)
{
  static lf_string_t s;
  char *p;
  int i;

  p = s;
  *p = '\0';
  for (i=0; i<route_len; ++i) {
    p += sprintf(p, "%d ", LF_ROUTE_TO_DELTA(route[i]));
  }
  return s;
}

static int
fma_mf_get_host_index(
  lf_mac_addr_t mac,
  int *nic_id)
{
  struct fma_map_progress *mpp;
  struct fma_mf_host_record *hrp;
  int h;

  mpp = A.myri->map_progress;

  /* find it if we can */
  hrp = mpp->host_table;
  for (h=0; h<mpp->num_hosts; ++h) {
    if (LF_MAC_CMP(mac, hrp->max_mac) == 0) {
      *nic_id = hrp->num_nics++;
      return h;
    }
    ++hrp;
  }

  /*
   * We have a new MAC address.  First, make sure there is room
   * in the table and grow it if not.
   */
  if (h <= mpp->host_table_size) {
    mpp->host_table_size += 100;
    mpp->host_table = (struct fma_mf_host_record *) realloc(mpp->host_table,
	sizeof(struct fma_mf_host_record) * mpp->host_table_size);
    if (mpp->host_table == NULL) {
      LF_ERROR(("Error growing host_table"));
    }
  }

  /* copy in the mac address */
  hrp = &mpp->host_table[h];
  LF_MAC_COPY(&hrp->max_mac, mac);
  hrp->num_nics = 1;
  ++mpp->num_hosts;
  
  *nic_id = 0;		/* first NIC is ID 0 */
  return h;

 except:
  fma_perror_exit(1);
  return -1;
}

/*
 * Generate a map of the fabric using node and xbar queries
 */
void
fma_start_mapping_fabric(
  void (*map_complete_rtn)(struct lf_topo_map *topo_map, int topo_map_size),
  int standalone)
{
  struct fma_myri *fmp;
  struct fma_nic_info *nip;
  struct fma_map_progress *mpp;
  int n;
  int p;

  fmp = A.myri;

  /* don't start again if one mapping already in progress */
  if (fmp->map_progress != NULL) {
    return;
  }

  printf("Mapping fabric...\n");

  /* start the new map, allocate and fill in a progress struct */
  LF_CALLOC(mpp, struct fma_map_progress, 1);
  fmp->map_progress = mpp; /* save away map progress pointer */
  LF_CALLOC(mpp->my_nics, struct lf_nic *, fmp->num_nics);

  /* Increment session ID */
  ++fmp->map_session_id;

  /* clear the list anchors */
  mpp->xbar_anchor.mf_next = &mpp->xbar_anchor;
  mpp->xbar_anchor.mf_prev = &mpp->xbar_anchor;
  mpp->unresolved_xbar_anchor.mf_next = &mpp->unresolved_xbar_anchor;
  mpp->unresolved_xbar_anchor.mf_prev = &mpp->unresolved_xbar_anchor;
  mpp->nic_anchor.mf_next = &mpp->nic_anchor;
  mpp->nic_anchor.mf_prev = &mpp->nic_anchor;

  /* save completion routine */
  mpp->map_complete_rtn = map_complete_rtn;

  /* Decide what NIC scout reason to use */
  mpp->nic_scout_reason = standalone ? LF_NSR_STANDALONE_MAP : LF_NSR_FMS_MAP;

  /* clear mapped flags on all our NICs */
  for (n=0; n<fmp->num_nics; ++n) {
    struct myri_nic_info *mip;
    struct lf_nic *nicp;

    nip = fmp->nic_info[n];
    mip = &nip->myri_info;

    for (p=0; p<mip->num_ports; ++p) {
      nip->mf_mapped[p] = FALSE;
    }

    /* create the NIC struct */
    nicp = fma_create_nic(mip->num_ports, mip->mac_addr);
    if (nicp == NULL) LF_ERROR(("Error creating NIC"));
    mpp->my_nics[n] = nicp;

    /* Fill FMA-private structs */
    fma_mf_init_nic_private(mpp->my_nics[n], A.stand->my_max_mac);

    fma_mf_add_nic_to_map(mpp->my_nics[n]);	/* add it to the map */
  }

  /* set some bookkeeping fields in the map descriptor */
  mpp->nic_index = 0;
  mpp->nic_port = 0;

  /* now, start mapping on the next next NIC */
  fma_start_mapping_next_nic_port();

  return;

 except:
  if (mpp != NULL) LF_FREE(mpp->my_nics);
  LF_FREE(fmp->map_progress);
  fma_perror_exit(1);
}

static void
fma_start_mapping_next_nic_port()
{
  int n;
  struct fma_map_progress *mpp;
  struct fma_myri *fmp;

  fmp = A.myri;
  mpp = fmp->map_progress;

  for (n=0; n<fmp->num_nics; ++n) {
    struct fma_nic_info *nip;
    int p;

    nip = fmp->nic_info[n];
    for (p=0; p<nip->myri_info.num_ports; ++p) {
      if (!nip->mf_mapped[p]) {
	fma_mf_scout_frontier(nip, p, LF_NODE(mpp->my_nics[n]), p);
	nip->mf_mapped[p] = TRUE;
	return;
      }
    }
  }

  fma_mf_mapping_complete();		/* mapping is all done */
}

/*
 * If there are no more pending probes and there are no unresolved xbars,
 * then it is time to move on to the next NIC
 */
static void
fma_mf_check_for_more_probes()
{
  struct fma_map_progress *mpp;

  mpp = A.myri->map_progress;

  /* If no outstanding probes and no unresolved anchors, go to next NIC */
  if (mpp->mf_outstanding_probes <= 0 &&
      mpp->unresolved_xbar_anchor.mf_next == &mpp->unresolved_xbar_anchor) {
    fma_start_mapping_next_nic_port();
  }
}

/*
 * Send three probes for each frontier port - x32, NIC, and xbar, in that order.  If an xbar
 * is not dumb, we want the ID back before the simple echo.
 */
static void
fma_mf_scout_frontier(
  struct fma_nic_info *nip,
  int p,
  struct lf_node *np,
  int port)
{
  unsigned char route[FMA_IFC_ROUTE_LEN*2+1];
  int route_len;
  struct fma_probe_desc *pdp;
  struct fma_probe_desc *probe_anchor;
  struct fma_mf_scout *msp;
  struct fma_map_progress *mpp;
  struct fma_mf_info *mfip;

  mpp = A.myri->map_progress;

  /*
   * If coming from a NIC, route is empty, else it's the route to this
   * xbar plus a byte to leave by the appropriate port
   */
  if (np->type == LF_NODE_NIC) {
    route_len = 0;

  } else if (np->type == LF_NODE_XBAR) {
    struct fma_xbar *fxp;

    /* add one byte to route to this xbar */
    fxp = FMA_XBAR(np);
    route_len = fxp->mf_route_len+1;
    memcpy(route, fxp->mf_route, fxp->mf_route_len);
    route[route_len-1] = LF_DELTA_TO_ROUTE(port - fxp->mf_in_port);
    
  } else {
    LF_ERROR(("Bad node type"));
  }

  /* get probe anchor */
  mfip = fma_mf_get_mf_info(np, port);
  probe_anchor = &mfip->mf_probe_anchor;

  /*
   * start the x32 probe unless we know there are none
   */
  if (A.xbar_types != FMA_XT_NO_IDS) {
    pdp = fma_create_x32_scout_probe(nip, p, FMA_PT_MF_X32_SCOUT,
				     fma_mf_got_x32_resp,
				     fma_mf_scout_timeout,
				     route, route_len);

    /* associate this probe with frontier node */
    msp = &pdp->u.mf_scout;
    msp->frontier_node = np;
    msp->frontier_port = port;
    pdp->user_next = probe_anchor->user_next;
    pdp->user_prev = probe_anchor;
    pdp->user_next->user_prev = pdp;
    pdp->user_prev->user_next = pdp;
    ++mpp->mf_outstanding_probes;
    
    /* send the packet */
    fma_send_probe_packet(pdp);

    /* link this into list of active probes */
    fma_link_probe_descriptor(pdp);
  }

  /*
   * start the NIC probe
   */
  pdp = fma_create_nic_scout_probe(nip, p, FMA_PT_MF_NIC_SCOUT,
                                   fma_mf_got_nic_resp,
				   fma_mf_scout_timeout,
				   route, route_len,
				   mpp->nic_scout_reason);

  /* associate this probe with frontier node */
  msp = &pdp->u.mf_scout;
  msp->frontier_node = np;
  msp->frontier_port = port;
  pdp->user_next = probe_anchor->user_next;
  pdp->user_prev = probe_anchor;
  pdp->user_next->user_prev = pdp;
  pdp->user_prev->user_next = pdp;
  ++mpp->mf_outstanding_probes;

  /* send the packet */
  fma_send_probe_packet(pdp);

  /* link this into list of active probes */
  fma_link_probe_descriptor(pdp);

  /*
   * start the xbar probe unless we know there are none
   */
  if (A.xbar_types != FMA_XT_ID_ONLY) {
    pdp = fma_create_xbar_scout_probe(nip, p, FMA_PT_MF_XBAR_SCOUT,
				     fma_mf_got_xbar_resp,
				     fma_mf_scout_timeout,
				     route, route_len);

    /* associate this probe with frontier node */
    msp = &pdp->u.mf_scout;
    msp->frontier_node = np;
    msp->frontier_port = port;
    pdp->user_next = probe_anchor->user_next;
    pdp->user_prev = probe_anchor;
    pdp->user_next->user_prev = pdp;
    pdp->user_prev->user_next = pdp;
    ++mpp->mf_outstanding_probes;

    /* send the packet */
    fma_send_probe_packet(pdp);

    /* link this into list of active probes */
    fma_link_probe_descriptor(pdp);
  }

  return;

 except:
  dump_map();
  fma_perror_exit(1);
}

/*
 * Fill in the MF-private fields of an xbar
 */
static void
fma_mf_init_xbar_private(
  struct lf_xbar *xp,
  struct fma_nic_info *nip,
  int nic_port,
  unsigned char *route,
  int route_len,
  int in_port)
{
  int p;
  struct fma_mf_info *mfip;

  /* array of flags for recording xbar scout responses */
  LF_CALLOC(mfip, struct fma_mf_info, xp->num_ports);
  FMA_XBAR(xp)->mf_info = mfip;

  /* initialize mapping probe list anchors */
  for (p=0; p<xp->num_ports; ++p) {
    mfip->mf_probe_anchor.user_next = &mfip->mf_probe_anchor;
    mfip->mf_probe_anchor.user_prev = &mfip->mf_probe_anchor;
    ++mfip;
  }

  /* record which NIC to start from for this xbar */
  FMA_XBAR(xp)->mf_nic = nip;
  FMA_XBAR(xp)->mf_nic_port = nic_port;

  /* copy over route info */
  memcpy(FMA_XBAR(xp)->mf_route, route, route_len);
  FMA_XBAR(xp)->mf_route_len = route_len;
  FMA_XBAR(xp)->mf_in_port = in_port;
  return;

 except:
  dump_map();
  fma_perror_exit(1);
}

/*
 * Fill in the MF-private fields of a NIC
 */
static void
fma_mf_init_nic_private(
  struct lf_nic *nicp,
  lf_mac_addr_t max_mac_addr)
{
  int p;
  struct fma_mf_info *mfip;

  /* array of flags for recording xbar scout responses */
  LF_CALLOC(mfip, struct fma_mf_info, nicp->num_ports);
  FMA_NIC(nicp)->mf_info = mfip;

  /* initialize mapping probe list anchors */
  for (p=0; p<nicp->num_ports; ++p) {
    mfip->mf_probe_anchor.user_next = &mfip->mf_probe_anchor;
    mfip->mf_probe_anchor.user_prev = &mfip->mf_probe_anchor;

    /* save max mac address */
    LF_MAC_COPY(mfip->mf_max_mac_addr, max_mac_addr);

    ++mfip;
  }

  /* set host index */
  FMA_NIC(nicp)->mf_host_index = fma_mf_get_host_index(max_mac_addr,
                                                       &nicp->host_nic_id);

  return;

 except:
  dump_map();
  fma_perror_exit(1);
}

/*
 * Add a NIC to the map we are creating
 */
static void
fma_mf_add_nic_to_map(
  struct lf_nic *nicp)
{
  struct fma_nic *fnp;
  struct fma_nic *anchor;

  anchor = &A.myri->map_progress->nic_anchor;
  fnp = FMA_NIC(nicp);

  /* Add this NIC to the map list */
  fnp->mf_next = anchor->mf_next;
  fnp->mf_prev = anchor;
  fnp->mf_next->mf_prev = fnp;
  fnp->mf_prev->mf_next = fnp;

  return;
}

/*
 * Add an xbar to the map we are creating
 */
static void
fma_mf_add_xbar_to_map(
  struct lf_xbar *xp)
{
  struct fma_xbar *fxp;
  struct fma_xbar *anchor;

  anchor = &A.myri->map_progress->xbar_anchor;
  fxp = FMA_XBAR(xp);

  /* Add this xbar to the map list */
  fxp->mf_next = anchor->mf_next;
  fxp->mf_prev = anchor;
  fxp->mf_next->mf_prev = fxp;
  fxp->mf_prev->mf_next = fxp;

  return;
}

/*
 * Add an xbar to the list of unresolved xbars
 */
static void
fma_mf_add_unresolved_xbar(
  struct lf_xbar *xp)
{
  struct fma_xbar *fxp;
  struct fma_xbar *anchor;

  anchor = &A.myri->map_progress->unresolved_xbar_anchor;
  fxp = FMA_XBAR(xp);

  /* Add this xbar to the unresolved list */
  fxp->mf_next = anchor->mf_next;
  fxp->mf_prev = anchor;
  fxp->mf_next->mf_prev = fxp;
  fxp->mf_prev->mf_next = fxp;
}

/*
 * Unlink a scout probe and reduce the number of pending probes
 */
static void
fma_mf_unlink_probe(
  struct fma_probe_desc *pdp)
{
  /* unlink this probe descriptor from node's scout list */
  pdp->user_next->user_prev = pdp->user_prev;
  pdp->user_prev->user_next = pdp->user_next;
  --A.myri->map_progress->mf_outstanding_probes;
}

static int
fma_mf_scout_list_empty(
  struct lf_node *np,
  int port)
{
  struct fma_probe_desc *probe_anchor;
  struct fma_mf_info *mfip;

  mfip = fma_mf_get_mf_info(np, port);
  probe_anchor = &mfip->mf_probe_anchor;
  
  return probe_anchor->user_next == probe_anchor;
}

/*
 * Get fabric mapping info for this node
 */
static struct fma_mf_info *
fma_mf_get_mf_info(
  struct lf_node *np,
  int port)
{
  if (np->type == LF_NODE_XBAR) {
    return FMA_XBAR(np)->mf_info + port;
  } if (np->type == LF_NODE_NIC) {
    return FMA_NIC(np)->mf_info + port;
  } else {
    LF_ERROR(("Bad node type"));
  }
 except:
  dump_map();
  fma_perror_exit(1);
  return NULL;
}

/*
 * We had a scout timeout - if this was an x32 scout and
 * we have gotten a regular xbar response, create
 * an "unresolved xbar" and start xbar resolution on it.
 * Otherwise, if this was the last active probe on the port, it is considered
 * disconnected.
 */
static void
fma_mf_scout_timeout(
  struct fma_probe_desc *pdp)
{
  struct lf_node *np;
  int port;

  // if (mf_debug) printf("probe %d timed out, type = %d\n", pdp->serial, pdp->type);

  /* unlink this probe descriptor from node's list */
  fma_mf_unlink_probe(pdp);

  /* get node and port */
  np = pdp->u.mf_scout.frontier_node;
  port = pdp->u.mf_scout.frontier_port;

  /* if no more pending scouts and not connected,
   * port is resolved and disconnected
   */
  if (fma_mf_scout_list_empty(np, port)) {
    struct lf_node *onp;

    onp = lf_follow_topo_link(np, port, NULL);
    if (onp == NULL) {
      fma_mf_node_port_resolved(np, port);

      if (mf_debug && np->type == LF_NODE_XBAR) printf("x%d:%d is disconnected\n", FMA_XBAR(np)->node_id, port);
    }
  }

  /* check to see if we're done */
  fma_mf_check_for_more_probes();
}

/*
 * Mark a port on this node as resolved
 */
static void
fma_mf_node_port_resolved(
  struct lf_node *np,
  int port)
{
  struct fma_mf_info *mfip;

  mfip = fma_mf_get_mf_info(np, port);

  /* mark this port as resolved */
  mfip->mf_resolved = TRUE;
}

/*
 * Look for a NIC in the map
 */
static struct lf_nic *
fma_mf_find_nic_by_mac(
  struct fma_map_progress *mpp,
  lf_mac_addr_t mac_addr)
{
  struct fma_nic *fnicp;

  /* loop through all NICs */
  for (fnicp=mpp->nic_anchor.mf_next;
       fnicp != &mpp->nic_anchor;
       fnicp=fnicp->mf_next) {
    if (LF_MAC_CMP(fnicp->parent->mac_addr, mac_addr) == 0) {
      return fnicp->parent;
    }
  }

  return NULL;
}

/*
 * Got a response from a NIC probe.
 * If this is not one of our NICs, add it to the map.
 * In any case, make a connection from this frontier node to the NIC.
 */
static void
fma_mf_got_nic_resp(
  struct fma_probe_desc *pdp,
  struct lf_myri_packet *pkt,
  int len)
{
  struct fma_myri *fmp;
  struct fma_map_progress *mpp;
  struct lf_nic *nnicp;
  struct lf_node *fnp;
  int fport;
  int nport;
  int n;

  fmp = A.myri;
  mpp = fmp->map_progress;

  if (mf_debug) {
    printf("got resp from NIC, probe %d, " LF_MAC_FORMAT ", port %d\n",
	pdp->serial, LF_MAC_ARGS(pkt->u.nic_scout_reply.mac_addr),
	pkt->u.nic_scout_reply.port_8);
  }

  /* unlink this probe descriptor from node's list */
  fma_mf_unlink_probe(pdp);

  /* port we scouted */
  nport = pkt->u.nic_scout_reply.port_8;

  /* see if this is one of our NIC ports */
  nnicp = NULL;
  for (n=0; n<fmp->num_nics; ++n) {
    if (LF_MAC_CMP(mpp->my_nics[n]->mac_addr,
	           pkt->u.nic_scout_reply.mac_addr) == 0) {
      struct fma_nic_info *nip;

      nip = fmp->nic_info[n];
      nnicp = mpp->my_nics[n];

      nip->mf_mapped[nport] = TRUE;
      break;
    }
  }

  /*
   * tell standalone that we got a scout.  This may cause mapping to
   * be cancelled, so recheck that map_progress is not NULL.  If it is,
   * just return now.
   */
  if (A.standalone) {
    fma_standalone_got_contact(pkt->u.nic_scout_reply.level_8,
	                       pkt->u.nic_scout_reply.max_mac_addr,
	                       pdp->origin, pdp->port,
	                       pdp->route, pdp->route_len,
			       FALSE);

    /* if map_progress got NULLED, all done */
    if (fmp->map_progress == NULL) return;

  }

  /* get reference to the frontier node and port */
  fnp = pdp->u.mf_scout.frontier_node;
  fport = pdp->u.mf_scout.frontier_port;

  /* if not one of ours, see if it is already known to us */
  if (nnicp == NULL) {

    /* Look for this NIC in our map */
    nnicp = fma_mf_find_nic_by_mac(mpp, pkt->u.nic_scout_reply.mac_addr);

    /* If not already in the map, add the this NIC */
    if (nnicp == NULL) {
      /* create the NIC struct */
      nnicp = fma_create_nic(pkt->u.nic_scout_reply.num_ports_8,
		      pkt->u.nic_scout_reply.mac_addr);
      if (nnicp == NULL) LF_ERROR(("Creating NIC"));

      /* Fill FMA-private structs */
      fma_mf_init_nic_private(nnicp, pkt->u.nic_scout_reply.max_mac_addr);

      fma_mf_add_nic_to_map(nnicp);	/* add it to the map */
    }
  }

  /* connect this NIC to frontier node */
  fma_mf_connect_nodes(fnp, fport, LF_NODE(nnicp), nport, TRUE);

  /*
   * Remove all scouts from the frontier node
   */
  fma_mf_clear_probes(fnp, fport, FMA_PT_ANY);

  /* check to see if we're done */
  fma_mf_check_for_more_probes();
  return;

 except:
  dump_map();
  fma_perror_exit(1);
}

/*
 * Clear probes associated with this node / port pair of a specified type.
 * A type of FMA_PT_ANY means everything.
 */
static void
fma_mf_clear_probes(
  struct lf_node *np,
  int port,
  int type)
{
  struct fma_mf_info *mfip;
  struct fma_probe_desc *anchor;
  struct fma_probe_desc *next;
  struct fma_probe_desc *pdp;

  mfip = fma_mf_get_mf_info(np, port);
  anchor = &mfip->mf_probe_anchor;

  /* start at the anchor and loop through all probes */
  pdp = anchor->user_next;
  while (pdp != anchor) {
    next = pdp->user_next;

    /* delete this record if type matches or type is wildcard */
    if (pdp->type == type || type == FMA_PT_ANY) {

      /* unlink from the node list */
      fma_mf_unlink_probe(pdp);

      /* unlink from probe list */
      if (pdp->timer != NULL) {
	lf_remove_event(pdp->timer);
        pdp->timer = NULL;
      }
      fma_unlink_probe_descriptor(pdp);

      /* free it if no sends pending, else mark it stale */
      if (pdp->sends_pending <= 0) {
        LF_FREE(pdp);
      } else {
	pdp->stale = TRUE;
      }
    }

    pdp = next;
  }
}

/*
 * got an xbar response - Remove this probe from the list as well as any NIC
 * probes. Add this xbar to the list of unresolved xbars.
 * An x32 probe may still be pending, but if it *does* come back, we'll
 * deal with it then.
 */
static void
fma_mf_got_xbar_resp(
  struct fma_probe_desc *pdp,
  struct lf_myri_packet *pkt,
  int len)
{
  struct fma_myri *fmp;
  struct fma_map_progress *mpp;
  struct lf_node *fnp;
  struct lf_xbar *nxp;
  int fport;
  int nport;

  if (mf_debug) printf("got return from probe %d, xbar response r:%s\n", pdp->serial, route_str(pdp->route, pdp->route_len));

  fmp = A.myri;
  mpp = fmp->map_progress;

  /* get reference to the frontier node and port */
  fnp = pdp->u.mf_scout.frontier_node;
  fport = pdp->u.mf_scout.frontier_port,

  /* unlink this probe descriptor from node's list */
  fma_mf_unlink_probe(pdp);

  /* clear all the NIC probes, since we know there can't be a NIC */
  fma_mf_clear_probes(fnp, fport, FMA_PT_MF_NIC_SCOUT);

  /* create unresolved xbar */
  nxp = fma_create_xbar(FMA_MF_DFLT_XBAR_PORTS, 0, 0);
  if (nxp == NULL) LF_ERROR(("error reating xbar"));

  FMA_XBAR(nxp)->mf_ref_count = 1;
  FMA_XBAR(nxp)->node_id = ++mpp->next_node_id;

  if (mf_debug) {
    printf("created xbar x%d on ", FMA_XBAR(nxp)->node_id);
    if (fnp->type == LF_NODE_NIC) {
      printf("NIC");
    } else {
      printf("x%d", FMA_XBAR(fnp)->node_id);
    }
  printf(":%d\n", fport);
  }

  /* record info about this xbar */
  nport = FMA_MF_INITIAL_XBAR_PORT;
  fma_mf_init_xbar_private(nxp, pdp->origin, pdp->port,
                           pdp->route, (pdp->route_len-1)/2, nport);
  fma_mf_connect_nodes(fnp, fport, LF_NODE(nxp), nport, FALSE);	/* unresolved */
  fma_mf_add_unresolved_xbar(nxp);

  /* start any xbar compares needed */
  fma_mf_start_next_xbar_compares();

  /* check to see if we're done */
  fma_mf_check_for_more_probes();
  return;

 except:
  dump_map();
  fma_perror_exit(1);
}

/*
 * got an x32 response - Remove all probes from the list, and search for
 * this xid.  If it is found, just make the link.  If it is not found, add a
 * new resolved xbar to the map.  If there is already something attached to
 * the frontier port, that means an xbar scout beat us back, and we need to
 * unlink the xbar and reinitialize it.
 */
static void
fma_mf_got_x32_resp(
  struct fma_probe_desc *pdp,
  struct lf_myri_packet *pkt,
  int len)
{
  struct fma_myri *fmp;
  struct fma_map_progress *mpp;
  struct lf_node *fnp;
  int fport;
  struct lf_xbar *nxp;
  int nport;
  struct lf_xbar *cxp;
  int cport;
  int xid;
  int quad_dis;

  fmp = A.myri;
  mpp = fmp->map_progress;

  /* get reference to the frontier node and port */
  fnp = pdp->u.mf_scout.frontier_node;
  fport = pdp->u.mf_scout.frontier_port,

  /* unlink from the node list */
  fma_mf_unlink_probe(pdp);

  /* clear all the probes from this node */
  fma_mf_clear_probes(fnp, fport, FMA_PT_ANY);

  /* get port and xid of new xbar */
  nport = pkt->u.xbar32_scout.x32_insert.absolute_port;
  xid = lf_swap_l(ntohl(pkt->u.xbar32_scout.x32_insert.id));
  quad_dis = pkt->u.xbar32_scout.x32_insert.quadrant_enabled;

  if (mf_debug) printf("got return from probe %d, x32 response, id=%d, p=%d, qd=%d\n", pdp->serial, xid, nport, quad_dis);

  /*
   * See if this xbar is already known to us.  If so, make the link.
   * Also check now to see if there is already something attached to this
   * frontier port.
   */
  nxp = fma_mf_find_xbar_by_xbar_id(xid);
  cxp = LF_XBAR(lf_get_topo_link(fnp, fport, &cport));

  /* if true, this xid has already been seen */
  if (nxp != NULL) {

  if (mf_debug) printf("this is x%d\n", FMA_XBAR(nxp)->node_id);
    /* something already attached to this frontier port ? */
    if (cxp != NULL) {

      /* these shouldn't happen */
      if (cxp == nxp) {
	LF_ERROR(("pre-attached x32?"));
      }
      if (cxp->xbar_id != 0) LF_ERROR(("conflicting x32 connections"));

      /* clear any probes originating from the xbar we just found */
      fma_mf_clear_probes(LF_NODE(nxp), nport, FMA_PT_ANY);

      /* If this xbar is resolved, subsume it into the x32 xbar */
      if (FMA_XBAR(cxp)->mf_xbar_resolved) {

	/* merge cxp into nxp and destroy it */
  if (mf_debug) printf("merging in x%d\n", FMA_XBAR(cxp)->node_id);
	fma_mf_merge_xbars(cxp, cport, nxp, nport);

      /* if it's not resolved, just eliminate and connect to the xbar */
      } else {
  if (mf_debug) printf("destroying x%d\n", FMA_XBAR(cxp)->node_id);
	fma_mf_destroy_xbar(cxp);
	fma_mf_connect_nodes(fnp, fport, LF_NODE(nxp), nport, TRUE);
      }

    /* there is nothing attached to our frontier port,
     * is there something attached to the x32 at this spot?  If so, 
     * make sure it is not resolved and then merge it into us
     */
    } else {
      struct lf_xbar *ncxp;

      ncxp = LF_XBAR(nxp->topo_ports[nport]);

      /* if nothing there, a new connection, link it up */
      if (ncxp == NULL) {
	fma_mf_clear_probes(LF_NODE(nxp), nport, FMA_PT_ANY);
	fma_mf_connect_nodes(fnp, fport, LF_NODE(nxp), nport, TRUE);

      /* something is there - validate and merge */
      } else {
        int ncport;

	if (FMA_XBAR(ncxp)->mf_xbar_resolved) {
	  LF_ERROR(("Unexpected resolved xbar"));
	}

	ncport = nxp->topo_rports[nport];

	/* the merge will hook us to this x32 */
	fma_mf_merge_xbars(LF_XBAR(fnp), fport, ncxp, ncport);
      }
    }

  /* no matching xid, so this is a new unique xbar */
  } else {

    /* is something already attached to this port ? */
    if (cxp != NULL) {

      /* If this xbar is resolved, convert it to an x32 */
      if (FMA_XBAR(cxp)->mf_xbar_resolved) {

	/* create a new x32 and merge this xbar into it */
	nxp = fma_create_xbar(FMA_MF_DFLT_XBAR_PORTS, xid, quad_dis);
	if (nxp == NULL) LF_ERROR(("error reating xbar"));

	FMA_XBAR(nxp)->mf_ref_count = 1;
	FMA_XBAR(nxp)->node_id = ++mpp->next_node_id;

	if (mf_debug) {
	  printf("created xbar 32 x%d, merging in x%d\n", FMA_XBAR(nxp)->node_id, FMA_XBAR(cxp)->node_id);
	}

	/* merge cxp into nxp and destroy it */
	fma_mf_merge_xbars(cxp, cport, nxp, nport);

	fma_mf_add_xbar_to_map(nxp);	/* and add this new one to the map */

	fma_mf_explore_xbar(nxp); 	/* explore this xbar */

      /* not resolved, clobber all compare probes and hook it up as an x32 */
      } else {

	/* stop any xbar compares in progress on this node */
	fma_mf_stop_compares(cxp);

	fma_mf_unlink_xbar_from_mf_list(cxp);	/* remove from unres list */

	cxp->xbar_id = xid;		/* set xid, everything else is right */

	if (mf_debug) printf("turning unresolved x%d into x32\n", FMA_XBAR(cxp)->node_id);

	/* link to this node */
	fma_mf_connect_nodes(fnp, fport, LF_NODE(cxp), nport, TRUE);

	fma_mf_add_xbar_to_map(cxp);		/* and add this to the map */

	fma_mf_explore_xbar(cxp); 		/* explore this xbar */
      }

    /* completely new x32 - allocate, fill in, link, and explore */
    } else {
      /* create xbar and add it to map */
      nxp = fma_create_xbar(FMA_MF_DFLT_XBAR_PORTS, xid, quad_dis);
      if (nxp == NULL) LF_ERROR(("error reating xbar"));

      FMA_XBAR(nxp)->mf_ref_count = 1;
      FMA_XBAR(nxp)->node_id = ++mpp->next_node_id;

      if (mf_debug) {
	printf("created xbar 32 x%d\n", FMA_XBAR(nxp)->node_id);
      }

      /* record info about this xbar */
      fma_mf_init_xbar_private(nxp, pdp->origin, pdp->port,
                               pdp->route, (pdp->route_len-1)/2, nport);

      fma_mf_add_xbar_to_map(nxp);    /* add this new xbar to the map */

      /* link to this node */
      fma_mf_connect_nodes(fnp, fport, LF_NODE(nxp), nport, TRUE);
      fma_mf_explore_xbar(nxp); 			/* explore this xbar */

      /* x32s are resolved automatically */
      FMA_XBAR(nxp)->mf_xbar_resolved = TRUE;
    }
  }

  /* start any new xbar compares needed */
  fma_mf_start_next_xbar_compares();

  /* check to see if we're done */
  fma_mf_check_for_more_probes();
  return;

 except:
  dump_map();
  fma_perror_exit(1);
}

/*
 * Stop any compares in progress on this xbar
 */
static void
fma_mf_stop_compares(
  struct lf_xbar *xp)
{
  struct fma_map_progress *mpp;
  struct fma_xbar *fxp;

  mpp = A.myri->map_progress;
  fxp = FMA_XBAR(xp);

  if (fxp->mf_now_comparing) {

    /* no longer comparing */
    fxp->mf_now_comparing = FALSE;
    --mpp->mf_outstanding_probes;

    /*
     * Clear compare timeout timer if active
     */
    if (fxp->mf_compare_timer != NULL) {
      lf_remove_event(fxp->mf_compare_timer);
      fxp->mf_compare_timer = NULL;
    }
  }

  /* IF this xbar is under compare, clear it */
  if (mpp->xbar_comparing == xp) {
    mpp->xbar_comparing = NULL;		/* comparing nothing now */
  }
}

/*
 * connect two nodes together and mark the link as resolved
 */
static void
fma_mf_connect_nodes(
  struct lf_node *np1,
  int port1,
  struct lf_node *np2,
  int port2,
  int resolved)
{
  /* link the ports */
  lf_make_topo_link(np1, port1, np2, port2);
  lf_make_topo_link(np2, port2, np1, port1);

  /* mark both ports as resolved */
  if (resolved) {
    fma_mf_node_port_resolved(np1, port1);
    fma_mf_node_port_resolved(np2, port2);
  }
}

/*
 * Explore all the ports on this xbar
 */
static void
fma_mf_explore_xbar(
  struct lf_xbar *xp)
{
  int p;
  struct fma_mf_info *mfip;

  /* explore every unresolved port */
  for (p=0; p<xp->num_ports; ++p) {

    mfip = FMA_XBAR(xp)->mf_info + p;

    if (!mfip->mf_resolved) {
      fma_mf_scout_frontier(FMA_XBAR(xp)->mf_nic, FMA_XBAR(xp)->mf_nic_port,
	                    LF_NODE(xp), p);
    }
  }
}

/*
 * Start xbar compares on this xbar
 */
static void
fma_mf_start_xbar_compares(
  struct lf_xbar *xp,
  int port)
{
  struct fma_map_progress *mpp;
  int sent;
  int rc;

  mpp = A.myri->map_progress;

  /* Send all the compares for this xbar */
  sent = fma_mf_send_xbar_compares(xp, port);

  /* If compares in progress, note this and start timer */
  if (sent > 0) {

    /* save as xbar under compare */
    mpp->xbar_comparing = xp;

    /* The whole slew of compare packets counts as one probe */
    ++mpp->mf_outstanding_probes;
    FMA_XBAR(xp)->mf_now_comparing = TRUE;

    /* set up retry count */
    FMA_XBAR(xp)->mf_compare_port = port;
    FMA_XBAR(xp)->mf_compare_retries = FMA_XBAR_SCOUT_RETRY;

    /* start timer when sends complete */
    rc = fma_myri_sync_callback(FMA_XBAR(xp)->mf_nic->nic_handle,
	                   FMA_XBAR(xp)->mf_nic_port,
			   fma_mf_set_compare_timer, xp);
    if (rc != 0) {
      LF_ERROR(("Error setting up sync callback"));
    }

    /* this is now referenced in callback */
    ++FMA_XBAR(xp)->mf_ref_count;

  } else {
    fma_mf_xbar_is_unique(xp);
  }
  return;

 except:
  fma_perror_exit(1);
}

static void
fma_mf_set_compare_timer(
  void *vxp)
{
  struct lf_xbar *xp;
  struct fma_xbar *fxp;

  xp = vxp;
  fxp = FMA_XBAR(xp);

  /* if still comparing, start the timer */
  if (fxp->mf_now_comparing) {
    fxp->mf_compare_timer =
      lf_schedule_event(fma_mf_compare_timeout, xp, FMA_XBAR_SCOUT_TIMEOUT);
    if (fxp->mf_compare_timer == NULL) {
      LF_ERROR(("Error scheduling xbar compare timeout"));
    }
  }

  /* dereference the xbar (might free it) */
  fma_mf_dereference_xbar(xp);

  return;

 except:
  fma_perror_exit(1);
}

static int
fma_mf_send_xbar_compares(
  struct lf_xbar *xp,
  int port)
{
  struct fma_map_progress *mpp;
  struct fma_xbar *fxp;
  struct lf_xbar *txp;
  int tport;
  int sent;

  mpp = A.myri->map_progress;
  sent = 0;

  if (mf_debug) printf("sending compares for x%d:%d\n", FMA_XBAR(xp)->node_id, port);

  /* loop through all resolved xbars for possible compares */
  for (fxp=mpp->xbar_anchor.mf_next;
       fxp != &mpp->xbar_anchor;
       fxp=fxp->mf_next) {
    txp = fxp->parent;

    /* start a compare on each unresolved port of each resolved xbar */
    for (tport=0; tport<txp->num_ports; ++tport) {
      struct fma_mf_info *mfip;
      mfip = fxp->mf_info + tport;

      /* if this port is not resolved, test to see if we just found it */
      if (!mfip->mf_resolved) {
	fma_mf_compare_xbar_port(xp, port, txp, tport);
	++sent;
      }
    }
  }

  /* loop through all unresolved xbars for possible compares */
  for (fxp=mpp->unresolved_xbar_anchor.mf_next;
       fxp != &mpp->unresolved_xbar_anchor;
       fxp=fxp->mf_next) {
    txp = fxp->parent;

    if (txp == xp) continue;	/* don't compare to self */

    /* don't compare to unresolved farther away than us */
    if (fxp->mf_route_len > FMA_XBAR(xp)->mf_route_len) continue;

    /* start a compare on each unresolved port of each resolved xbar */
    for (tport=0; tport<txp->num_ports; ++tport) {
      struct fma_mf_info *mfip;

      mfip = fxp->mf_info + tport;

      /* if this port is not resolved, test to see if we just found it */
      if (!mfip->mf_resolved) {
	fma_mf_compare_xbar_port(xp, port, txp, tport);
	++sent;
      }
    }
  }
  return sent;
}

/*
 * Send an xbar compare to test whether we are this port
 */
static void
fma_mf_compare_xbar_port(
  struct lf_xbar *xp,
  int port,
  struct lf_xbar *txp,
  int tport)
{
  struct lf_myri_packet pkt;
  struct fma_xbar_compare *compp;
  struct fma_map_progress *mpp;
  unsigned char route[MAX_FMA_ROUTE_LEN];
  int route_len;
  int rc;

  mpp = A.myri->map_progress;

  /* Fill in the packet */
  compp = &pkt.u.xbar_compare;
  pkt.type_16 = htons(FMA_PACKET_TYPE);
  pkt.subtype_16 = htons(FMA_SUBTYPE_XBAR_COMPARE);
  compp->map_session_id_32 = htonl(A.myri->map_session_id);
  compp->unres_node_id_32 = htonl(FMA_XBAR(xp)->node_id);
  compp->unres_port_32 = htonl(port);
  compp->test_node_id_32 = htonl(FMA_XBAR(txp)->node_id);
  compp->test_port_32 = htonl(tport);

  /*
   * create route that would be loop if xp:port == txp:tport.  We will send a
   * route that will go out to xp, and come back along the txp route inverted.
   */
  memcpy(route, FMA_XBAR(txp)->mf_route, FMA_XBAR(txp)->mf_route_len);
  route[FMA_XBAR(txp)->mf_route_len] =
       LF_DELTA_TO_ROUTE(tport - FMA_XBAR(txp)->mf_in_port);
  lf_reverse_route(route + FMA_XBAR(txp)->mf_route_len + 1,
                   FMA_XBAR(xp)->mf_route, FMA_XBAR(xp)->mf_route_len);
  route_len = FMA_XBAR(xp)->mf_route_len + FMA_XBAR(txp)->mf_route_len + 1;

#if 0
  if (mf_debug) printf("compare x%d:%d to x%d:%d, rt = %s\n", FMA_XBAR(xp)->node_id, port, FMA_XBAR(txp)->node_id, tport, route_str(route, route_len));
#endif

  /* send the probe */
  rc = fma_myri_raw_send(FMA_XBAR(xp)->mf_nic->nic_handle,
                         FMA_XBAR(xp)->mf_nic_port,
			 route, route_len,
      			 &pkt, sizeof(pkt), NULL, NULL);
  if (rc != 0) LF_ERROR(("Error sending xbar compare packet"));

  return;

 except:
  dump_map();
  fma_perror_exit(1);
}

/*
 * xbar comparisons timed out, retry if needed, else this xbar is unique.
 */
static void
fma_mf_compare_timeout(
  void *vxp)
{
  struct lf_xbar *xp;
  int unique;
  int sent;

  xp = vxp;
  unique = FALSE;
  FMA_XBAR(xp)->mf_compare_timer = NULL;

  /*
   * If retries left, re-send the packets
   */
  if (FMA_XBAR(xp)->mf_compare_retries > 0) {
    
    --FMA_XBAR(xp)->mf_compare_retries;

    /* Send all the compares for this xbar */
    sent = fma_mf_send_xbar_compares(xp, FMA_XBAR(xp)->mf_compare_port);

    /* If compares in progress, note this and start timer */
    if (sent > 0) {
      int rc;

      /* start timer when sends complete */
      rc = fma_myri_sync_callback(FMA_XBAR(xp)->mf_nic->nic_handle,
			     FMA_XBAR(xp)->mf_nic_port,
			     fma_mf_set_compare_timer, xp);
      if (rc != 0) {
	LF_ERROR(("Error setting up sync callback"));
      }

      /* this is now referenced in callback */
      ++FMA_XBAR(xp)->mf_ref_count;
      
    } else {
      unique = TRUE;
    }
  } else {
      unique = TRUE;
  }

  /*
   * This xbar is unique if either retries have been exhausted, or if we ran
   * out of things to compare to.
   */
  if (unique) {
    --A.myri->map_progress->mf_outstanding_probes;
    fma_mf_xbar_is_unique(xp);
    fma_mf_stop_compares(xp);
    fma_mf_start_next_xbar_compares();
  }
  return;

 except:
  fma_perror_exit(1);
}

/*
 * Declare an unresolved xbar unique
 */
static void
fma_mf_xbar_is_unique(
  struct lf_xbar *xp)
{
  int port;

  /* mark both ends of any links to NICs or resolved xbars as resolved */
  for (port=0; port<xp->num_ports; ++port) {
    struct lf_node *onp;
    int oport;

    onp = xp->topo_ports[port];
    if (onp == NULL) continue;

    oport = xp->topo_rports[port];

    if (onp->type == LF_NODE_NIC
        || (onp->type == LF_NODE_XBAR && FMA_XBAR(onp)->mf_xbar_resolved)) {
      fma_mf_node_port_resolved(LF_NODE(xp), port);
      fma_mf_node_port_resolved(onp, oport);

      /* clear any outstanding compares on my port */
      fma_mf_clear_probes(LF_NODE(xp), port, FMA_PT_MF_XBAR_COMPARE);
    }
  }
  
  /* mark this xbar as resolved */
  FMA_XBAR(xp)->mf_xbar_resolved = TRUE;

  if (mf_debug) printf("xbar x%d is unique\n", FMA_XBAR(xp)->node_id);

  /* move the xbar from unresolved list to map */
  fma_mf_unlink_xbar_from_mf_list(xp);
  fma_mf_add_xbar_to_map(xp);

  fma_mf_explore_xbar(xp);	/* start exploring it */
}

/*
 * Unlink and destroy this xbar
 */
static void
fma_mf_destroy_xbar(
  struct lf_xbar *xp)
{
  int p;

  /* clear any outstanding probes of any kind from any port */
  for (p=0; p<xp->num_ports; ++p) {
    fma_mf_clear_probes(LF_NODE(xp), p, FMA_PT_ANY);
  }

  /* Stop any compares from this xbar */
  fma_mf_stop_compares(xp);

  fma_mf_unlink_xbar_from_mf_list(xp);	/* unlink the xbar */
  fma_mf_dereference_xbar(xp);
}

/*
 * De-reference an xbar.  This may result in freeing it.
 */
static void
fma_mf_dereference_xbar(
  struct lf_xbar *xp)
{
  if (--FMA_XBAR(xp)->mf_ref_count <= 0) {
    fma_free_xbar(xp);
  }
}

/*
 * Find an xbar in the map by its xbar ID
 */
static struct lf_xbar *
fma_mf_find_xbar_by_xbar_id(
  int xbar_id)
{
  struct fma_xbar *anchor;
  struct fma_xbar *fxp;

  anchor = &A.myri->map_progress->xbar_anchor;

  for (fxp=anchor->mf_next; fxp != anchor; fxp=fxp->mf_next) {
    if (fxp->parent->xbar_id == xbar_id) {
      return fxp->parent;
    }
  }
  return NULL;
}

/*
 * Find an xbar in the map by its node ID
 */
static struct lf_xbar *
fma_mf_find_xbar_by_node_id(
  int node_id)
{
  struct fma_xbar *anchor;
  struct fma_xbar *fxp;

  anchor = &A.myri->map_progress->xbar_anchor;
  for (fxp=anchor->mf_next; fxp != anchor; fxp=fxp->mf_next) {
    if (fxp->node_id == node_id) {
      return fxp->parent;
    }
  }

  anchor = &A.myri->map_progress->unresolved_xbar_anchor;
  for (fxp=anchor->mf_next; fxp != anchor; fxp=fxp->mf_next) {
    if (fxp->node_id == node_id) {
      return fxp->parent;
    }
  }
  return NULL;
}

/*
 * Unlink an xbar from an mf_next based list
 */
static void
fma_mf_unlink_xbar_from_mf_list(
  struct lf_xbar *xp)
{
  FMA_XBAR(xp)->mf_prev->mf_next = FMA_XBAR(xp)->mf_next;
  FMA_XBAR(xp)->mf_next->mf_prev = FMA_XBAR(xp)->mf_prev;
}

/*
 * Got a response from an xbar compare packet.  This means we found the
 * crossbar under test to be a duplicate of an existing one.  Merge them.
 */
void
fma_mf_got_compare_resp(
  struct lf_myri_packet *pkt)
{
  struct fma_xbar_compare *compp;
  int unid;
  struct lf_xbar *uxp;
  int uport;
  int tnid;
  struct lf_xbar *txp;
  int tport;
  struct lf_node *cnp;

  compp = &pkt->u.xbar_compare;	/* get to xbar compare section */

  /* Drop this if not mapping or not for this session */
  if (A.myri->map_progress == NULL
      || ntohl(compp->map_session_id_32) != A.myri->map_session_id) {
    return;
  }

  /* pull data from the packet */
  unid = ntohl(compp->unres_node_id_32);
  uport = ntohl(compp->unres_port_32);
  tnid = ntohl(compp->test_node_id_32);
  tport = ntohl(compp->test_port_32);

  if (mf_debug) printf("unres x%d:%d matches x%d:%d\n", unid, uport, tnid, tport);

  uxp = fma_mf_find_xbar_by_node_id(unid);
  if (uxp == NULL) {
    if (mf_debug) printf("stale: unres x%d not found\n", unid);
    return;
  }

  /* 
   * If compare is not active, this set of compares got cancelled
   * or something.
   */
  if (!FMA_XBAR(uxp)->mf_now_comparing) {
    return;
  }

  txp = fma_mf_find_xbar_by_node_id(tnid);
  if (txp == NULL) {
    if (mf_debug) printf("stale: test x%d not found\n", tnid);
    return;
  }

  /* is this other port already connected to something? */
  cnp = txp->topo_ports[tport];
  if (cnp != NULL) {
    struct lf_xbar *cxp;

    LF_ERROR(("This is not happening."));

    /* make sure it's at least an xbar */
    if (cnp->type != LF_NODE_XBAR) {
      LF_ERROR(("Bad node type collides with compare"));
    }

    cxp = LF_XBAR(cnp);
    if (FMA_XBAR(cxp)->mf_xbar_resolved) {
      LF_ERROR(("Resolved xbar collides with compare"));
    }

    /* check for self so we don't destroy the wrong xbar */
    if (cxp == uxp) {
      LF_ERROR(("xbar connection already made??"));
    }

    /* now we know we have a ghost xbar at this spot - nuke it */
    fma_mf_destroy_xbar(cxp);
    txp->topo_ports[tport] = NULL;
  }

  /* Merge all port connections */
  if (FMA_XBAR(txp)->mf_xbar_resolved) {

    /* We are now all done with this xbar compare, clean up */
    lf_remove_event(FMA_XBAR(uxp)->mf_compare_timer);
    FMA_XBAR(uxp)->mf_compare_timer = NULL;
    --A.myri->map_progress->mf_outstanding_probes;

    fma_mf_merge_xbars(uxp, uport, txp, tport);

    /* remove any probes from the resolved port we tested against */
    fma_mf_clear_probes(LF_NODE(txp), tport, FMA_PT_ANY);

  } else {
    fma_mf_merge_xbars(txp, tport, uxp, uport);
  }

  /* start up the next set of compares */
  fma_mf_start_next_xbar_compares();

  return;

 except:
  dump_map();
  fma_perror_exit(1);
}

/*
 * Start xbar compares on the xbar at tail of unresolved list, if there is one
 */
static void
fma_mf_start_next_xbar_compares()
{
  struct fma_xbar *fxp;
  struct fma_xbar *anchor;
  struct fma_map_progress *mpp;

  mpp = A.myri->map_progress;

  while (mpp->xbar_comparing == NULL) {
    anchor = &mpp->unresolved_xbar_anchor;

    /* If list is not empty, start a compare on the last one */
    fxp = anchor->mf_prev;
    if (fxp != anchor) {
      fma_mf_start_xbar_compares(fxp->parent, fxp->mf_in_port);

    } else {	/* list empty - nothing to do */
      break;
    }
  }
}

/*
 * Merge all xp1 links into xp2 and destroy xp1
 */
static void
fma_mf_merge_xbars(
  struct lf_xbar *xp1,
  int port1,
  struct lf_xbar *xp2,
  int port2)
{
  int diff;
  int p1;
  int p2;

  diff = port2 - port1;

  if (mf_debug) printf("merge x%d into x%d, diff=%d\n", FMA_XBAR(xp1)->node_id, FMA_XBAR(xp2)->node_id, diff);

  for (p1=0; p1<xp1->num_ports; ++p1) {
    struct lf_node *onp;
    int oport;

    onp = xp1->topo_ports[p1];
    if (onp != NULL) {
      p2 = p1 + diff;
      if (p2 < 0 || p2 >= xp2->num_ports) {
	LF_ERROR(("Impossible port %d in xbar"));
      }
      if (xp2->topo_ports[p2] != NULL && xp2->topo_ports[p2] != onp) {
	LF_ERROR("Irreconcilable port connect conflict");	/* XXX but is it?? */
      }

      /* connect xp2 to the node */
      oport = xp1->topo_rports[p1];
      
if (mf_debug) printf("  connect x%d:%d to x%d:%d\n", FMA_XBAR(xp2)->node_id, p2, FMA_XBAR(onp)->node_id, oport);

      fma_mf_connect_nodes(LF_NODE(xp2), p2, onp, oport,
                           FMA_XBAR(xp2)->mf_xbar_resolved);
      xp1->topo_ports[p1] = NULL;		/* disconnect for safety */
    }
  }

  fma_mf_destroy_xbar(xp1);	/* destroy the old xbar */
  return;

 except:
  dump_map();
  fma_perror_exit(1);
}

static void
dump_map()
{
  struct fma_map_progress *mpp;
  struct lf_node *onp;
  struct fma_xbar *fxp;

  mpp = A.myri->map_progress;

  /* loop through all xbars in the map */
  for (fxp=mpp->xbar_anchor.mf_next;
       fxp != &mpp->xbar_anchor;
       fxp=fxp->mf_next) {
    struct lf_xbar *xp;
    int p;

    xp = fxp->parent;
    printf("x%d:\n", fxp->node_id);

    for (p=0; p<xp->num_ports; ++p) {
      struct fma_mf_info *mfip;

      printf("%d ", p);
      mfip = fxp->mf_info + p;
      onp = xp->topo_ports[p];
      if (onp == NULL) {
        if (mfip->mf_resolved) {
	  printf("disconnected\n");
	} else {
	  printf("unknown\n");
	}
      } else if (onp->type == LF_NODE_NIC) {
	struct lf_nic *nicp;

	nicp = LF_NIC(onp);
	printf("NIC " LF_MAC_FORMAT ", p%d \n", LF_MAC_ARGS(nicp->mac_addr),
	      xp->topo_rports[p]);
      } else if (onp->type == LF_NODE_XBAR) {
	struct lf_xbar *oxp;

	oxp = LF_XBAR(onp);
	printf("x%d, p%d", FMA_XBAR(oxp)->node_id, xp->topo_rports[p]);

	if (mfip->mf_resolved) {
	  printf(" (resolved)");
	}
	printf("\n");
      } else {
        printf("weird node type = %d\n", onp->type);
      }
    }
    printf("\n");
  }
}

/*
 * Mapping is complete, turn it into a topo map, call completion routine,
 * and then free everything.
 */
static void
fma_mf_mapping_complete()
{
  struct fma_map_progress *mpp;

  mpp = A.myri->map_progress;

  printf("mapping complete!\n");
  if (A.debug > 1) dump_map();

  fma_mf_make_topo_map();		/* make a topo map */

  /* call completion routine.  this must free or keep the topo map */
  mpp->map_complete_rtn(mpp->topo_map, mpp->topo_map_size);
  mpp->topo_map = NULL;

  fma_mf_free_mapping();
}

/*
 * create a topo map from the map we just generated
 */
static void
fma_mf_make_topo_map()
{
  int num_nics;
  int num_links;
  int num_xbars;
  struct lf_topo_map *topo_map;
  struct fma_map_progress *mpp;
  void *tm_buf;

  mpp = A.myri->map_progress;

  /*
   * we need to make 2 passes through the fabric - once to count everything
   * so we can allocate space, then once more to copy things into the
   * topo_map struct
   */

  /* count the NICs and start on links */
  {
    struct fma_nic *fnicp;

    num_nics = 0;
    for (fnicp=mpp->nic_anchor.mf_next;
	 fnicp != &mpp->nic_anchor;
	 fnicp=fnicp->mf_next) {
      ++num_nics;
    }
  }

  /* count all links, counting each only once.  this is a little tricky in
   * the case of an xbarport connected back to itself, make sure not to count
   * that as only half a link.
   * The rule is this: visit all xbar ports.  If connected, count the link iff:
   * 	the other node is not an xbar  OR
   * 	the other node is an xbar with an address > this xbar  OR
   * 	the other node is this same xbar, and the port # is >= current port #
   */
  {
    struct fma_xbar *fxp;

    num_xbars = 0;
    num_links = 0;

    /* loop through all xbars */
    for (fxp=mpp->xbar_anchor.mf_next;
	 fxp != &mpp->xbar_anchor;
	 fxp=fxp->mf_next) {
      struct lf_xbar *xp;
      int p;

      xp = fxp->parent;
      ++num_xbars;

      for (p=0; p<xp->num_ports; ++p) {
	struct lf_xbar *oxp;

	oxp = LF_XBAR(xp->topo_ports[p]);

	if (oxp == NULL
	    || (oxp->n.type == LF_NODE_XBAR && oxp < xp)
	    || (oxp == xp && xp->topo_rports[p] < p)) {
	  continue;
	}

	++num_links;
      }
    }
  }

  if (A.debug) printf("new map has %d nics, %d xbars, %d links\n",
      num_nics, num_xbars, num_links);

  /*
   * Everything is counted, allocate the space for the topo map
   */
  mpp->topo_map_size = LF_TOPO_MAP_SIZE(num_xbars, num_nics, num_links);
  LF_CALLOC(tm_buf, char, mpp->topo_map_size);
  mpp->topo_map = tm_buf;

  /* fill in map header */
  topo_map = mpp->topo_map;
  topo_map->num_hosts_32 = htonl(mpp->num_hosts);
  topo_map->num_xbars_32 = htonl(num_xbars);
  topo_map->num_nics_32 = htonl(num_nics);
  topo_map->num_links_32 = htonl(num_links);

  /* fill in all the xbars */
  {
    struct lf_topo_xbar *topo_xbar;
    struct fma_xbar *fxp;
    int tx;

    topo_xbar = LF_TOPO_XBAR_ARRAY(tm_buf);
    tx = 0;

    /* loop through all xbars */
    for (fxp=mpp->xbar_anchor.mf_next;
	 fxp != &mpp->xbar_anchor;
	 fxp=fxp->mf_next) {
      struct lf_xbar *xp;
      uint32_t node_id;

      xp = fxp->parent;

      node_id = LF_TOPO_XBAR_NODE_ID(tx/128, 0, tx%128);
      FMA_XBAR(xp)->map_index = tx;
      topo_xbar[tx].node_id_32 = htonl(node_id);
      topo_xbar[tx].xbar_id_32 = htonl(xp->xbar_id);
      topo_xbar[tx].num_ports_16 = htons(xp->num_ports);
      topo_xbar[tx].clos_level_8 = 0;		/* each computes this */
      topo_xbar[tx].quadrant_disable_8 = xp->quadrant_disable;

      ++tx;
    }
    if (tx != num_xbars) LF_ERROR(("xbar count mismatch!"));
  }

  /* fill in all the NICs */
  {
    struct fma_nic *fnicp;
    struct lf_topo_nic *topo_nic;
    int tn;

    topo_nic = LF_TOPO_NIC_ARRAY(tm_buf);
    tn = 0;
    for (fnicp=mpp->nic_anchor.mf_next;
	 fnicp != &mpp->nic_anchor;
	 fnicp=fnicp->mf_next) {
      struct lf_nic *nicp;
      uint32_t node_id;

      nicp = fnicp->parent;

      node_id = LF_TOPO_NIC_NODE_ID(fnicp->mf_host_index, nicp->host_nic_id);
      FMA_NIC(nicp)->map_index = tn;
      topo_nic->node_id_32 = htonl(node_id);
      memcpy(topo_nic->mac_addr, nicp->mac_addr, sizeof(lf_mac_addr_t));
      topo_nic->num_ports_16 = htons(nicp->num_ports);

      ++topo_nic;
      ++tn;
    }
    if (tn != num_nics) LF_ERROR(("NIC count mismatch!"));
  }

  /* Fill in links.  yes, could be done in xbar loop
   * but that'd be even uglier
   */
  {
    struct fma_xbar *fxp;
    struct lf_topo_link *topo_link;
    lf_topo_link_state_t *topo_link_state;
    int tl;

    topo_link = LF_TOPO_LINK_ARRAY(tm_buf);
    topo_link_state = LF_TOPO_LINK_STATE_ARRAY(tm_buf);
    tl = 0;

    /* loop through all xbars */
    for (fxp=mpp->xbar_anchor.mf_next;
	 fxp != &mpp->xbar_anchor;
	 fxp=fxp->mf_next) {
      struct lf_xbar *xp;
      int p;

      xp = fxp->parent;

      for (p=0; p<xp->num_ports; ++p) {
	struct lf_node *onp;
	uint32_t node_index;

	onp = xp->topo_ports[p];
	if (onp == NULL
	    || (onp->type == LF_NODE_XBAR && onp < LF_NODE(xp))
	    || (onp == LF_NODE(xp) && xp->topo_rports[p] < p)) {
	  continue;
	}

	/* save this xbar as node_index[0] */
	node_index = LF_TOPO_XBAR_INDEX(FMA_XBAR(xp)->map_index, p);
	topo_link->node_index_32[0] = htonl(node_index);

	/* fill in node_index[1] with link partner */
	if (onp->type == LF_NODE_XBAR) {
	  struct lf_xbar *oxp;

	  oxp = LF_XBAR(onp);
	  node_index = LF_TOPO_XBAR_INDEX(FMA_XBAR(oxp)->map_index,
	                                 xp->topo_rports[p]);
	  topo_link->node_index_32[1] = htonl(node_index);

	} else if (onp->type == LF_NODE_NIC) {
	  struct lf_nic *onicp;

	  onicp = LF_NIC(onp);
	  node_index = LF_TOPO_NIC_INDEX(FMA_NIC(onicp)->map_index,
	                                xp->topo_rports[p]);
	  topo_link->node_index_32[1] = htonl(node_index);

	} else {
	  LF_ERROR(("Bad type in topo_ports!"));
	}

	/* Save link state and topo index */
	topo_link_state[tl] = LF_LINK_STATE_UP;

	++topo_link;
	++tl;
      }
    }

    if (tl != num_links) LF_ERROR(("link count mismatch!"));
  }

  return;

 except:
  fma_perror_exit(1);
}

/*
 * Send the map we just created to the FMS
 */
void
fma_mf_send_map_to_fms(
  struct lf_topo_map *topo_map,
  int topo_map_size)
{
  struct fms_msg_header hdr;
  int sock;
  int rc;

  sock = A.fms->socket;

  if (sock == -1) LF_ERROR(("FMS socket not open"));

  /* build header */
  hdr.msg_type_32 = htonl(FMA_FMS_MAP_UPLOAD);
  hdr.length_32 = htonl(topo_map_size);

  /* send the header and the message */
  rc = lf_write(sock, &hdr, sizeof(hdr));
  if (rc != sizeof(hdr)) FMA_ERROR("error sending mapping header");
  rc = lf_write(sock, topo_map, topo_map_size);
  if (rc != topo_map_size) FMA_ERROR("error sending mapping data");

  LF_FREE(topo_map);		/* all done with map */

  return;

 except:
  fma_perror();		/* basically print error and ignore it */
}

/*
 * Halt a map in progress and free all memory associated with it
 */
void
fma_mf_cancel_mapping()
{
  /* Only work to do if a mapping actually in progress */
  if (A.myri->map_progress != NULL) {
    fma_mf_free_mapping();	/* free all the data for the mapping process */
    LF_FREE(A.myri->map_progress);
  }
}

/*
* Release all memory associated with mapping the fabric
*/
static void
fma_mf_free_mapping()
{
  struct fma_map_progress *mpp;
  struct fma_xbar *fxp;
  struct fma_nic *fnicp;
  struct fma_nic *nfnicp;

  if (mf_debug) printf("Freeing mapping\n");

  mpp = A.myri->map_progress;

  /* free all resolved xbars */
  for (fxp=mpp->xbar_anchor.mf_next;
       fxp != &mpp->xbar_anchor;
       fxp=mpp->xbar_anchor.mf_next) {
    fma_mf_destroy_xbar(fxp->parent);
  }

  for (fxp=mpp->unresolved_xbar_anchor.mf_next;
       fxp != &mpp->unresolved_xbar_anchor;
       fxp=mpp->unresolved_xbar_anchor.mf_next) {
    fma_mf_destroy_xbar(fxp->parent);
  }

  /* free all the NICs */
  for (fnicp = mpp->nic_anchor.mf_next;
       fnicp != &mpp->nic_anchor;
       fnicp = nfnicp) {
    struct lf_nic *nicp;
    int p;

    nfnicp = fnicp->mf_next;
    nicp = fnicp->parent;

    /* Clear all probes for this NIC */
    for (p=0; p<nicp->num_ports; ++p) {
      fma_mf_clear_probes(LF_NODE(nicp), p, FMA_PT_ANY);
    }
    fma_free_nic(nicp);
  }

  LF_FREE(A.myri->map_progress);
  return;
}
