/*
 * Code for sending, receiving, and responding to probes
 */
#include <stdio.h>
#include <sys/time.h>
#include <netinet/in.h>

#include "libfma.h"
#include "lf_scheduler.h"
#include "lf_fms_comm.h"
#include "lf_xbar32.h"
#include "lf_myri_packet.h"
#include "lf_fabric.h"

#include "libmyri.h"

#include "fma.h"
#include "fma_myri.h"
#include "fma_probe.h"
#include "fma_standalone.h"

/*
 * local prototypes
 */
static struct fma_probe_desc *fma_find_probe(unsigned int serial);
static struct fma_probe_desc *fma_get_matching_probe(
      struct lf_myri_packet *pkt, struct fma_nic_info *nip, int ifc);
static void fma_got_nic_scout_resp(struct fma_probe_desc *pdp,
	struct lf_myri_packet *pkt, int len);
static void fma_got_xbar32_scout_resp(struct fma_probe_desc *pdp,
	struct lf_myri_packet *pkt, int len);
static void fma_got_xbar_scout_resp(struct fma_probe_desc *pdp,
	struct lf_myri_packet *pkt, int len);
static void fma_x32_scout_failure(struct fma_probe_desc *pdp);
static void fma_xbar_scout_failure(struct fma_probe_desc *pdp);
static void fma_nic_scout_failure(struct fma_probe_desc *pdp);
static void fma_set_probe_timeout(void *vpdp);

/*
 * Handle returning probe
 */
void
fma_got_probe_resp(
  struct fma_nic_info *nip,
  int ifc,
  struct lf_myri_packet *pkt,
  int len)
{
  struct fma_probe_desc *pdp;

  /* find and validate matching probe */
  pdp = fma_get_matching_probe(pkt, nip, ifc);

  if (A.debug > 1) {
    int subtype;

    subtype = htons(pkt->subtype_16);
    if (subtype == FMA_SUBTYPE_NIC_SCOUT_REPLY) {
      printf("NIC scout reply  %d from " LF_MAC_FORMAT ", pdp=%p\n",
	  ntohl(pkt->serial_32),
	  LF_MAC_ARGS(pkt->u.nic_scout_reply.mac_addr), pdp);
    }
  }

  if (pdp == NULL) {
    return;
  }

  /* Cancel timeout timer */
  if (pdp->timer != NULL) {
    lf_remove_event(pdp->timer);
    pdp->timer = NULL;
  }

  /* call completion routine */
  pdp->resp_rtn(pdp, pkt, len);
   
  /* probe is no longer active */
  fma_unlink_probe_descriptor(pdp);

  /* If no outstanding sends, we can free the packet */
  if (pdp->sends_pending <= 0) {

    /* destroy this unless persistant */
    if (!pdp->persistant) {
	LF_FREE(pdp);
    }

  /* sends pending - marh this as "stale" */
  } else {
    pdp->stale = TRUE;	/* possibly free this when the send completes */
  }
}

/*
 * Find a matching probe record for a packet and validate the
 * packet against it.
 */
static struct fma_probe_desc *
fma_get_matching_probe(
  struct lf_myri_packet *pkt,
  struct fma_nic_info *nip,
  int ifc)
{
  struct fma_probe_desc *pdp;
  unsigned int serial;

  /* Use the serial number to find matching probe */
  serial = ntohl(pkt->serial_32);
  pdp = fma_find_probe(serial);
  if (pdp == NULL) {
    return NULL;
  }

  /* If it did not return the way we sent it, it doesn't count */
  if (LF_MAC_CMP(pkt->origin_mac_addr, nip->myri_info.mac_addr) != 0
      || pdp->port != ifc
      || pdp->origin != nip) {
    return NULL;
  }

  return pdp;
}

/*
 * Start a probe to see what's directly connected to one of our NICs.
 * We'll send both an x32 probe a few times, and then try an x16 probe a
 * few more times if that doesn't work.
 */
void
fma_start_direct_probe(
  struct fma_nic_info *nip,
  int ifc)
{
  struct fma_probe_desc *pdp;

  /* get a probe descriptor */
  pdp = fma_create_x32_scout_probe(nip, ifc, FMA_PT_X32_SCOUT,
				   fma_got_xbar32_scout_resp,
                                   fma_x32_scout_failure,
				   NULL, 0);
  /* send the packet */
  fma_send_probe_packet(pdp);

  /* packet sent OK, save the probe descriptor */
  fma_link_probe_descriptor(pdp);
}

/*
 * Create a probe for scouting an xbar32
 */
struct fma_probe_desc *
fma_create_x32_scout_probe(
  struct fma_nic_info *nip,
  int port,
  int type,
  fma_probe_resp_rtn_t resp_rtn,
  fma_probe_fail_rtn_t fail_rtn,
  unsigned char *route,
  int route_len)
{
  struct lf_myri_packet *pkt;
  struct fma_probe_desc *pdp;
  int i;

  /* get a probe descriptor */
  LF_CALLOC(pdp, struct fma_probe_desc, 1);
  pkt = &pdp->packet;

  pdp->serial = fma_myri_next_probe_serial();
  pdp->retries_left = FMA_X32_SCOUT_RETRY;
  pdp->type = type;
  pdp->origin = nip;
  pdp->port = port;
  pdp->timeout = FMA_X32_SCOUT_TIMEOUT;
  pdp->resp_rtn = resp_rtn;
  pdp->fail_rtn = fail_rtn;

  /* create loop route */
  for (i=0; i<route_len; ++i) {
    pdp->route[i] = route[i];
    pdp->route[2*route_len-i] = LF_INVERT_ROUTE(route[i]);
  }
  pdp->route[route_len] = LF_XBAR32_QUERY_ROUTE;
  pdp->route_len = route_len*2+1;

  /* Fill in an x32 scout packet */
  fma_fill_x32_scout(pkt, nip, port, pdp->serial);

  return pdp;

 except:
  fma_perror_exit(1);
  return NULL;
}

/*
 * Create a probe for scouting a generic xbar
 */
struct fma_probe_desc *
fma_create_xbar_scout_probe(
  struct fma_nic_info *nip,
  int port,
  int type,
  fma_probe_resp_rtn_t resp_rtn,
  fma_probe_fail_rtn_t fail_rtn,
  unsigned char *route,
  int route_len)
{
  struct lf_myri_packet *pkt;
  struct fma_probe_desc *pdp;

  /* get a probe descriptor */
  LF_CALLOC(pdp, struct fma_probe_desc, 1);
  pkt = &pdp->packet;

  pdp->serial = fma_myri_next_probe_serial();
  pdp->retries_left = FMA_XBAR_SCOUT_RETRY;
  pdp->type = type;
  pdp->origin = nip;
  pdp->port = port;
  pdp->timeout = FMA_XBAR_SCOUT_TIMEOUT;
  pdp->resp_rtn = resp_rtn;
  pdp->fail_rtn = fail_rtn;

  /* create loop route */
  memcpy(pdp->route, route, route_len);
  lf_reverse_route(pdp->route+route_len+1, route, route_len);
  pdp->route[route_len] = LF_DELTA_TO_ROUTE(0);
  pdp->route_len = route_len*2+1;

  /* Fill in an xbar scout packet */
  fma_fill_xbar_scout(pkt, nip, port, pdp->serial);

  return pdp;

 except:
  fma_perror_exit(1);
  return NULL;
}

/*
 * Create a probe for scouting for a NIC
 */
struct fma_probe_desc *
fma_create_nic_scout_probe(
  struct fma_nic_info *nip,
  int port,
  int type,
  fma_probe_resp_rtn_t resp_rtn,
  fma_probe_fail_rtn_t fail_rtn,
  unsigned char *route,
  int route_len,
  lf_nic_scout_reason_t reason)
{
  struct lf_myri_packet *pkt;
  struct fma_probe_desc *pdp;

  /* use the route we were passed */
  if (route_len > MAX_FMA_ROUTE_LEN) {
    LF_ERROR(("Route too long in fma_probe"));
  }

  /* get a probe descriptor */
  LF_CALLOC(pdp, struct fma_probe_desc, 1);
  pkt = &pdp->packet;

  pdp->serial = fma_myri_next_probe_serial();
  pdp->retries_left = FMA_NIC_SCOUT_RETRY;
  pdp->type = type;
  pdp->origin = nip;
  pdp->port = port;
  pdp->timeout = FMA_NIC_SCOUT_TIMEOUT;
  pdp->resp_rtn = resp_rtn;
  pdp->fail_rtn = fail_rtn;

  /* just copy the route */
  memcpy(pdp->route, route, route_len);
  pdp->route_len = route_len;

  /* Fill in a NIC scout packet */
  fma_fill_nic_scout(pkt, nip, port, pdp->serial, route, route_len, reason);

  return pdp;

 except:
  fma_perror_exit(1);
  return NULL;
}



/*
 * Fill in Myrinet packet for an x32 probe
 */
void
fma_fill_x32_scout(
  struct lf_myri_packet *pkt,
  struct fma_nic_info *nip,
  int ifc,
  int serial)
{
  pkt->type_16 = htons(FMA_PACKET_TYPE);
  pkt->subtype_16 = htons(FMA_SUBTYPE_XBAR32_SCOUT);
  memset(&pkt->u.xbar32_scout.x32_insert, 0, sizeof(struct lf_xbar32_insert));
  LF_MAC_COPY(pkt->origin_mac_addr, nip->myri_info.mac_addr);
  pkt->origin_ifc_8 = ifc;
  pkt->serial_32 = htonl(serial);
}

/*
 * Fill in Myrinet packet for an xbar probe
 */
void
fma_fill_xbar_scout(
  struct lf_myri_packet *pkt,
  struct fma_nic_info *nip,
  int ifc,
  int serial)
{
  pkt->type_16 = htons(FMA_PACKET_TYPE);
  pkt->subtype_16 = htons(FMA_SUBTYPE_XBAR_SCOUT);
  LF_MAC_COPY(pkt->origin_mac_addr, nip->myri_info.mac_addr);
  pkt->origin_ifc_8 = ifc;
  pkt->serial_32 = htonl(serial);
}

/*
 * Fill in Myrinet packet for an FMA probe
 */
void
fma_fill_nic_scout(
  struct lf_myri_packet *pkt,
  struct fma_nic_info *nip,
  int ifc,
  int serial,
  unsigned char *route,
  int route_len,
  lf_nic_scout_reason_t reason)
{
  int i;
  struct fma_nic_scout *sp;

  pkt->type_16 = htons(FMA_PACKET_TYPE);
  pkt->subtype_16 = htons(FMA_SUBTYPE_NIC_SCOUT);
  LF_MAC_COPY(pkt->origin_mac_addr, nip->myri_info.mac_addr);
  pkt->origin_ifc_8 = ifc;
  pkt->serial_32 = htonl(serial);

  /* pointer to nic scout specific part of packet */
  sp = &pkt->u.nic_scout;

  /* Fill in our mapping info */
  sp->level_8 = A.stand->my_level;
  LF_MAC_COPY(sp->max_mac_addr, A.stand->my_max_mac);

  /* copy over the reverse route */
  for (i=0; i<route_len; ++i) {
    sp->reply_route[route_len-i-1] =
	LF_DELTA_TO_ROUTE(-LF_ROUTE_TO_DELTA(route[i]));
  }
  sp->reply_route_len_8 = route_len;

  /* Tell why we are sending this scout */
  sp->reason_8 = reason;
}

/*
 * Send the packet for a probe
 */
void
fma_send_probe_packet(
  struct fma_probe_desc *pdp)
{
  int rc;

  /* send the packet using info from probe descriptor */
  rc = fma_myri_raw_send(pdp->origin->nic_handle, pdp->port,
                     pdp->route, pdp->route_len,
		     &pdp->packet, sizeof(pdp->packet),
		     fma_set_probe_timeout, pdp);
  if (rc != 0) LF_ERROR(("Error sending packet"));

  ++pdp->sends_pending;

  return;
 except:
  fma_perror_exit(1);
}

/*
 * an x32 scout sequence failed - switch to scouting for xbar 16s.
 */
static void
fma_x32_scout_failure(
  struct fma_probe_desc *pdp)
{
  struct fma_probe_desc *npdp;

  /* switch to looking for xbar with no ID */
  npdp = fma_create_xbar_scout_probe(pdp->origin, pdp->port, FMA_PT_XBAR_SCOUT,
				   fma_got_xbar_scout_resp,
                                   fma_xbar_scout_failure,
				   NULL, 0);

  fma_send_probe_packet(npdp);	/* send the packet */

  /* packet sent OK, save the probe descriptor */
  fma_link_probe_descriptor(npdp);
}

/*
 * an xbar scout timed out hard
 */
static void
fma_xbar_scout_failure(
  struct fma_probe_desc *pdp)
{
  /* XXX - need more variety here */

  /* mark connection "unknown" and give up */
  pdp->origin->xbar_id[pdp->port] = -1;
  fma_note_query_completion();
}

/*
 * a NIC scout failed hard
 */
static void
fma_nic_scout_failure(
  struct fma_probe_desc *pdp)
{
  fma_note_query_completion();
}

/*
 * Got an x32 scout response
 */
static void
fma_got_xbar32_scout_resp(
  struct fma_probe_desc *pdp,
  struct lf_myri_packet *pkt,
  int len)
{
  int xbar_id;
  int xbar_port;

  /* get xbar ID and port from packet */
  xbar_id = lf_swap_l(ntohl(pkt->u.xbar32_scout.x32_insert.id));
  xbar_port = pkt->u.xbar32_scout.x32_insert.absolute_port;

  /* Fill in the xbar info for this NIC */
  pdp->origin->xbar_id[pdp->port] = xbar_id;
  pdp->origin->xbar_port[pdp->port] = xbar_port;

  /* Notify of query completion */
  fma_note_query_completion();

  return;
}

/*
 * Got an xbar scout response
 */
static void
fma_got_xbar_scout_resp(
  struct fma_probe_desc *pdp,
  struct lf_myri_packet *pkt,
  int len)
{
  /* Fill in the xbar info for this NIC */
  pdp->origin->xbar_id[pdp->port] = 0;

  /* Notify of query completion */
  fma_note_query_completion();

  return;
}

/*
 * Got an FMA scout response
 */
static void
fma_got_nic_scout_resp(
  struct fma_probe_desc *pdp,
  struct lf_myri_packet *pkt,
  int len)
{
printf("NIC %d:%d (%d) -> NIC " LF_MAC_FORMAT ":%d\n",
       pdp->origin->nic_id, pdp->port, LF_ROUTE_TO_DELTA(pdp->route[0]),
	LF_MAC_ARGS(pkt->u.nic_scout_reply.mac_addr),
        pkt->u.nic_scout_reply.port_8);

  if (A.standalone) {
    fma_standalone_got_contact(pkt->u.nic_scout_reply.level_8,
			       pkt->u.nic_scout_reply.max_mac_addr,
			       pdp->origin, pdp->port,
			       pdp->route, pdp->route_len, FALSE);
  }

  /* Notify of query completion */
  fma_note_query_completion();

  return;
}

/*
 * Find a probe by serial number
 */
static struct fma_probe_desc *
fma_find_probe(
  unsigned int serial)
{
  struct fma_probe_desc *pdp;
  struct fma_probe_desc *anchor;

  anchor = A.myri->probe_anchor;
  for (pdp=anchor->next; pdp!=anchor; pdp=pdp->next) {
    if (pdp->serial == serial) {
      return pdp;
    }
  }
  return NULL;
}

/*
 * Got an FMA scout
 * We need to do different things with this based on what the reason is for 
 * this scout.
 * LF_NSR_NEIGHBOR_SCOUT - a startup check, just reply.
 * LF_NSR_VERIFY - verifying the fabric, just reply.
 * LF_NSR_FMS_MAP - An FMS requested map ...
 * LF_NSR_STANDALONE_MAP - Map scout from FMA in standalone mode.
 */
void
fma_got_nic_scout(
  struct fma_nic_info *nip,
  int ifc,
  struct lf_myri_packet *pkt,
  int len)
{
  struct lf_myri_packet resp;
  struct fma_nic_scout *sp;
  struct fma_nic_scout_reply *rp;
  int rc;

  sp = &pkt->u.nic_scout;

  if (0) {
    int i;
    printf("got scout from " LF_MAC_FORMAT ", port %d, reply_route = ",
	   LF_MAC_ARGS(pkt->origin_mac_addr), pkt->origin_ifc_8);
    for (i=0; i<sp->reply_route_len_8; ++i) {
      printf("%d ", LF_ROUTE_TO_DELTA(sp->reply_route[i]));
    }
    printf("\n");
  }

  /* tell standalone that we got a scout */
  if (A.standalone) {
    int is_lost;

    /* If not verifying, this poor child is lost */
    is_lost = (sp->reason_8 != LF_NSR_VERIFY);

    fma_standalone_got_contact(pkt->u.nic_scout.level_8,
	                       pkt->u.nic_scout.max_mac_addr,
			       nip, ifc,
			       sp->reply_route, sp->reply_route_len_8,
			       is_lost);
  }

  resp.type_16 = htons(FMA_PACKET_TYPE);
  resp.subtype_16 = htons(FMA_SUBTYPE_NIC_SCOUT_REPLY);
  resp.serial_32 = pkt->serial_32;
  LF_MAC_COPY(resp.origin_mac_addr, pkt->origin_mac_addr);
  resp.origin_ifc_8 = pkt->origin_ifc_8;

  /* Fill in the type specific part */
  rp = &resp.u.nic_scout_reply;
  rp->fma_protocol_32 = htonl(FMA_PROTOCOL_ID);
  LF_MAC_COPY(rp->mac_addr, nip->myri_info.mac_addr);
  rp->port_8 = ifc;
  rp->num_ports_8 = nip->myri_info.num_ports;
  rp->firmware_type_8 = 0; /* XXX */

  /* Fill in our mapping info */
  rp->level_8 = A.stand->my_level;
  LF_MAC_COPY(rp->max_mac_addr, A.stand->my_max_mac);

  rc = fma_myri_raw_send(nip->nic_handle, ifc,
                     sp->reply_route, sp->reply_route_len_8,
                     &resp, sizeof(resp), NULL, NULL);
  if (rc != 0) LF_ERROR(("Error sending scout reply"));
 
  return;

 except:
  fma_perror_exit(1);
}

/*
 * Start a probe of a neighboring NIC
 */

void
fma_start_neighbor_probe(
  struct fma_nic_info *nip,
  int ifc,
  unsigned char *route,
  int route_len)
{
  fma_start_nic_probe(nip, ifc, route, route_len, LF_NSR_NEIGHBOR_SCOUT);
}

/*
 * Start a probe to a NIC
 */
void
fma_start_nic_probe(
  struct fma_nic_info *nip,
  int ifc,
  unsigned char *route,
  int route_len,
  lf_nic_scout_reason_t reason)
{
  struct fma_probe_desc *pdp;

  /* get a probe descriptor */
  pdp = fma_create_nic_scout_probe(nip, ifc, FMA_PT_NIC_SCOUT,
                                   fma_got_nic_scout_resp,
				   fma_nic_scout_failure,
				   route, route_len, reason);

  /* send the packet */
  fma_send_probe_packet(pdp);

  /* packet sent OK, save the probe descriptor */
  fma_link_probe_descriptor(pdp);
}

/*
 * Generic timeout handler for probes - if retry count not exhausted, re-send
 * the packet, else call the failure routine for the probe
 */
void
fma_probe_timeout(
  void *vpdp)
{
  struct fma_probe_desc *pdp;

  pdp = vpdp;
#if 0
printf("probe type %d timeout on %d:%d, %d retries left\n", pdp->type,
	pdp->origin->nic_id, pdp->port, pdp->retries_left);
#endif
  pdp->timer = NULL;		/* expired */

  /* If any retries left, re-send packet with the same serial number */
  if (pdp->retries_left-- > 0) {

    fma_send_probe_packet(pdp);	/* send the packet */

  /* No retries remaining, call failure routine */
  } else {

    pdp->fail_rtn(pdp);

    fma_unlink_probe_descriptor(pdp);
    if (pdp->sends_pending <= 0) {
      if (!pdp->persistant) {
	LF_FREE(pdp);
      }
    } else {
      pdp->stale = TRUE;
printf("%p timeout with sends pending?\n", pdp);
fflush(stdout);
abort();		/* is this possible? */
    }
  }
}

/*
 * the send for a probe completed, start its timeout
 */
static void
fma_set_probe_timeout(
  void *vpdp)
{
  struct fma_probe_desc *pdp;

  pdp = vpdp;

  /* If more sends pending, do nothing.  Otherwise, free if stale and
   * start timeout timer if not stale.
   */
  if (--pdp->sends_pending <= 0) {

    if (pdp->stale) {
      if (!pdp->persistant) {
	LF_FREE(pdp);		/* free if stale and not persistant */
      }

    /* Not stale, start a timeout timer for response */
    } else {
      /* schedule a new timeout for this probe */
      pdp->timer = lf_schedule_event(fma_probe_timeout, pdp, pdp->timeout);
      if (pdp->timer == NULL) LF_ERROR(("Error rescheduling probe timeout"));
    }
  }
  return;

 except:
  fma_perror_exit(1);
}
