#ifndef _fma_probe_h_
#define _fma_probe_h_

#include "lf_myri_packet.h"

/*
 * Definitions for sending and tracking probes
 */
enum {		/* probe types */
  FMA_PT_NIC_SCOUT,
  FMA_PT_X32_VERIFY,
  FMA_PT_X32_SCOUT,
  FMA_PT_XBAR_VERIFY,
  FMA_PT_XBAR_SCOUT,
  FMA_PT_NIC_VERIFY,
  FMA_PT_EMPTY_PORT_VERIFY,
  FMA_PT_MF_X32_SCOUT,
  FMA_PT_MF_XBAR_SCOUT,
  FMA_PT_MF_NIC_SCOUT,
  FMA_PT_MF_XBAR_COMPARE,
  FMA_PT_ANY		/* matches any, a wildcard */
};

#define FMA_NIC_SCOUT_TIMEOUT 250
#define FMA_NIC_SCOUT_RETRY 3

#define FMA_X32_SCOUT_TIMEOUT 30
#define FMA_X32_SCOUT_RETRY 3

#define FMA_X32_VERIFY_TIMEOUT 100
#define FMA_X32_VERIFY_RETRY 20

#define FMA_XBAR_SCOUT_TIMEOUT 30
#define FMA_XBAR_SCOUT_RETRY 3

struct fma_probe_desc;		/* forward */
typedef void (*fma_probe_resp_rtn_t)(struct fma_probe_desc *pdp,
                                      struct lf_myri_packet *pp, int);
typedef void (*fma_probe_fail_rtn_t)(struct fma_probe_desc *);

struct fma_probe_desc {
  uint32_t serial;		/* serial number of expected packet */
  int type;			/* type of probe */

  /* retry information */
  int retries_left;		/* number of times to retry */
  int timeout;			/* timeout interval */
  struct lf_event *timer;	/* timeout event */
  fma_probe_resp_rtn_t resp_rtn;	/* call if got response */
  fma_probe_fail_rtn_t fail_rtn;	/* call if all retries fail */
  int persistant;			/* if false, free when fail or resp */

  /* NIC and port from which probe was sent */
  struct fma_nic_info *origin;
  int port;

  int sends_pending;		/* sends pending for this probe */
  int stale;			/* when a reply comes right after a retrans */

  struct lf_myri_packet packet;		/* the packet sent */
  unsigned char route[MAX_FMA_ROUTE_LEN];
  int route_len;

  union {

    /* FMA_PT_X32_VERIFY - verify an x32 is where / what it is supposed to be */
    struct fma_x32_verify {
      struct lf_xbar *xbar;		/* xbar we expect to see */
      int port;				/* port we expect to see */
    } x32_verify;

    /* FMA_PT_NIC_VERIFY - verify a NIC is where it should be */
    struct fma_nic_verify {
      struct lf_nic *nic;		/* nic we expect to see */
      int port;				/* port we expect to see */

      int topo_index;		/* link we are probing */
      int link_down;		/* If true, link is down */
    } nic_verify;

    /* FMA_PT_XBAR_VERIFY - verify an xbar is where it should be */

    /* FMA_PT_EMPTY_PORT_VERIFY - check "empty" xbar ports */
    struct fma_empty_port_verify {
      int xbar_no;
      int port;
    } empty_port_verify;

    /* FMA_PT_*_MAP_SCOUT - what frontier node are we scouting? */
    struct fma_mf_scout {
      struct lf_node *frontier_node;
      int frontier_port;
    } mf_scout;

    /* FMA_PT_XBAR_COMPARE - info about xbar compare */
    struct fma_mf_compare {
      struct lf_xbar *unres_xbar;
      int unres_port;
      struct lf_xbar *test_xbar;
      int test_port;
    } mf_compare;


  } u;

  /* doubly linked list for easy insert/remove */
  struct fma_probe_desc *user_prev;
  struct fma_probe_desc *user_next;
  struct fma_probe_desc *prev;
  struct fma_probe_desc *next;
};

/*
 * Some inlines for list manipulation
 */
static inline void
fma_link_probe_descriptor(
  struct fma_probe_desc *pdp)
{
  /* link this descriptor to end of list */
  pdp->next = A.myri->probe_anchor;
  pdp->prev = A.myri->probe_anchor->prev;

  /* insert into the list */
  pdp->prev->next = pdp;
  pdp->next->prev = pdp;
}

static inline void
fma_unlink_probe_descriptor(
  struct fma_probe_desc *pdp)
{
  /* remove from the list, checking for NULL to prevent double unlinks */
  if (pdp->next != NULL) {
    pdp->prev->next = pdp->next;
    pdp->next->prev = pdp->prev;
    pdp->next = NULL;
  }
}

/*
 * prototypes
 */
void fma_got_nic_scout(struct fma_nic_info *nip, int ifc,
                       struct lf_myri_packet *pkt, int len);
void fma_got_probe_resp(struct fma_nic_info *nip, int ifc,
                             struct lf_myri_packet *pkt, int len);
struct fma_probe_desc *fma_create_x32_scout_probe(struct fma_nic_info *nip,
  int port, int type, fma_probe_resp_rtn_t resp_rtn,
  fma_probe_fail_rtn_t fail_rtn, unsigned char *route, int route_len);
struct fma_probe_desc *fma_create_xbar_scout_probe(struct fma_nic_info *nip,
  int port, int type, fma_probe_resp_rtn_t resp_rtn,
  fma_probe_fail_rtn_t fail_rtn, unsigned char *route, int route_len);
struct fma_probe_desc *fma_create_nic_scout_probe(struct fma_nic_info *nip,
  int port, int type, fma_probe_resp_rtn_t resp_rtn,
  fma_probe_fail_rtn_t fail_rtn, unsigned char *route, int route_len,
  lf_nic_scout_reason_t reason);
void fma_start_nic_probe(struct fma_nic_info *nip, int ifc,
        unsigned char *route, int route_len, lf_nic_scout_reason_t reason);
void fma_send_probe_packet(struct fma_probe_desc *pdp);
void fma_probe_timeout(void *vpdp);
void fma_fill_xbar_scout(struct lf_myri_packet *pkt,
		        struct fma_nic_info *nip, int ifc, int serial);
void fma_fill_x32_scout(struct lf_myri_packet *pkt,
		       struct fma_nic_info *nip, int ifc, int serial);
void fma_fill_nic_scout(struct lf_myri_packet *pkt,
		               struct fma_nic_info *nip, int ifc, int serial,
			       unsigned char *route, int route_len,
			       lf_nic_scout_reason_t reason);

#endif /* _fma_probe_h_ */
