/*
 * Routines dealing with routing
 */

#include "libfma.h"
#include "lf_fabric.h"
#include "lf_fms_comm.h"
#include "lf_topo_map.h"
#include "libmyri.h"

#include "fma.h"
#include "fma_myri.h"
#include "fma_map.h"
#include "fma_fabric.h"
#include "fma_standalone.h"

/*
 * local prototypes
 */
static void fma_calc_quick_routes(struct fma_nic_info *nip, int port);
void print_fanout(struct fma_nic_info *nip, int port);

void
show_routes()
{
  struct lf_fabric *fp;

  struct fma_nic_info *lcl_nic;   /* this host fma struct */
  int lcl_nic_id;                 /* this hosts nic id */
  int lcl_port;                    /* local nic interface */
  struct lf_nic *rmt_nic;         /* remote nic struct */
  int rmt_nic_id;                 /* remote nic id */
  int rmt_port;                    /* remote nic interface */
  unsigned char *route;
  int rlen, rlenx;
  int r;
  int rindex;

  fp = A.fabric;

  for (rmt_nic_id = 0; rmt_nic_id < FMA_FABRIC(fp)->num_nics; ++rmt_nic_id) {

    rmt_nic = FMA_FABRIC(fp)->nics[rmt_nic_id];

    for (lcl_nic_id = 0; lcl_nic_id < A.myri->num_nics; ++lcl_nic_id) {

      lcl_nic = A.myri->nic_info[lcl_nic_id];

      for (lcl_port = 0; lcl_port < 2; lcl_port++) {
        for (rmt_port = 0; rmt_port < 2; rmt_port++) {
          rindex = FMA_ROUTE_INDEX(lcl_nic_id, lcl_port, rmt_port, 0);
          route = FMA_NIC(rmt_nic)->route_buf + FMA_ROUTE_OFFSET(rindex);
          rlen = FMA_NIC(rmt_nic)->route_len[rindex];

          rlenx = FMA_ROUTE_LEN(rmt_nic, lcl_nic_id, lcl_port, rmt_port, 0);

          if (rlen != 0) {
            printf("%d p%d -> " LF_MAC_FORMAT " p%d: ", lcl_nic_id, lcl_port,
                LF_MAC_ARGS(rmt_nic->mac_addr), rmt_port);
            for (r=0; r<rlen; ++r) {
              printf("%x ", route[r] & 0xff);
            }
            printf(" len: %d\n", rlenx);
          }
        }
      }
    }
  }
}

/*
 * generate a really quick and dirty set of routes
 */
void
fma_calc_all_quick_routes()
{
  int n;
  int port;
  struct lf_fabric *fp;

  fp = A.fabric;

  for (n=0; n<fp->num_xbars; ++n) {
    FMA_XBAR(fp->xbars[n])->bf_serialp = NULL;
  }

  for (n=0; n<A.myri->num_nics; ++n) {
    struct fma_nic_info *nip;

    nip = A.myri->nic_info[n];
    
    /* calculate routes for each port on this NIC */
    for (port=0; port < nip->myri_info.num_ports; ++port) {
      fma_calc_quick_routes(nip, port);
    }
  }

}

/*
 * A debugging aid to show the fanout from a NIC
 */
void
print_fanout(
  struct fma_nic_info *nip,
  int port)
{
  struct lf_xbar *xp;
  struct lf_nic *nicp;
  enum { UP, DOWN, FLAT, NUM_DIRS };
  struct lf_xbar *work_list[NUM_DIRS];
  struct lf_xbar *new_work_list[NUM_DIRS];
  void *serialp;
  struct lf_fabric *fp;

  fp = A.fabric;

  printf("\nfanout:\n");
  /* find directly connected xbar */
  nicp = FMA_FABRIC(fp)->nics[nip->map_index];
  xp = LF_XBAR(nicp->topo_ports[port]);
  if (xp == NULL) {
    printf("nothing connected\n");
    return;		/* just return if no xbar connected... */
  }

  /* set up route info for this xbar */
  FMA_XBAR(xp)->in_port = nicp->topo_rports[port];
  FMA_XBAR(xp)->route_len = 0;

  printf("direct: xbar at %p\n", xp);

  /* this is the tag we'll use for bfs serial no. */
  serialp = &serialp;

  /* initialize UP work_list to just be first xbar */
  new_work_list[UP] = xp;
  new_work_list[DOWN] = NULL;
  new_work_list[FLAT] = NULL;
  FMA_XBAR(xp)->bf_next = NULL;
  FMA_XBAR(xp)->bf_serialp = serialp;

  /* Breadth-first until nothing more to do */
  while (new_work_list[UP] != NULL
         || new_work_list[DOWN] != NULL
         || new_work_list[FLAT] != NULL) {
    int dir;

    /* Copy over work lists */
    for (dir=0; dir<NUM_DIRS; ++dir) {
      work_list[dir] = new_work_list[dir];
      new_work_list[dir] = NULL;
    }

    /* process all lists each time through */
    for (dir=0; dir<NUM_DIRS; ++dir) {

      /* save this list and start building new work_list */
      xp = work_list[dir];
      work_list[dir] = NULL;

      /* process every xbar in old work_list */
      while (xp != NULL) {
	int p;

	printf("\n");
	printf("xbar %p, level = %d\n", xp, xp->clos_level);

	/* Check all ports for NICs and unvisited xbars */
	for (p=0; p<xp->num_ports; ++p) {
	  struct lf_node *np2;
	  int port2;

	  printf("%p:%d - ", xp, p);

	  /* if this link is down, skip it */
	  if (FMA_XBAR(xp)->link_down[p]) {
	    printf("DOWN\n");
	    continue;
	  }

	  np2 = xp->topo_ports[p];
	  port2 = xp->topo_rports[p];
	  if (np2 == NULL) {
	    printf("nothing\n");
	    continue;
	  }

	  /* If this is a NIC, save the route! */
	  if (np2->type == LF_NODE_NIC) {

	    printf("NIC " LF_MAC_FORMAT "\n",
		LF_MAC_ARGS(LF_NIC(np2)->mac_addr));

	  } else if (np2->type == LF_NODE_XBAR) {
	    struct lf_xbar *xp2;
	    int new_dir;

	    xp2 = LF_XBAR(np2);
	    printf("xbar %p:%d ", xp2, port2);

	    if (xp2->clos_level > xp->clos_level) {
	      printf("(UP) ");
	      new_dir = UP;
	    } else if (xp2->clos_level < xp->clos_level) {
	      printf("(DOWN) ");
	      new_dir = DOWN;
	    } else {
	      printf("(FLAT) ");
	      new_dir = FLAT;
	    }

	    if (FMA_XBAR(np2)->bf_serialp == serialp) {
	      printf("SEEN\n");
	      continue;
	    }

	    /* don't go from flat or down to anything but down! */
	    if (dir != UP && new_dir != DOWN) {
	      printf("wrongway\n");
	      continue;
	    }
	    printf("adding\n");

	    /* add to new work_list */
	    FMA_XBAR(xp2)->bf_next = new_work_list[new_dir];
	    new_work_list[new_dir] = xp2;
	    
	    /* mark as visited to prevent multiple visitors */
	    FMA_XBAR(xp2)->bf_serialp = serialp;
	  } else {
            printf("???\n");
	  }
	}

	/* next in list */
	xp = FMA_XBAR(xp)->bf_next;
      }
    }
  }
}

/*
 * Calculate all routes for a given NIC interface
 * We will only calculate one route for each destination and will
 * only go from interface N to interface N
 *
 * start from directly connected xbar and breadth-first search
 * everywhere until we run out of places to go.  We go up then down
 * with one flat hop allowed.
 */

static void
fma_calc_quick_routes(
  struct fma_nic_info *nip,
  int port)
{
  int rlen;
  struct lf_xbar *xp;
  struct lf_nic *nicp;
  enum { UP, DOWN, FLAT, NUM_DIRS };
  struct lf_xbar *work_list[NUM_DIRS];
  struct lf_xbar *new_work_list[NUM_DIRS];
  void *serialp;
  struct lf_fabric *fp;

  fp = A.fabric;

  /* find directly connected xbar */
  nicp = FMA_FABRIC(fp)->nics[nip->map_index];
  xp = LF_XBAR(nicp->topo_ports[port]);
  if (xp == NULL) {
    return;		/* just return if no xbar connected... */
  }

  /* set up route info for this xbar */
  FMA_XBAR(xp)->in_port = nicp->topo_rports[port];
  FMA_XBAR(xp)->route_len = 0;

  /* this is the tag we'll use for bfs serial no. */
  serialp = &(nicp->topo_ports[port]);

  /* initialize UP work_list to just be first xbar */
  new_work_list[UP] = xp;
  new_work_list[DOWN] = NULL;
  new_work_list[FLAT] = NULL;
  FMA_XBAR(xp)->bf_next = NULL;
  FMA_XBAR(xp)->bf_serialp = serialp;
  rlen = 1;

  /* Breadth-first until nothing more to do */
  while (new_work_list[UP] != NULL
         || new_work_list[DOWN] != NULL
         || new_work_list[FLAT] != NULL) {
    int dir;

    /* Copy over work lists */
    for (dir=0; dir<NUM_DIRS; ++dir) {
      work_list[dir] = new_work_list[dir];
      new_work_list[dir] = NULL;
    }

    /* process all lists each time through */
    for (dir=0; dir<NUM_DIRS; ++dir) {

      /* save this list and start building new work_list */
      xp = work_list[dir];
      work_list[dir] = NULL;

      /* process every xbar in old work_list */
      while (xp != NULL) {
	int p;
	int i;

	/* Check all ports for NICs and unvisited xbars */
	p = random() % xp->num_ports;
	for (i=0; i<xp->num_ports; ++i) {
	  struct lf_node *np2;
	  int port2;

	  p = (p+1) % xp->num_ports;

	  /* if this link is down, skip it */
	  if (FMA_XBAR(xp)->link_down[p]) continue;

	  np2 = xp->topo_ports[p];
	  port2 = xp->topo_rports[p];
	  if (np2 == NULL) continue;

	  /* If this is a NIC, save the route! */
	  if (np2->type == LF_NODE_NIC) {
	    unsigned char *route;
	    int rindex;

	    rindex = FMA_ROUTE_INDEX(nip->nic_index, port, port2, 0);
	    route = FMA_NIC(np2)->route_buf + FMA_ROUTE_OFFSET(rindex);
	    memcpy(route, FMA_XBAR(xp)->route, rlen-1);
	    route[rlen-1] = LF_DELTA_TO_ROUTE(p - FMA_XBAR(xp)->in_port);
	    FMA_NIC(np2)->route_len[rindex] = rlen;

	  } else if (np2->type == LF_NODE_XBAR
		     && FMA_XBAR(np2)->bf_serialp != serialp) {
	    struct lf_xbar *xp2;
	    int new_dir;

	    xp2 = LF_XBAR(np2);
	    if (xp2->clos_level > xp->clos_level) {
	      new_dir = UP;
	    } else if (xp2->clos_level < xp->clos_level) {
	      new_dir = DOWN;
	    } else {
	      new_dir = FLAT;
	    }

	    /* don't go from flat or down to anything but down! */
	    if (dir != UP && new_dir != DOWN) {
	      continue;
	    }

	    /* record route to this xbar */
	    memcpy(FMA_XBAR(xp2)->route, FMA_XBAR(xp)->route, rlen-1);
	    FMA_XBAR(xp2)->route[rlen-1] =
		LF_DELTA_TO_ROUTE(p - FMA_XBAR(xp)->in_port);
	    FMA_XBAR(xp2)->route_len = rlen;
	    FMA_XBAR(xp2)->in_port = port2;

	    /* add to new work_list */
	    FMA_XBAR(xp2)->bf_next = new_work_list[new_dir];
	    new_work_list[new_dir] = xp2;
	    
	    /* mark as visited to prevent multiple visitors */
	    FMA_XBAR(xp2)->bf_serialp = serialp;
	  }
	}

	/* next in list */
	xp = FMA_XBAR(xp)->bf_next;
      }
    }

    /* increment distance for each go-round */
    ++rlen;
  }

  /* show_routes(); */	/* XXX */
}

/*
 * Set routes from lcl NIC interfaces to rmt NIC interfaces. The routing between
 * the local and remote interfaces is described with 2 lcl and 2 rmt ports
 * and are iterated over for the number of routes to be set.
 *
 * XXX - only works for for NICS with 2 or fewer interfaces
 */
void
fma_set_route_group(
  struct fma_nic_info *lcl_nic,
  struct lf_nic *rmt_nic,
  int lcl_port_0,
  int rmt_port_0,
  int lcl_port_1,
  int rmt_port_1)
{
  int rt_num;
  int rc;
  int lcl_port_idx = 0;
  int rmt_port_idx = 0;
  int lcl_ports[2];
  int rmt_ports[2];

  lcl_ports[0] = lcl_port_0;
  lcl_ports[1] = lcl_port_1;
  rmt_ports[0] = rmt_port_0;
  rmt_ports[1] = rmt_port_1;

  for (rt_num = 0; rt_num < lcl_nic->myri_info.num_routes; rt_num++) {
    rc = fma_set_route(lcl_nic, lcl_ports[lcl_port_idx], rmt_nic,
          rmt_ports[rmt_port_idx], rt_num);
    if (rc == -1) {
      printf("Error setting route to " LF_MAC_FORMAT "\n",
              LF_MAC_ARGS(rmt_nic->mac_addr));
    }
    lcl_port_idx = (lcl_port_idx + 1) % 2;
    rmt_port_idx = (rmt_port_idx + 1) % 2;
  }
}

/*
 * set route for all NICs
 */
int
fma_set_all_routes()
{
  struct lf_fabric *fp;
  struct fma_nic_info *lcl_nic;   /* this host fma struct */
  int lcl_nic_id;                 /* this hosts nic id */
  int lcl_port;                    /* local nic interface */
  struct lf_nic *rmt_nic;         /* remote nic struct */
  int rmt_nic_id;                 /* remote nic id */
  int rmt_port;                    /* remote nic interface */
  int rt_0_len[LF_MAX_NIC_PORTS][LF_MAX_NIC_PORTS];
                                  /* route 0 lengths. 1st desc is local port, 
                                   * 2nd is remote port */
  int rc;

  fp = A.fabric;

  for (lcl_nic_id = 0; lcl_nic_id < A.myri->num_nics; ++lcl_nic_id) {
    lcl_nic = A.myri->nic_info[lcl_nic_id];

    /* Start setting routes */
    rc = myri_set_route_start(lcl_nic->nic_handle);
    if (rc == -1) LF_ERROR(("Error from myri_set_route_start."));

    for (rmt_nic_id = 0; rmt_nic_id < FMA_FABRIC(fp)->num_nics; ++rmt_nic_id) {

      rmt_nic = FMA_FABRIC(fp)->nics[rmt_nic_id];

      /* for singular routes, route from interface 0 to interface 0 */
      if (lcl_nic->myri_info.num_routes == 1) {
        rc = fma_set_route(lcl_nic, 0, rmt_nic, 0, 0);
	if (rc == -1) LF_ERROR(("Error from fma_set_route."));
        continue;
      }

      /* get route 0 lengths for all possible route combos between
       * lcl and rmt nic interfaces. 
       */
      memset(rt_0_len, 0, sizeof(rt_0_len));
      for (lcl_port = 0; lcl_port < lcl_nic->nic_ptr->num_ports; lcl_port++) {
	for (rmt_port = 0; rmt_port < rmt_nic->num_ports; rmt_port++) {
	  rt_0_len[lcl_port][rmt_port] =
		FMA_ROUTE_LEN(rmt_nic, lcl_nic_id, lcl_port, rmt_port, 0);
	}
      }

      /* See if routes between multiple interfaces on lcl and rmt nics exist.
       * XXX - assumes LF_MAX_NIC_PORTS == 2 !!!
       */
      if ((rt_0_len[0][0] > 0 && rt_0_len[1][1] > 0) ||
          (rt_0_len[0][1] > 0 && rt_0_len[1][0] > 0)) {

        /* routes between multiple interfaces exist. Decide which ports to
         * route between based which ones exist and are the shortest.
         */
        if ((rt_0_len[0][0] > 0 && rt_0_len[1][1] > 0) &&
            (rt_0_len[0][1] > 0 && rt_0_len[1][0] > 0)) {

          /* all combos exist, pick the shortest */ 
          if (rt_0_len[0][0] + rt_0_len[1][1]
              <= rt_0_len[0][1] + rt_0_len[1][0]) {
            /* Set up to use routes from 0->0 and 1->1 because they are shorter
             * than routes from 0->1 and 1->0 */
            fma_set_route_group(lcl_nic, rmt_nic, 0, 0, 1, 1);

          } else {
            /* use routes from 0->1 and 1->0 because they are shorter */
            fma_set_route_group(lcl_nic, rmt_nic, 0, 1, 1, 0);

          }
        } else if (rt_0_len[0][0] > 0 && rt_0_len[1][1] > 0) {
          /* 0->0 and 1->1 exist, so use them */
          fma_set_route_group(lcl_nic, rmt_nic, 0, 0, 1, 1);

        } else if (rt_0_len[0][1] > 0 && rt_0_len[1][0] > 0) {
          /* 0->1 and 1->0 exist, so use them */
          fma_set_route_group(lcl_nic, rmt_nic, 0, 1, 1, 0);

        } else {
          /* this can't happen */
          LF_ERROR(("Multiple interface routing confusion"));

        }

      } else if (rt_0_len[0][0] > 0 && rt_0_len[0][1] > 0) {
        /* Use routes from lcl port 0 to rmt port 0 and 1 */
        fma_set_route_group(lcl_nic, rmt_nic, 0, 0, 0, 1);

      } else if (rt_0_len[1][0] > 0 && rt_0_len[1][1] > 0) {
        /* Use routes from lcl port 1 to rmt port 0 and 1 */
        fma_set_route_group(lcl_nic, rmt_nic, 1, 0, 1, 1);

      } else if (rt_0_len[0][0] > 0 && rt_0_len[1][0] > 0) {
        /* Use routes from lcl port 0 and 1 to rmt port 0 */
        fma_set_route_group(lcl_nic, rmt_nic, 0, 0, 1, 0);

      } else if (rt_0_len[0][1] > 0 && rt_0_len[1][1] > 0) {
        /* Use routes from lcl port 0 and 1 to rmt port 1 */
        fma_set_route_group(lcl_nic, rmt_nic, 0, 1, 1, 1);

      } else {
        /* Next try 1->1 routes.  Of all route groups, chose the one with
           the shortest routes.  (There may be more than one group of routes
           if some of the code above is disabled.) */
        int best_lcl_port=0, best_rmt_port=0;
        unsigned int shortest_len = FMA_IFC_ROUTE_LEN + 1;

        /* Find the group with the shortest routes. */
        for (lcl_port = 0; lcl_port < 2; lcl_port++) {
          for (rmt_port = 0; rmt_port < 2 ; rmt_port++) {
            if (rt_0_len[lcl_port][rmt_port] > 0 &&
                  rt_0_len[lcl_port][rmt_port] < shortest_len) {
               shortest_len = rt_0_len[lcl_port][rmt_port];
               best_lcl_port = lcl_port;
               best_rmt_port = rmt_port;
            }
          }
        }

        /* If there is a usable group of routes, use them. */
        if (shortest_len <= FMA_IFC_ROUTE_LEN) {
          fma_set_route_group(lcl_nic, rmt_nic, best_lcl_port, best_rmt_port,
                best_lcl_port, best_rmt_port);
	}
      }
    }

    /* Let the API know we are done for this NIC */
    rc = myri_set_route_end(lcl_nic->nic_handle,
			    A.stand->mapper_mac,
			    FMA_FABRIC(fp)->num_nics);
    if (rc == -1) LF_ERROR(("Error from myri_set_route_end."));

  }
  return 0;

 except:
  fma_perror();
  return -1;
}

/*
 * set a route from one NIC/port to another. fma_set_route() must be
 * bracketed by calls to myri_set_route_start() and myri_set_route_end() 
 */
int
fma_set_route(
  struct fma_nic_info *nip,
  int local_port,
  struct lf_nic *nicp2,
  int remote_port,
  int route_num)
{
  int rc;
  int rindex;
  int rlen;
  unsigned char *route;

  rindex = FMA_ROUTE_INDEX(nip->nic_index, local_port, remote_port, route_num);


  /* get the route and length */
  rlen = FMA_NIC(nicp2)->route_len[rindex];
  route= FMA_NIC(nicp2)->route_buf + FMA_ROUTE_OFFSET(rindex);

  /*
   * Save the first route to this NIC, just so we have one.
   * This may get done a couple of times per host (once per NIC) but
   * keeping the last one is easier than keeping the first one and
   * it doesn't matter which one we send to anyhow.
   */
  if (route_num == 0) {
    struct fma_host *fhp;

    fhp = FMA_HOST(nicp2->host);
    fhp->quick_route_len = rlen;
    fhp->quick_route = route;
    fhp->quick_nic = nip;
    fhp->quick_port = local_port;
  }

  if (rlen == 0) {
    return 0;
  }

  rc = myri_set_route(nip->nic_handle,
  	nicp2->mac_addr,
	remote_port,
	route_num,
	local_port,
	route, rlen);
  return rc;
}

/*
 * Figure out how far every xbar is from this host
 */
void
fma_calc_xbar_dists()
{
  int n;
  int port;
  struct lf_xbar *work_list;
  struct lf_xbar *new_list;
  struct lf_fabric *fp;

  fp = A.fabric;

  /* Initialize all xbars to have unknown distance */
  for (n=0; n<fp->num_xbars; ++n) {
    FMA_XBAR(fp->xbars[n])->host_dist = -1;
  }

  /* Start with all the distance 1 xbars */
  work_list = NULL;
  for (n=0; n<A.myri->num_nics; ++n) {
    struct fma_nic_info *nip;
    struct lf_nic *nicp;

    nip = A.myri->nic_info[n];
    nicp = nip->nic_ptr;
    
    /* start with xbar attached to each port on each NIC */
    for (port=0; port < nicp->num_ports; ++port) {
      struct lf_node *np;

      /* If this link is down, skip it */
      if (FMA_NIC(nicp)->link_down[port]) continue;

      /* get connection and make sure it's an xbar */
      np = nicp->topo_ports[port];
      if (np == NULL) continue;
      if (np->type != LF_NODE_XBAR) {
	LF_ERROR(("Non xbar attached to NIC!"));
      }

      /* If already counted, skip it */
      if (FMA_XBAR(np)->host_dist >= 0) continue;

      /* Add this xbar to the work list */
      FMA_XBAR(np)->bf_next = work_list;
      work_list = LF_XBAR(np);
      FMA_XBAR(np)->host_dist = 0;
      FMA_XBAR(np)->host_nic = nip;
      FMA_XBAR(np)->host_nic_port = port;
      FMA_XBAR(np)->host_in_port = nicp->topo_rports[port];
    }
  }

  /* Work the list BFS */
  new_list = NULL;
  while (work_list != NULL) {
    struct lf_xbar *xp;
    int p;

    /* take the top item off the work list */
    xp = work_list;
    work_list = FMA_XBAR(xp)->bf_next;

    /* Check each port for an xbar to add */
    for (p=0; p<xp->num_ports; ++p) {
      struct lf_node *np;

      /* If this link is down, skip it */
      if (FMA_XBAR(xp)->link_down[p]) continue;

      /* If this is an xbar not yet distanced, process it! */
      np = xp->topo_ports[p];
      if (np != NULL
	  && np->type == LF_NODE_XBAR
	  && FMA_XBAR(np)->host_dist < 0) {

	/* Add this xbar to the work list */
	FMA_XBAR(np)->bf_next = new_list;
	new_list = LF_XBAR(np);

	/* and set its host distance values */
	FMA_XBAR(np)->host_dist = FMA_XBAR(xp)->host_dist + 1;
	FMA_XBAR(np)->host_nic = FMA_XBAR(xp)->host_nic;
	FMA_XBAR(np)->host_nic_port = FMA_XBAR(xp)->host_nic_port;
	FMA_XBAR(np)->host_in_port = xp->topo_rports[p];
	memcpy(FMA_XBAR(np)->host_route, FMA_XBAR(xp)->host_route,
	       FMA_XBAR(xp)->host_dist);
	FMA_XBAR(np)->host_route[FMA_XBAR(xp)->host_dist] = 
	  LF_DELTA_TO_ROUTE(p - FMA_XBAR(xp)->host_in_port);
      }
    }

    /* If the list is empty, switch to new list */
    if (work_list == NULL) {
      work_list = new_list;
      new_list = NULL;
    }
  }
  return;

 except:
  fma_perror_exit(1);
}
