/*
 * Routines for running in standalone mode
 */
#include <stdio.h>
#include <unistd.h>
#include <netinet/in.h>

#include "libfma.h"
#include "lf_scheduler.h"
#include "lf_fabric.h"
#include "lf_topo_map.h"
#include "libmyri.h"

#include "fma.h"
#include "fma_fabric.h"
#include "fma_myri.h"
#include "fma_map.h"
#include "fma_fms.h"
#include "fma_standalone.h"
#include "fma_tunnel.h"
#include "fma_verify.h"

/*
 * Local prototypes
 */
static void fma_standalone_map_timeout(void *v);
static void fma_standalone_dist_route_done(void *v, void *buf);
static void fma_standalone_dist_route_failed(void *v, void *buf);
static void fma_standalone_distribute_topo_map(void);
static void fma_request_remap_from_mapper(void);
static void fma_copy_clos_levels_to_topo(void);
static int fma_is_link_xbar_xbar(int topo_index);
static void fma_standalone_assign_probes(void);
static int fma_compare_to_mapper(int level, lf_mac_addr_t mac_addr);

void
fma_init_standalone_vars()
{
  LF_CALLOC(A.stand, struct fma_standalone_data, 1);
  return;

 except:
  fma_perror_exit(1);
}

void
fma_init_standalone()
{
  struct fma_standalone_data *sdp;
  struct myri_nic_info *mip;
  struct fma_myri *mp;
  int n;

  sdp = A.stand;
  mp = A.myri;

  if (A.fms->num_hosts < 1) {
    fma_enter_standalone_mode();
  } else {
    fma_exit_standalone_mode();
  }

  /* Find my highest ordinal MAC address */
  mip = &mp->nic_info[0]->myri_info;
  LF_MAC_COPY(sdp->my_max_mac, mip->mac_addr);
  LF_MAC_COPY(sdp->mapper_mac, mip->mac_addr);

  for (n=1; n<mp->num_nics; ++n) {
    mip = &mp->nic_info[n]->myri_info;
    if (LF_MAC_CMP(mip->mac_addr, sdp->my_max_mac) > 0) {
      LF_MAC_COPY(sdp->my_max_mac, mip->mac_addr);
      LF_MAC_COPY(sdp->mapper_mac, mip->mac_addr);
    }
  }
}

/*
 * Leave standalone mode - we are now an agent of the FMS
 */
void
fma_exit_standalone_mode()
{
  A.standalone = FALSE;
  A.stand->my_level = FMA_FMS_MAPPER_LEVEL;
}

void
fma_enter_standalone_mode()
{
  struct fma_standalone_data *sdp;

  sdp = A.stand;

  /* indicate we are in standalone mode */
  A.standalone = TRUE;

  /* Active by default */
  sdp->active = TRUE;

  sdp->my_level = 1;		/* XXX */
  sdp->mapper_level = 1;
}

/*
 * Myrinet is ready - if we are the active mapper, start mapping
 */
void
fma_standalone_myrinet_ready()
{
  if (A.stand->active) {
    printf("Standalone mode - start mapping fabric\n");

    /* initiate mapping, TRUE means standalone mode */
    fma_start_mapping_fabric(fma_standalone_map_done, TRUE);
  }
}

/*
 * We got contact from another NIC out there.
 */
void
fma_standalone_got_contact(
  int level,
  lf_mac_addr_t max_mac_addr,
  struct fma_nic_info *nip,
  int port,
  uint8_t *route,
  int route_len,
  int is_lost)
{
  struct fma_standalone_data *sdp;
  int cmp;

  sdp = A.stand;

  /* If this mapper if higher than highest so far, then make this the
   * new mapper */
  cmp = fma_compare_to_mapper(level, max_mac_addr);
  if (cmp > 0) {

printf("new mapper is " LF_MAC_FORMAT ", level %d\n", LF_MAC_ARGS(max_mac_addr), level);

    /* Save this as address of new mapper */
    sdp->mapper_level = level;
    LF_MAC_COPY(sdp->mapper_mac, max_mac_addr);

    /* remember how to get to mapper */
    sdp->mapper_route_nic = nip;
    sdp->mapper_route_port = port;
    sdp->mapper_route_len = route_len;
    memcpy(sdp->mapper_route, route, route_len);

    /* If we used to be the active, cancel any mapping in progress and 
     * start waiting for a map */
    if (sdp->active) {
      sdp->active = FALSE;
      fma_mf_cancel_mapping();

    /* Cancel any map timeout we have going, and start waiting for 
     * a map from this new mapper */
    } else {

      /* remove any possible pending timeout events */
      lf_remove_event(sdp->map_wait_timer);
      sdp->map_wait_timer = NULL;
    }

    /* Start waiting for a map from this new mapper */
    sdp->map_wait_timer = lf_schedule_event(fma_standalone_map_timeout, NULL,
		                            FMA_MAP_WAIT_TIMEOUT);
    if (sdp->map_wait_timer == NULL) {
      LF_ERROR(("Error scheduling map wait timeout"));
    }

  /* not higher, but if this is a new NIC or a NIC not yet
   * in verify mode, cause a remap */
  } else if (cmp < 0 
             && (is_lost
                 || fma_find_host_by_max_mac_addr(max_mac_addr) == NULL)) {
    fma_standalone_map_is_invalid();
  }
  return;

 except:
  fma_perror_exit(1);
}

/*
 * The map wait timeout expired - become the active mapper again and start
 * mapping!
 */
static void
fma_standalone_map_timeout(
  void *v)
{
  struct fma_standalone_data *sdp;

  sdp = A.stand;

  sdp->map_wait_timer = NULL;		/* indicate the timer has expired */

printf("timeout waiting for map, going active again\n");

  /* go back to being the active mapper */
  sdp->active = TRUE;
  LF_MAC_COPY(sdp->mapper_mac, sdp->my_max_mac);
  sdp->mapper_level = sdp->my_level;


  fma_start_mapping_fabric(fma_standalone_map_done, TRUE);
}

/*
 * Our existing map is invalid, we need a new one.
 * If we are the active mapper, start mapping, else send a request
 * to the active mapper to re-map
 */
void
fma_standalone_map_is_invalid()
{
  struct fma_standalone_data *sdp;

  /* If Myrinet not even ready yet, just wait */
  if (!A.myri_ready) return;

  sdp = A.stand;

  /* if we're the mapper, map it! */
  if (sdp->active) {
    fma_start_mapping_fabric(fma_standalone_map_done, TRUE);

  /* otherwise, if we're not already waiting for a map, request one */
  } else if (sdp->map_wait_timer == NULL) {
    fma_request_remap_from_mapper();

    /* Start waiting for a map from this new mapper */
    sdp->map_wait_timer = lf_schedule_event(fma_standalone_map_timeout, NULL,
		                            FMA_MAP_WAIT_TIMEOUT);
    if (sdp->map_wait_timer == NULL) {
      LF_ERROR(("Error scheduling map wait timeout"));
    }
  }
  return;

 except:
  fma_perror_exit(1);
}

/*
 * Request a re-map from another standalone mapper
 */
static void
fma_request_remap_from_mapper()
{
  struct fma_standalone_data *sdp;

  sdp = A.stand;

  printf("Requesting remap from " LF_MAC_FORMAT "\n", LF_MAC_ARGS(sdp->mapper_mac));

  fma_tunnel_send(NULL, 0, LF_FMA_MAP_FABRIC,
	sdp->mapper_route, sdp->mapper_route_len,
	sdp->mapper_route_nic, sdp->mapper_route_port,
	NULL, NULL, NULL);
}

/*
 * we got a map, clear the timer, make probe assignments
 */
void
fma_standalone_got_map()
{
  struct fma_standalone_data *sdp;

  sdp = A.stand;

  /* Clear map receipt timeout timer */
  if (sdp->map_wait_timer != NULL) {
    lf_remove_event(sdp->map_wait_timer);
    sdp->map_wait_timer = NULL;
  }

  /* clear old probe assignments, make new ones */
  fma_standalone_assign_probes();
}

/*
 * Compare a given level/max_mac to the current mapper and return <0, 0, >0
 * a la memcmp()
 */
static int
fma_compare_to_mapper(
  int level,
  lf_mac_addr_t mac_addr)
{
  struct fma_standalone_data *sdp;
  int cmp;

  sdp = A.stand;

  cmp = level - sdp->mapper_level;
  if (cmp != 0) return cmp;

  return LF_MAC_CMP(mac_addr, sdp->mapper_mac);
}

/*
 * We have completed a map in standalone mode.  Start sending it to
 * everyone in it.
 */
void
fma_standalone_map_done(
  struct lf_topo_map *topo_map,
  int topo_map_size)
{
  struct fma_myri *fmp;
  struct fma_standalone_data *sp;
  int rc;

  printf("loading standalone map\n");

  /* load pointers */
  fmp = A.myri;
  sp = A.stand;

  /* save topo map buffer */
  sp->topo_map = topo_map;
  sp->topo_map_size = topo_map_size;

  /* load the map we generated */
  fma_load_map(topo_map);

  /* calculate the clos_level for each xbar */
  rc = lf_clos_label_xbars(A.fabric);
  if (rc == -1) LF_ERROR(("Error getting clos levels for xbars"));

  /* copy clos levels into the topo map */
  fma_copy_clos_levels_to_topo();

  /* generate and load routes from this map */
  rc = fma_process_topo_map();
  if (rc != 0) {
    FMA_ERROR(("Error processing self-generated map"));
  }

  /* Send the topo map to everyone else */
  fma_standalone_distribute_topo_map();

  /* Make probes assignments */
  fma_standalone_assign_probes();

  return;

 except:
  fma_perror_exit(1);
}

/*
 * Distribute this topo map to all other hosts in the topo map.
 * Distribute only once to each host, even where hosts have multiple NICs
 */
static void
fma_standalone_distribute_topo_map()
{
  struct lf_fabric *fp;
  struct fma_standalone_data *sp;
  int h;

  sp = A.stand;
  fp = A.fabric;			/* our fabric */

  printf("Distributing topo map to %d hosts\n", fp->num_hosts);

  /* For each host, send a map */
  for (h=0; h<fp->num_hosts; ++h) {
    struct lf_host *hp;
    struct fma_host *fhp;

    hp = fp->hosts[h];
    fhp = FMA_HOST(hp);

    /* If there is a route to this NIC, send a map */
    if (hp != A.my_host) {
      FMA_HOST(hp)->map_send_id = fma_tunnel_send(
	  sp->topo_map, sp->topo_map_size,
	  LF_FMA_TOPO_MAP,
	  fhp->quick_route, fhp->quick_route_len,
	  fhp->quick_nic, fhp->quick_port,
	  fma_standalone_dist_route_done,
	  fma_standalone_dist_route_failed,
	  (void *) ((uintptr_t)h));
      ++sp->map_sends_pending;
    }
  }
}

/*
 * Copy clos_level values from fabric to the topo map.  Assume the xbar
 * arrays are a 1:1 mapping.
 */
static void
fma_copy_clos_levels_to_topo()
{
  struct lf_fabric *fp;
  struct lf_topo_xbar *topo_xbar;
  int x;

  fp = A.fabric;
  topo_xbar = LF_TOPO_XBAR_ARRAY(A.stand->topo_map);
  for (x=0; x<fp->num_xbars; ++x) {
    topo_xbar[x].clos_level_8 = fp->xbars[x]->clos_level;
  }
}


static void
fma_standalone_dist_route_done(
  void *v,
  void *buf)
{
  struct fma_standalone_data *sp;
  int hid;

  sp = A.stand;

  hid = (int) ((uintptr_t)v);
  printf("Map distribution to %d complete, MAC=" LF_MAC_FORMAT "\n", hid, LF_MAC_ARGS(A.fabric->hosts[hid]->nics[0]->mac_addr));

  /* If the map has been successfully sent to everyone, free it */
  if (--sp->map_sends_pending <= 0) {
    LF_FREE(sp->topo_map);
    printf("map distribution complete\n");
  }
}

static void
fma_standalone_dist_route_failed(
  void *v,
  void *buf)
{
  struct fma_standalone_data *sp;
  int hid;

  hid = (int) ((uintptr_t)v);

  sp = A.stand;

  printf("Map distribution to %d failed? MAC=" LF_MAC_FORMAT "\n", hid, LF_MAC_ARGS(A.fabric->hosts[hid]->nics[0]->mac_addr));

  /* If the map has been successfully sent to everyone, free it */
  if (--sp->map_sends_pending <= 0) {
    printf("map distribution complete\n");
    LF_FREE(sp->topo_map);
  }
}

/*
 * Make probe assignments
 */
void
fma_standalone_assign_probes()
{
  int l;
  struct lf_fabric *fp;
  struct lf_host *my_hp;
  int my_id;
  int mod;
  int x;

  fp = A.fabric;

  /* first, clear all old probes */
  fma_clear_verify_probes();

  /* Now, take every Nth link and probe it, starting with M
   * where N is the number of hosts, and M is our host index
   */
  my_hp = A.my_host;
  my_id = FMA_HOST(my_hp)->host_id;
  l = my_id;
  while (l < FMA_FABRIC(fp)->num_links) {

    if (fma_is_link_xbar_xbar(l)) {

      fma_define_verify_probe(l);
    }

    /* skip to the next link for this host */
    l += fp->num_hosts;
  }

  /*
   * Start probes on every Nth empty xbar port where N is our host index
   */
  mod = 0;
  for (x=0; x<fp->num_xbars; ++x) {
    struct lf_xbar *xp;
    int p;

    xp = fp->xbars[x];
    for (p=0; p<xp->num_ports; ++p) {
      if (xp->topo_ports[p] == NULL) {
	if (mod == my_id) {
	  fma_define_empty_port_probe(xp, p);
	}
	mod = (mod+1) % fp->num_hosts;
      }
    }
  }
}

/*
 * Test whether a given topo link is xbar-xbar or nor
 */
static int
fma_is_link_xbar_xbar(
  int topo_index)
{
  struct lf_fabric *fp;
  struct lf_xbar *xp1;
  struct lf_xbar *xp2;
  int p1;

  fp = A.fabric;
  
  /* Get pointer to first end of the link */
  xp1 = LF_XBAR(FMA_FABRIC(fp)->link_end_node[topo_index]);
  p1 = FMA_FABRIC(fp)->link_end_port[topo_index];

  /* Make this end is an xbar */
  if (xp1->n.type != LF_NODE_XBAR) {
    return FALSE;
  }

  /* get the other end */
  xp2 = LF_XBAR(xp1->topo_ports[p1]);

  /* Make sure it is an xbar */
  if (xp2->n.type != LF_NODE_XBAR) {
    return FALSE;
  }

  return TRUE;
}
