/*
 * Code for sending, receiving, and responding to probes
 */
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <netinet/in.h>

#include "libfma.h"
#include "lf_scheduler.h"
#include "lf_fma_comm.h"
#include "lf_xbar32.h"
#include "lf_myri_packet.h"
#include "lf_topo_map.h"
#include "lf_fabric.h"

#include "libmyri.h"

#include "fma.h"
#include "fma_fabric.h"
#include "fma_myri.h"
#include "fma_map.h"
#include "fma_probe.h"
#include "fma_verify.h"
#include "fma_settings.h"

/*
 * Local prototypes
 */
static void fma_initiate_verify_probe(void *vpdp);
static void fma_got_nic_verify_resp(struct fma_probe_desc *pdp,
                                    struct lf_myri_packet *pkt, int len);
static void fma_nic_verify_failed(struct fma_probe_desc *pdp);
static int fma_find_verify_route(struct fma_probe_desc *pdp);
static void fma_define_one_empty_port_probe( struct lf_xbar *xp,
  int port, int x32);
static int fma_find_empty_port_route( struct fma_probe_desc *pdp,
  struct lf_xbar *xp, int port, int x32);
static void fma_empty_port_resp(struct fma_probe_desc *pdp,
  struct lf_myri_packet *pkt, int len);
static void fma_empty_port_failed(struct fma_probe_desc *pdp);
static void fma_got_xbar_verify_resp(struct fma_probe_desc *pdp,
                                    struct lf_myri_packet *pkt, int len);
static void fma_xbar_verify_failed(struct fma_probe_desc *pdp);

void dumpit(struct lf_fabric *);

/*
 * Clear all verify probes
 */
void
fma_clear_verify_probes()
{
  struct fma_probe_desc *pdp;
  struct fma_probe_desc *npdp;

#if 1
printf("clearing verify probes\n");
#endif
  npdp = A.myri->verify_anchor->user_next;
  while (npdp != A.myri->verify_anchor) {
    pdp = npdp;
    npdp = npdp->user_next;

    /* unlink this from user list */
    pdp->user_prev->user_next = pdp->user_next;
    pdp->user_next->user_prev = pdp->user_prev;

    /* cancel any pending timeout */
    if (pdp->timer != NULL) {
      lf_remove_event(pdp->timer);
      pdp->timer = NULL;
    }

    /* destroy the verify probe */
    fma_unlink_probe_descriptor(pdp);
    if (pdp->sends_pending <= 0) {
      LF_FREE(pdp);
    } else {
      pdp->stale = TRUE;
      pdp->persistant = FALSE;
    }
  }
}

/*
 * Make a probe assignment based on a message
 */
void
fma_define_verify_probe_from_msg(
  struct lf_fma_define_probe *dpp)
{
  struct fma_settings *asp;

  asp = A.settings;

  fma_define_verify_probe(ntohl(dpp->topo_index_32));
}

/*
 * Define a verify probe given a topo index
 */
void
fma_define_verify_probe(
  int topo_index)
{
  struct fma_settings *asp;

  struct fma_probe_desc *pdp;
  struct fma_probe_desc *anchor;
  int first_interval;
  int rc;

  anchor = A.myri->verify_anchor;
  asp = A.settings;

  /* allocate a probe descriptor for this verify probe */
  LF_CALLOC(pdp, struct fma_probe_desc, 1);

  /* fill in fields for the probe */
  pdp->type = FMA_PT_NIC_VERIFY;
  pdp->timeout = asp->verify_timeout;
  pdp->u.nic_verify.topo_index = topo_index;
  pdp->resp_rtn = fma_got_nic_verify_resp;
  pdp->fail_rtn = fma_nic_verify_failed;
  pdp->persistant = TRUE;

  /* find the local NIC, port, and route we will use to verify this link */
  rc = fma_find_verify_route(pdp);
  if (rc < 0) {

    /* rc of -1 means map is likely invalid */
    if (rc == -1) {
      fma_save_error(__FILE__, __LINE__, 0,
	  "Error calculating verify route, invalid map?");
      fma_perror();

      /* Free this fledgling probe descriptor and invalidate the map */
      fma_map_is_invalid();	/* try getting a new map */
    }

    /* Free descriptor and return */
    LF_FREE(pdp);
    return;
  }

  /* attach it to list of verify probes */
  pdp->user_next = anchor->user_next;
  pdp->user_prev = anchor;
  pdp->user_next->user_prev = pdp;
  pdp->user_prev->user_next = pdp;

  /* Fill in the packet for this probe, except for serial # */
  fma_fill_nic_scout(&pdp->packet, pdp->origin, pdp->port,
                     0, pdp->route, pdp->route_len, LF_NSR_VERIFY);

#if 0
{int i;
printf("define verify link %d, nic %d, port %d route=", pdp->u.nic_verify.topo_index,
  pdp->origin->nic_id, pdp->port);
for (i=0; i<pdp->route_len; ++i) printf("%d ",
  LF_ROUTE_TO_DELTA(pdp->route[i]));
printf(" reply=");
for (i=0; i<pdp->packet.u.nic_scout.reply_route_len_8; ++i) printf("%d ",
  LF_ROUTE_TO_DELTA(pdp->packet.u.nic_scout.reply_route[i]));
printf("\n");
}
#endif

  /* schedule the verify probe for the first time */
  if (asp->verify_interval > 0) {
    first_interval = random() % asp->verify_interval;
    pdp->timer = lf_schedule_event(fma_initiate_verify_probe, pdp,
	first_interval);
    if (pdp->timer == NULL) LF_ERROR(("Error rescheduling verify probe"));
  }
  return;

 except:
  fma_perror_exit(1);
}

/*
 * Start a verify probe cycle on a link
 */
static void
fma_initiate_verify_probe(
  void *vpdp)
{
  struct fma_probe_desc *pdp;
  struct fma_settings *asp;

  asp = A.settings;
  pdp = vpdp;

  pdp->timer = NULL;	/* no timer active now */

  /* 
   * If verify interval has been set to zero, then just shut this probe down for now
   */
  if (asp->verify_interval == 0) {
    return;
  }

  /* If sends are still pending, skip this time around */
  if (pdp->sends_pending > 0) {

    printf("Deferring verify, sends still pending\n");

    /* start re-probe timer */
    pdp->timer = lf_schedule_event(fma_initiate_verify_probe, pdp,
	asp->verify_interval);
    if (pdp->timer == NULL) LF_ERROR(("Error rescheduling verify probe"));

    return;
  }

  /* fill in serial number and retries */
  pdp->serial = fma_myri_next_probe_serial();
  pdp->packet.serial_32 = htonl(pdp->serial);

  /* If the link is expected to be down, just try once,
   * else specified number of times
   */
  if (pdp->type == FMA_PT_NIC_VERIFY) {
    if (pdp->u.nic_verify.link_down) {
      pdp->retries_left = 0;
    } else {
      pdp->retries_left = asp->verify_retries;
    }

  /* same for xbar verifies */
  } else if (pdp->type == FMA_PT_XBAR_VERIFY) {
    struct lf_nic *nicp;
    struct lf_xbar *xp;
    int xport;

    /* get NIC pointer */
    nicp = pdp->origin->nic_ptr;
    xp = LF_XBAR(nicp->topo_ports[pdp->port]);
    xport = nicp->topo_rports[pdp->port];

    /* If the link is expected to be down, ask for a remap to pick it up */
    if (xp == NULL || FMA_XBAR(xp)->link_down[xport]) {
      pdp->retries_left = 0;
    } else {
      pdp->retries_left = asp->verify_retries;
    }

  /* just try once when checking dead ports */
  } else if (pdp->type == FMA_PT_EMPTY_PORT_VERIFY) {
    pdp->retries_left = asp->verify_retries;
  }

  /* start the send of the packet */
  fma_send_probe_packet(pdp);			/* send the packet */

  /* put this on probe list to be found */
  fma_link_probe_descriptor(pdp);

  return;

 except:
  fma_perror_exit(1);
}

/*
 * Got a verify probe response.
 * If we were expecting the link to be down,
 * XXX request a remap
 * If we are expecting it to be up, deactivate the probe and schedule
 * a re-probe for later
 */
static void
fma_got_nic_verify_resp(
  struct fma_probe_desc *pdp,
  struct lf_myri_packet *pkt,
  int len)
{
  struct fma_settings *asp;

  asp = A.settings;

  /* If the link is expected to be down, ask for a remap to pick it up */
  if (pdp->u.nic_verify.link_down) {
{int i;
printf("resp from down link %d, nic %d, port %d route=", pdp->u.nic_verify.topo_index,
  pdp->origin->nic_id, pdp->port);
for (i=0; i<pdp->route_len; ++i) printf("%d ",
  LF_ROUTE_TO_DELTA(pdp->route[i]));
printf(" reply=");
for (i=0; i<pdp->packet.u.nic_scout.reply_route_len_8; ++i) printf("%d ",
  LF_ROUTE_TO_DELTA(pdp->packet.u.nic_scout.reply_route[i]));
printf("\n");
}
    fma_map_is_invalid();
    pdp->u.nic_verify.link_down = FALSE;
  }

  /* start re-probe timer */
  if (asp->verify_interval > 0) {
    pdp->timer = lf_schedule_event(fma_initiate_verify_probe, pdp,
	asp->verify_interval);
    if (pdp->timer == NULL) LF_ERROR(("Error rescheduling verify probe"));
  }

  return;

 except:
  fma_perror_exit(1);
}

/*
 * Failed to get a response from a verify probe
 * Beginner: just request a re-map of the fabric
 * Advanced: try to isolate the actual non-working link in the chain
 */
static void
fma_nic_verify_failed(
  struct fma_probe_desc *pdp)
{
  struct fma_settings *asp;

  asp = A.settings;

  /* If we expected the link to be down, just re-schedule the probe */
  if (pdp->u.nic_verify.link_down) {

    /* if interval still > 0, start re-probe timer */
    if (asp->verify_interval > 0) {
      pdp->timer = lf_schedule_event(fma_initiate_verify_probe, pdp,
	  asp->verify_interval);
      if (pdp->timer == NULL) LF_ERROR(("Error rescheduling verify probe"));
    }

  /*
   * If we expect the link to be up, report that this map is invalid
   * and destroy the probe.
   */
  } else {

{int i;
printf("verify failed link %d, nic %d, port %d route=", pdp->u.nic_verify.topo_index,
  pdp->origin->nic_id, pdp->port);
for (i=0; i<pdp->route_len; ++i) printf("%d ",
  LF_ROUTE_TO_DELTA(pdp->route[i]));
printf(" reply=");
for (i=0; i<pdp->packet.u.nic_scout.reply_route_len_8; ++i) printf("%d ",
  LF_ROUTE_TO_DELTA(pdp->packet.u.nic_scout.reply_route[i]));
printf("\n");
}

    /* link is already removed from the probe list, go ahead and clobber the
     * whole verify record
     */
    pdp->user_prev->user_next = pdp->user_next;
    pdp->user_next->user_prev = pdp->user_prev;
pdp->user_next = NULL;
if (pdp->timer != NULL) printf("%p: timeout timer still active??\n", pdp);

    /* allow probe subsystem to free this */
    pdp->persistant = FALSE;

    /* request re-map */ 
    fma_map_is_invalid();
  }
  return;

 except:
  fma_perror_exit(1);
}

/*
 * Find a NIC on the far side of this xbar link so we can test it.  We
 * ensure that this is a link between two xbars, and find which one it closer
 * to one of the NICs on this host.  We then go from the other end until we
 * find a NIC which is not the origination NIC and that is the target
 * NIC for this verify probe.
 *
 * If all OK, return 0;
 * If a truly fatal error occurs, print error and exit.
 * If remap needed, return -1.
 * If we just can't do this probe and advise skipping it, return -2.
 *
 */
static int
fma_find_verify_route(
  struct fma_probe_desc *pdp)
{
  struct lf_fabric *fp;
  struct lf_xbar *xp1;
  struct lf_xbar *xp2;
  struct lf_xbar *work_list;
  struct lf_xbar *new_work_list;
  unsigned int serial;
  int old_link_down;
  int p1;
  int p2;
  int fatal;

  fp = A.fabric;
  fatal = 0;
  
  /* Get pointer to first end of the link */
  xp1 = LF_XBAR(FMA_FABRIC(fp)->link_end_node[pdp->u.nic_verify.topo_index]);
  p1 = FMA_FABRIC(fp)->link_end_port[pdp->u.nic_verify.topo_index];

  /* Make this end is an xbar */
  if (xp1->n.type != LF_NODE_XBAR) {
    LF_ERROR(("Link endpoint 1 is not an xbar!"));
  }

  /* get the other end */
  xp2 = LF_XBAR(xp1->topo_ports[p1]);
  p2 = xp1->topo_rports[p1];

  /* Make sure there is a link, and that it is an xbar */
  if (xp2 == NULL) {
    LF_ERROR(("Requested to test missing link!"));
  }
  if (xp2->n.type != LF_NODE_XBAR) {
    LF_ERROR(("Link endpoint 2 is not an xbar!"));
  }

  /* If either xbar is unreachable, we have no way to test this link,
   * and the map must be awry somehow.  Return an error.
   */
  if (FMA_XBAR(xp1)->host_dist == -1
      || FMA_XBAR(xp2)->host_dist == -1) {
    fatal = -1;			/* request remap */
    LF_ERROR(("Being asked to verify unreachable xbar"));
  }

  /* Record the state of the link - if it is down now, we are looking
   * for it to go up!
   */
  pdp->u.nic_verify.link_down = FMA_XBAR(xp1)->link_down[p1];

  /* We have both xbars, make xp1 be the one closest to this host */
  if (FMA_XBAR(xp2)->host_dist < FMA_XBAR(xp1)->host_dist) {
    struct lf_xbar *txp;
    int tp;

    txp = xp1;	/* swap 'em */
    xp1 = xp2;
    xp2 = txp;
    tp = p1;
    p1 = p2;
    p2 = tp;
  }

  /* Save the starting NIC information */
  pdp->origin = FMA_XBAR(xp1)->host_nic;
  pdp->port = FMA_XBAR(xp1)->host_nic_port;

  /* Find the closest NIC to xp2 that we can without using p2.
   * This is a BFS outwards from xp2 until we hit a NIC.
   */
  old_link_down = FMA_XBAR(xp2)->link_down[p2];
  FMA_XBAR(xp2)->link_down[p2] = TRUE;

  /* initial route to this xbar */
  memcpy(FMA_XBAR(xp2)->route, FMA_XBAR(xp1)->host_route,
	 FMA_XBAR(xp1)->host_dist);
  FMA_XBAR(xp2)->route[FMA_XBAR(xp1)->host_dist] =
    LF_DELTA_TO_ROUTE(p1 - FMA_XBAR(xp1)->host_in_port);
  FMA_XBAR(xp2)->route_len = FMA_XBAR(xp1)->host_dist + 1;
  FMA_XBAR(xp2)->in_port = xp1->topo_rports[p1];

  /* setup lists */
  work_list = xp2;
  FMA_XBAR(xp2)->bf_next = NULL;
  new_work_list =  NULL;

  pdp->u.nic_verify.nic = NULL;		/* no NIC yet */

  serial = ++A.myri->bfs_serial;	/* use serial numbers to test visited */
  FMA_XBAR(xp2)->bf_serial = serial;

  while (work_list != NULL) {
    struct lf_xbar *xp;
    int low_port;
    int i;
    int p;

    /* pull the top xbar off the list */
    xp = work_list;
    work_list = FMA_XBAR(xp)->bf_next;

    /* Don't violate quadrant disable */
    if (xp->quadrant_disable && FMA_XBAR(xp)->in_port < xp->num_ports/2) {
      low_port = xp->num_ports/2;
    } else {
      low_port = 0;
    }

    /* check each port */
    p = low_port + (random() % (xp->num_ports - low_port));
    for (i=low_port; i<xp->num_ports; ++i) {
      struct lf_node *np;

      /* p hits all values in [low_port,xp->num_ports) */
      if (++p >= xp->num_ports) p = low_port;

      /* see where this link goes */
      np = xp->topo_ports[p];

      /* skip over down or disconnected links */
      if (np == NULL || FMA_XBAR(xp)->link_down[p]) continue;

      /* If this is a NIC, use this as target and exit loop */
      if (np->type == LF_NODE_NIC) {

	/* Fill in route for probe */
	memcpy(pdp->route, FMA_XBAR(xp)->route, FMA_XBAR(xp)->route_len);
	pdp->route[FMA_XBAR(xp)->route_len] =
	  LF_DELTA_TO_ROUTE(p - FMA_XBAR(xp)->in_port);
	pdp->route_len = FMA_XBAR(xp)->route_len + 1;

	/* record the NIC we should find */
	pdp->u.nic_verify.nic = LF_NIC(np);
	pdp->u.nic_verify.port = xp->topo_rports[p];

	goto search_done;

      /* If it is an xbar and not visited, add it to new work list */
      } else if (np->type == LF_NODE_XBAR) {

	/* If visited, skip it */
	if (FMA_XBAR(np)->bf_serial == serial) continue;

	/* Not yet visited, build route to this xbar and add it to
	 * the new work list
	 */
	memcpy(FMA_XBAR(np)->route, FMA_XBAR(xp)->route,
	       FMA_XBAR(xp)->route_len);
	FMA_XBAR(np)->route[FMA_XBAR(xp)->route_len] =
	  LF_DELTA_TO_ROUTE(p - FMA_XBAR(xp)->in_port);
if (p == FMA_XBAR(xp)->in_port) {
printf("0 route delta!?!?\n");
}
	FMA_XBAR(np)->route_len = FMA_XBAR(xp)->route_len + 1;
	FMA_XBAR(np)->in_port = xp->topo_rports[p];

	FMA_XBAR(np)->bf_serial = serial;	/* mark as visited */

	/* Add to the next work list */
	FMA_XBAR(np)->bf_next = new_work_list;
	new_work_list = LF_XBAR(np);

      /* any other node type is an error */
      } else {
	LF_ERROR(("Bad node type"));
      }
    }

    /* If done with this level of BFS, advance to the next */
    if (work_list == NULL) {
      work_list = new_work_list;
      new_work_list = NULL;
    }
  }
 search_done:

  /* restore the old link state to the one we avoided */
  FMA_XBAR(xp2)->link_down[p2] = old_link_down;

  /* If no NIC found, we can't do this probe... */
  if (pdp->u.nic_verify.nic == NULL) {
    fatal = -2;
    goto except;
  }

  return 0;

 except:
  if (fatal == 0) {
    fma_perror_exit(1);
  }
  return fatal;
}

/*
 * Define a probe which looks for an empty port on an xbar to stop being 
 * empty.  It sends both x32 and xbar probes, since quadrant disable may
 * prevent x32s from responding to simple roundtrips.
 */
void
fma_define_empty_port_probe(
  struct lf_xbar *xp,
  int port)
{
  /* define one ID-less probe and one ID-full */
  if (A.xbar_types != FMA_XT_ID_ONLY) {
    fma_define_one_empty_port_probe(xp, port, FALSE);
  }
  if (A.xbar_types != FMA_XT_NO_IDS) {
    fma_define_one_empty_port_probe(xp, port, TRUE);
  }
}

static void
fma_define_one_empty_port_probe(
  struct lf_xbar *xp,
  int port,
  int x32)
{
  struct fma_settings *asp;
  struct fma_probe_desc *pdp;
  struct fma_probe_desc *anchor;
  int first_interval;
  int rc;

  anchor = A.myri->verify_anchor;
  asp = A.settings;

  /* allocate a probe descriptor for this verify probe */
  LF_CALLOC(pdp, struct fma_probe_desc, 1);

  /* fill in fields for the probe */
  pdp->type = FMA_PT_EMPTY_PORT_VERIFY;
  pdp->timeout = asp->verify_timeout;
  pdp->u.empty_port_verify.xbar_no = FMA_XBAR(xp)->map_index;
  pdp->u.empty_port_verify.port = port;
  pdp->resp_rtn = fma_empty_port_resp;
  pdp->fail_rtn = fma_empty_port_failed;
  pdp->persistant = TRUE;

  /* find the local NIC, port, and route we will use to probe this port */
  rc = fma_find_empty_port_route(pdp, xp, port, x32);
  if (rc < 0) {
    LF_FREE(pdp);
    LF_ERROR(("Cannot build route to empty port"));
  }

  /* attach it to list of verify probes */
  pdp->user_next = anchor->user_next;
  pdp->user_prev = anchor;
  pdp->user_next->user_prev = pdp;
  pdp->user_prev->user_next = pdp;

  /* Fill in the packet for this probe, except for serial # */
  if (x32) {
    fma_fill_x32_scout(&pdp->packet, pdp->origin, pdp->port, 0);
  } else {
    fma_fill_xbar_scout(&pdp->packet, pdp->origin, pdp->port, 0);
  }

#if 0
{int i;
printf("probe empty port, xbar %d, port %d route=",
  FMA_XBAR(xp)->map_index, port);
for (i=0; i<pdp->route_len; ++i) printf("%d ",
  LF_ROUTE_TO_DELTA(pdp->route[i]));
printf("\n");
}
#endif

  /* schedule the verify probe for the first time */
  if (asp->verify_interval > 0) {
    first_interval = random() % asp->verify_interval;
    pdp->timer = lf_schedule_event(fma_initiate_verify_probe, pdp,
	first_interval);
    if (pdp->timer == NULL) LF_ERROR(("Error scheduling empty port probe"));
  }
  return;

 except:
  fma_perror_exit(1);
}

/*
 * An empty port probe came up with nothing, just reschedule it
 */
static void
fma_empty_port_failed(
  struct fma_probe_desc *pdp)
{
  struct fma_settings *asp;

  asp = A.settings;
  
  /* if non-zero interval, start re-probe timer, else just hang out on the list */
  if (asp->verify_interval > 0) {
    pdp->timer = lf_schedule_event(fma_initiate_verify_probe, pdp,
	asp->verify_interval);
    if (pdp->timer == NULL) LF_ERROR(("Error rescheduling verify probe"));
  }
  return;

 except:
  fma_perror_exit(1);
}

/*
 * We got a response from a supposedly empty port - this map is invalid!
 */
static void
fma_empty_port_resp(
  struct fma_probe_desc *pdp,
  struct lf_myri_packet *pkt,
  int len)
{
  printf("empty port %d on xbar %d is no longer empty\n",
	pdp->u.empty_port_verify.port, pdp->u.empty_port_verify.xbar_no);

  /* just call the map invalid, this probe will get cleaned up later */
  fma_map_is_invalid();
}

/*
 * Find a route to test a supposedly disconnected port on an xbar
 */
static int
fma_find_empty_port_route(
  struct fma_probe_desc *pdp,
  struct lf_xbar *xp,
  int port,
  int x32)
{
  int dist;

  /* Save the starting NIC information */
  pdp->origin = FMA_XBAR(xp)->host_nic;
  pdp->port = FMA_XBAR(xp)->host_nic_port;

  /* initial route to this xbar */
  dist = FMA_XBAR(xp)->host_dist;
  memcpy(pdp->route, FMA_XBAR(xp)->host_route, dist);
  pdp->route[dist] = LF_DELTA_TO_ROUTE(port - FMA_XBAR(xp)->host_in_port);
  pdp->route[dist+1] = x32 ? LF_XBAR32_QUERY_ROUTE : LF_DELTA_TO_ROUTE(0);
  lf_reverse_route(pdp->route + dist+2, pdp->route, dist+1);

  pdp->route_len = (dist+1) * 2 + 1;
  
  return 0;
}

/*
 * Define a probe for a directly connected xbar
 */
void
fma_define_direct_xbar_verify_probe(
  struct fma_nic_info *nip,
  int port)
{
  struct fma_probe_desc *pdp;
  struct fma_settings *asp;
  int first_interval;

  asp = A.settings;

  /* if probe already defined, just return */
  if (nip->xbar_verify_probe != NULL) return;

  printf("creating direct verify for NIC %d p%d\n", nip->nic_id, port);

  /* allocate a probe descriptor for this verify probe */
  LF_CALLOC(pdp, struct fma_probe_desc, 1);

  /* fill in fields for the probe */
  pdp->type = FMA_PT_XBAR_VERIFY;
  pdp->timeout = asp->verify_timeout;
  pdp->resp_rtn = fma_got_xbar_verify_resp;
  pdp->fail_rtn = fma_xbar_verify_failed;
  pdp->persistant = TRUE;

  /* Fill in NIC, port, and route */
  pdp->origin = nip;
  pdp->port = port;
  pdp->route[0] = LF_DELTA_TO_ROUTE(0);
  pdp->route_len = 1;

  /* record this probe */
  nip->xbar_verify_probe = pdp;

  /* Fill in the packet for this probe, except for serial # */
  fma_fill_xbar_scout(&pdp->packet, pdp->origin, pdp->port, 0);

  /* schedule the verify probe for the first time */
  if (asp->verify_interval > 0) {
    first_interval = random() % asp->verify_interval;
    pdp->timer = lf_schedule_event(fma_initiate_verify_probe, pdp,
	first_interval);
    if (pdp->timer == NULL) LF_ERROR(("Error scheduling xbar verify probe"));
  }
  return;

 except:
  fma_perror_exit(1);
}

/*
 * Failed to get a response from a verify probe
 * Beginner: just request a re-map of the fabric
 * Advanced: try to isolate the actual non-working link in the chain
 */
static void
fma_xbar_verify_failed(
  struct fma_probe_desc *pdp)
{
  struct lf_nic *nicp;
  struct lf_xbar *xp;
  struct fma_settings *asp;
  int xport;

  asp = A.settings;

  /* get NIC pointer */
  nicp = pdp->origin->nic_ptr;
  xp = LF_XBAR(nicp->topo_ports[pdp->port]);
  xport = nicp->topo_rports[pdp->port];

  /*
   * If we expect the link to be up, report that this map is invalid
   */
  if (!FMA_XBAR(xp)->link_down[xport]) {

    /* request re-map */ 
    printf("Direct xbar verify failed on NIC %d p%d\n", pdp->origin->nic_id,
	pdp->port);
    fma_map_is_invalid();
  }

  /* if interval still non-zero, start re-probe timer */
  if (asp->verify_interval > 0) {
    pdp->timer = lf_schedule_event(fma_initiate_verify_probe, pdp,
	asp->verify_interval);
    if (pdp->timer == NULL) LF_ERROR(("Error rescheduling verify probe"));
  }

  return;

 except:
  fma_perror_exit(1);
}

/*
 * Got an xbar verify probe response.
 * If we were expecting the link to be down,
 * XXX request a remap
 * If we are expecting it to be up, deactivate the probe and schedule
 * a re-probe for later
 */
static void
fma_got_xbar_verify_resp(
  struct fma_probe_desc *pdp,
  struct lf_myri_packet *pkt,
  int len)
{
  struct lf_nic *nicp;
  struct lf_xbar *xp;
  struct fma_settings *asp;
  int xport;

  asp = A.settings;

  /* get NIC pointer */
  nicp = pdp->origin->nic_ptr;
  xp = LF_XBAR(nicp->topo_ports[pdp->port]);
  xport = nicp->topo_rports[pdp->port];

  /* If the link is expected to be down, ask for a remap to pick it up */
  if (xp == NULL || FMA_XBAR(xp)->link_down[xport]) {
    printf("Direct xbar verify response on NIC %d p%d\n", pdp->origin->nic_id,
	pdp->port);
    fma_map_is_invalid();
  }

  /* start re-probe timer */
  if (asp->verify_interval > 0) {
    pdp->timer = lf_schedule_event(fma_initiate_verify_probe, pdp,
	asp->verify_interval);
    if (pdp->timer == NULL) LF_ERROR(("Error rescheduling verify probe"));
  }

  return;

 except:
  fma_perror_exit(1);
}

void
dumpit(
  struct lf_fabric *fp)
{
  int i,n;
  int x;

  printf("Fabric loaded, %d hosts\n", fp->num_hosts);
  for (i=0; i<fp->num_hosts; ++i) {
   lf_host_t *hp;
   hp = fp->hosts[i];
   if (hp == NULL) { printf("%d: null\n", i); continue; }
   printf("\"%s\", %d NICs\n", hp->hostname, hp->num_nics);
   for (n=0; n<hp->num_nics;++n) {
     lf_nic_t *np;
     np = hp->nics[n];
     printf("\t%d " LF_MAC_FORMAT " sn=%s, pn=%s\n", n,
       LF_MAC_ARGS(np->mac_addr),
       np->serial_no, np->product_id);
   }
  }

  printf("%d xbars\n", fp->num_xbars);
  for (x=0; x<fp->num_xbars; ++x) {
    struct lf_xbar *xp;
    int p;

    xp = LF_XBAR(fp->xbars[x]);
    printf("      xbar[%d] - xid=%d, %d ports, clos=%d, dist=%d\n", x,
	xp->xbar_id, xp->num_ports,
	xp->clos_level,
	FMA_XBAR(xp)->host_dist);
    for (p=0; p<xp->num_ports; ++p) {
      struct lf_node *onp;

      printf("       %d -", p);

      onp = xp->topo_ports[p];
      if (onp == NULL) {
	printf("\n");
	continue;
      }

      if (onp->type == LF_NODE_XBAR) {
	struct lf_xbar *oxp;
	oxp = LF_XBAR(onp);
	printf(" x%d:%d, clos=%d", FMA_XBAR(oxp)->map_index,
		xp->topo_rports[p], oxp->clos_level);

      } else if (onp->type == LF_NODE_NIC) {
	struct lf_nic *onicp;
	onicp = LF_NIC(onp);
	printf(" " LF_MAC_FORMAT ":s%d:p%d",
		LF_MAC_ARGS(onicp->mac_addr), onicp->host_nic_id,
	    xp->topo_rports[p]);
      } else {
	printf(" ????");
      }

      printf(")\n");
    }
  }
}

/*
 * This is called when verify_interval is set from 0 to something non-zero.
 * It walks the verify * prove list looking for inactive probes and
 * restarts them.
 */
void
fma_new_verify_interval()
{
  struct fma_probe_desc *pdp;
  struct fma_settings *asp;
  int first_interval;

  asp = A.settings;

  /* If the interval is being reset, nothing to do right now */
  if (asp->verify_interval <= 0) {
    return;
  }

  pdp = A.myri->verify_anchor->user_next;
  while (pdp != A.myri->verify_anchor) {

    /* only restart disabled probes */
    if (pdp->timer == NULL) {
      first_interval = random() % asp->verify_interval;
      pdp->timer = lf_schedule_event(fma_initiate_verify_probe, pdp,
	  first_interval);
      if (pdp->timer == NULL) LF_ERROR(("Error rescheduling verify probe"));
    }

    pdp = pdp->user_next;
  }
  return;

 except:
  fma_perror_exit(1);
}
