/*
 * Routines for dealing with fabric maps on the FMAs
 */
#include <sys/types.h>
#include <netinet/in.h>
#include <sys/time.h>

#include "libfma.h"
#include "lf_fabric.h"
#include "lf_fma_comm.h"
#include "lf_topo_map.h"
#include "lf_channel.h"
#include "lf_scheduler.h"

#include "fms.h"
#include "fms_error.h"
#include "fms_fabric.h"
#include "fms_notify.h"
#include "fms_fma.h"
#include "fms_fma_map.h"
#include "fms_state.h"

/*
 * Create the topological representation of the fabric in a form suitable
 * for sending
 */
void
fms_create_topo_map(
  struct lf_fabric *fp)
{
  int num_nics;
  int num_links;
  int num_links_down;
  int topo_map_size;
  void *tm_buf;
  struct lf_topo_map *topo_map;

  /* free any previously existing topo map */
  LF_FREE(FMS_FABRIC(fp)->topo_map);

  /* If no hosts or xbars, then no need to proceed with a map */
  if (fp->num_hosts == 0 || fp->num_xbars == 0) {
    return;
  }

  /*
   * we need to make 2 passes through the fabric - once to count everything
   * so we can allocate space, then once more to copy things into the
   * topo_map struct
   */

  /* count the NICs and start on links */
  {
    int h;

    num_nics = 0;
    for (h=0; h<fp->num_hosts; ++h) {
      num_nics += fp->hosts[h]->num_nics;
    }
  }


  /* count all links, counting each only once.  this is a little tricky in
   * the case of an xbarport connected back to itself, make sure not to count
   * that as only half a link.
   * The rule is this: visit all xbar ports.  If connected, count the link iff:
   * 	the other node is not an xbar  OR
   * 	the other node is an xbar with an address > this xbar  OR
   * 	the other node is this same xbar, and the port # is >= current port #
   */
  {
    int x;

    num_links = 0;
    num_links_down = 0;
    for (x=0; x<fp->num_xbars; ++x) {
      struct lf_xbar *xp;
      int p;

      xp = fp->xbars[x];

      for (p=0; p<xp->num_ports; ++p) {
	struct lf_xbar *oxp;

	oxp = LF_XBAR(xp->topo_ports[p]);
	if (oxp == NULL
	    || (oxp->n.type == LF_NODE_XBAR && oxp < xp)
	    || (oxp == xp && xp->topo_rports[p] < p)) {
	  continue;
	}

	++num_links;
      }
    }
  }

  /*
   * Everything is counted, allocate the space for the topo map
   */
  topo_map_size = LF_TOPO_MAP_SIZE(fp->num_xbars, num_nics, num_links);
  FMS_FABRIC(fp)->topo_map_size = topo_map_size;
  LF_CALLOC(tm_buf, char, topo_map_size);
  FMS_FABRIC(fp)->topo_map = tm_buf;
  fms_notify(FMS_EVENT_DEBUG, "topo_map_size = %d", topo_map_size);

  /* fill in map header */
  topo_map = tm_buf;
  topo_map->num_hosts_32 = htonl(fp->num_hosts);
  topo_map->num_xbars_32 = htonl(fp->num_xbars);
  topo_map->num_nics_32 = htonl(num_nics);
  topo_map->num_links_32 = htonl(num_links);

  /* fill in all the xbars */
  {
    int e;
    int tx;
    struct lf_topo_xbar *topo_xbar;

    topo_xbar = LF_TOPO_XBAR_ARRAY(tm_buf);
    tx = 0;

    /* loop using enclosures */
    for (e=0; e<fp->num_enclosures; ++e) {
      struct lf_enclosure *ep;
      int l;

      ep = fp->enclosures[e];
      for (l=0; l<ep->num_slots; ++l) {
	struct lf_linecard *lp;
	int x;

	lp = ep->slots[l];
	if (lp == NULL) continue;

	for (x=0; x<lp->num_xbars; ++x) {
	  struct lf_xbar *xp;
	  uint32_t node_id;

	  xp = LF_XBAR(lp->xbars[x]);

	  node_id = LF_TOPO_XBAR_NODE_ID(FMS_ENC(ep)->eid, l, x);
	  FMS_XBAR(xp)->node_index = tx;
	  topo_xbar[tx].node_id_32 = htonl(node_id);
	  topo_xbar[tx].xbar_id_32 = htonl(xp->xbar_id);
	  topo_xbar[tx].num_ports_16 = htons(xp->num_ports);
	  topo_xbar[tx].clos_level_8 = xp->clos_level;
	  topo_xbar[tx].quadrant_disable_8 = xp->quadrant_disable;

	  ++tx;
	}
      }
    }
    if (tx != fp->num_xbars) LF_ERROR(("xbar count mismatch!"));
  }

  /* fill in all the NICs */
  {
    struct lf_topo_nic *topo_nic;
    int tn;
    int h;

    topo_nic = LF_TOPO_NIC_ARRAY(tm_buf);
    tn = 0;
    for (h=0; h<fp->num_hosts; ++h) {
      int n;
      struct lf_host *hp;

      hp = fp->hosts[h];
      for (n=0; n<hp->num_nics; ++n) {
	struct lf_nic *nicp;
	uint32_t node_id;

	nicp = hp->nics[n];

	node_id = LF_TOPO_NIC_NODE_ID(h, nicp->host_nic_id);
	FMS_NIC(nicp)->node_index = tn;
	topo_nic->node_id_32 = htonl(node_id);
	memcpy(topo_nic->mac_addr, nicp->mac_addr, sizeof(lf_mac_addr_t));
	topo_nic->num_ports_16 = htons(nicp->num_ports);

	++topo_nic;
	++tn;
      }
    }
    if (tn != num_nics) LF_ERROR(("NIC count mismatch!"));
  }

  /* Fill in links.  yes, could be done in xbar loop
   * but that'd be even uglier
   */
  {
    struct lf_topo_link *topo_link;
    lf_topo_link_state_t *topo_link_state;
    int x;
    int tl;

    topo_link = LF_TOPO_LINK_ARRAY(tm_buf);
    topo_link_state = LF_TOPO_LINK_STATE_ARRAY(tm_buf);
    tl = 0;

    for (x=0; x<fp->num_xbars; ++x) {
      struct lf_xbar *xp;
      int p;

      xp = fp->xbars[x];

      for (p=0; p<xp->num_ports; ++p) {
	struct lf_node *onp;
	uint32_t node_index;

	onp = xp->topo_ports[p];
	if (onp == NULL
	    || (onp->type == LF_NODE_XBAR && onp < LF_NODE(xp))
	    || (onp == LF_NODE(xp) && xp->topo_rports[p] < p)) {
	  continue;
	}

	/* save this xbar as node_index[0] */
	node_index = LF_TOPO_XBAR_INDEX(FMS_XBAR(xp)->node_index, p);
	topo_link->node_index_32[0] = htonl(node_index);

	/* fill in node_index[1] with link partner */
	if (onp->type == LF_NODE_XBAR) {
	  struct lf_xbar *oxp;

	  oxp = LF_XBAR(onp);
	  node_index = LF_TOPO_XBAR_INDEX(FMS_XBAR(oxp)->node_index,
	                                 xp->topo_rports[p]);
	  topo_link->node_index_32[1] = htonl(node_index);

	} else if (onp->type == LF_NODE_NIC) {
	  struct lf_nic *onicp;

	  onicp = LF_NIC(onp);
	  node_index = LF_TOPO_NIC_INDEX(FMS_NIC(onicp)->node_index,
	                                xp->topo_rports[p]);
	  topo_link->node_index_32[1] = htonl(node_index);

	} else {
	  LF_ERROR(("Bad type in topo_ports!"));
	}

	/* Save link state and topo index */
	topo_link_state[tl] = FMS_XBAR(xp)->ports[p]->link->link_state;
	FMS_XBAR(xp)->ports[p]->link->topo_index = tl;

	++topo_link;
	++tl;
      }
    }

    if (tl != num_links) LF_ERROR(("link count mismatch!"));
  }

  return;

 except:
  fms_perror_exit(1);
}

/*
 * update link state in the topo map
 */
void
fms_update_topo_link_state(
  struct fms_link *linkp)
{
  struct lf_fabric *fp;
  struct lf_topo_map *topo_map;
  lf_topo_link_state_t *lsp;

  fp = F.fabvars->fabric;

  topo_map = FMS_FABRIC(fp)->topo_map;
  if (topo_map != NULL) {
    lsp = LF_TOPO_LINK_STATE_ARRAY(topo_map);
    lsp[linkp->topo_index] = linkp->link_state;
  }
}

/*
 * Send the latest map to all registered FMAs that need it
 */
int
fms_send_topo_map_to_all_needy_fma(
  struct lf_fabric *fp)
{
  int h;

  /* There should be a map */
  if (FMS_FABRIC(fp)->topo_map == NULL) {
    LF_ERROR(("No map at push time!"));
  }

  fms_notify(FMS_EVENT_INFO, "Sending map to needy FMAs");

  for (h=0; h<fp->num_hosts; ++h) {
    struct fms_fma_desc *adp;

    adp = FMS_HOST(fp->hosts[h])->adp;
    if (adp != NULL) {
      int rc;

      if (adp->needs_map) {
	fms_notify(FMS_EVENT_DEBUG, "Sending topo map to %s", 
	    fp->hosts[h]->hostname);
	rc = fms_send_topo_map(F.fabvars->fabric, adp);
	if (rc != 0) FMS_ERROR(("Error sending topo map to %s", 
	    fp->hosts[h]->hostname));

      } else if (adp->needs_link_state) {
	rc = fms_send_topo_link_state(F.fabvars->fabric, adp);
	if (rc != 0) FMS_ERROR(("Error sending link state to %s", 
	    fp->hosts[h]->hostname));
      }
    }
  }

  /* update state machine that map has been sent */
  fms_state_map_sent();

  return 0;

 except:
  return -1;
}

/*
 * Mark every FMA as needing a map update
 */
void
fms_fma_invalidate_map()
{
  struct lf_fabric *fp;
  int h;

  fp = F.fabvars->fabric;

  fms_notify(FMS_EVENT_DEBUG, "All hosts need new map.");

  /* free the existing map */
  LF_FREE(FMS_FABRIC(fp)->topo_map);

  /* mark every host as needing a map */
  for (h=0; h<fp->num_hosts; ++h) {
    struct fms_fma_desc *adp;

    adp = FMS_HOST(fp->hosts[h])->adp;
    if (adp != NULL) {
      adp->needs_map = TRUE;
    }
  }

  return;
}

/*
 * Mark every FMA that doesn't need a map as needing a link state update
 */
void
fms_fma_invalidate_link_state(
  struct lf_fabric *fp)
{
  int h;

  fms_notify(FMS_EVENT_DEBUG, "All hosts need new link state.");

  /* mark every host as needing a map */
  for (h=0; h<fp->num_hosts; ++h) {
    struct fms_fma_desc *adp;

    adp = FMS_HOST(fp->hosts[h])->adp;
    if (adp != NULL && !adp->needs_map) {
      adp->needs_link_state = TRUE;
    }
  }
  return;
}

/*
 * Mark this FMA as needing a map, and schedule a map push
 */
void
fms_fma_needs_topo_map(
  struct fms_fma_desc *adp)
{
  struct lf_fabric *fp;

  fp = F.fabvars->fabric;

  adp->needs_map = TRUE;	/* mark this FMA as needing a map */

  /* Arrange for a map to be sent */
  fms_schedule_map_push();
}

void
fms_schedule_map_request()
{
  struct lf_fabric *fp;
  struct fms_fabric *fmsp;

  fp = F.fabvars->fabric;
  fmsp = FMS_FABRIC(fp);

  /* If a map request is not pending and not already scheduled, schedule now */
  if (fmsp->request_map_timeout == NULL  && fmsp->request_map_task == NULL) {
    fmsp->request_map_task = lf_schedule_event(fms_fma_request_map,
	                                        NULL, 5*1000);
  }
}

/*
 * Send the topological part of the map to an FMA
 */
int
fms_send_topo_map(
  struct lf_fabric *fp,
  struct fms_fma_desc *adp)
{
  int rc;

  /* send topo map message header */
  rc = fms_send_fma_msg_hdr(adp, FMS_FABRIC(fp)->topo_map_size,
                            LF_FMA_TOPO_MAP);
  if (rc == -1) {
    LF_ERROR(("Sending topo_map header to FMA"));
  }

  /* send the topo map itself */
  rc = lf_write(adp->chp->fd, FMS_FABRIC(fp)->topo_map,
                FMS_FABRIC(fp)->topo_map_size);
  if (rc != FMS_FABRIC(fp)->topo_map_size) {
    LF_ERROR(("Sending topo_map to FMA"));
  }

  adp->needs_map = FALSE;	/* map no longer needed */
  adp->needs_link_state = FALSE;
  
  return 0;

 except:
  return -1;
}

/*
 * Send the topological link state to an FMA
 */
int
fms_send_topo_link_state(
  struct lf_fabric *fp,
  struct fms_fma_desc *adp)
{
  int rc;
  int size;

  /* send topo link state message header */
  size = LF_TOPO_LINK_STATE_SIZE(FMS_FABRIC(fp)->topo_map);
  rc = fms_send_fma_msg_hdr(adp, size, LF_FMA_TOPO_LINK_STATE);
  if (rc == -1) {
    LF_ERROR(("Sending topo_map header to FMA"));
  }

  /* send the topo link state */
  rc = lf_write(adp->chp->fd,
  	        LF_TOPO_LINK_STATE_ARRAY(FMS_FABRIC(fp)->topo_map),
  	        size);
  if (rc != size) {
    LF_ERROR(("Sending topo link state to FMA"));
  }

  adp->needs_link_state = FALSE;	/* link state no longer needed */
  
  return 0;

 except:
  return -1;
}
