/*
 * Structure and definitions for describing fabric topology.
 * These structures integrate the physical and logical topologies of the
 * network in order to allow for easy routing and meaningful error
 * reporting.
 *
 * Crossbars, NICs, and transceivers are all represented by a "node" structure.
 * 
 * Each node structure has one or two arrays of connections, which are
 * pointers to other nodes.
 * For a transceiver, the 1st half of the ports array is the external
 * connections for that node, the 2nd half is the internal connections.
 * For a crossbar or a host, one array describes the physical connections
 * for that node, the other array describes the topological connections.
 *
 */

/*
 * Some defines for messing with route bytes
 */
#define LF_DELTA_TO_ROUTE(D) (((D)&0x3F) | 0x80)
#define LF_ROUTE_TO_DELTA(R) (((unsigned char)(R)>=0xA0)?((R)-0xC0):((unsigned char)(R)-0x80))
#define LF_INVERT_ROUTE(R) (LF_DELTA_TO_ROUTE(-LF_ROUTE_TO_DELTA(R)))

#define LF_MAX_XBAR_PORTS 32

/*
 * Status field values for database
 */
enum {
  LF_DS_CURRENT = 0,
  LF_DS_NEW,
  LF_DS_DELETED
};

/* states a link may be in */
enum topo_link_state {
  LF_LINK_STATE_UP,
  LF_LINK_STATE_DOWN,		/* down, not by our choice */
  LF_LINK_STATE_FLAKY,		/* deemed flaky by FMS (high badcrc or ...) */
  LF_LINK_STATE_MAINTENANCE	/* disabled by user for maintenance */
};

/*
 * structure describing the entire fabric
 */
struct lf_fabric {

  int num_hosts;		/* hosts in the fabric */
  struct lf_host **hosts;

  int num_enclosures;		/* enclosures */
  struct lf_enclosure **enclosures;

  /* this section is only valid after calling lf_build_xbar_array */
  int num_xbars;		/* number of xbars */
  struct lf_xbar **xbars;	/* array of xbar pointers */

  void *user;			/* user-defined data */
};
typedef struct lf_fabric lf_fabric_t;

/* node types */
enum lf_node_type {
  LF_NODE_XBAR='x', LF_NODE_LC_XCVR='s', LF_NODE_NIC_XCVR='h', LF_NODE_NIC='n'
};

/*
 * Transcivers, xbars, and NICs are all nodes in the graph.  Rather than
 * make a huge, ugly union, we just require that the structures for all of them
 * begin with a "struct lf_node" and all the pointers are pointers to 
 * lf_nodes that can be cast to the appropriate things.
 */
struct lf_node {
  enum lf_node_type type;
};
typedef struct lf_node lf_node_t;
#define LF_NODE(N) ((lf_node_t *)(N))

/*
 * A transceiver with 2N ports actually has N connections - N external
 * via fiber or other medium, and N internal via copper.  We keep one array 
 * of 2*N ports, with 0..N-1 being the "external" ports and N..2*N-1 being
 * the internal ports.  External port X is associated with internal port
 * X+N. num_conn will be N, and num_ports will be 2*N.
 */
struct lf_xcvr {		/* a transceiver - LF_NODE_LC_XCVR or
				   LF_NODE_NIC_XCVR */
  struct lf_node n;

  int num_ports;	/* number of port structs */
  int num_conns;	/* number of external connections (num_ports/2) */

  struct lf_node **ports;	/* internal, probably to xbars */
  int *rports;			/* remote port (on xbar) */

  union {
    struct lf_linecard *linecard;	/* parent linecard */
    struct lf_nic *nic;			/* parent NIC */
  } p;
  int port;			/* what port we are on parent */

  struct lf_xcvr_data *data;	/* data from the switch */
  				/* (only for lincard transceivers) */

  void *user;			/* user defined pointer */
};
typedef struct lf_xcvr lf_xcvr_t;
#define LF_XCVR(N) ((lf_xcvr_t *)(N))

struct lf_nic {			/* a NIC - LF_NODE_NIC */
  struct lf_node n;

  int slot;			/* NIC index in host's array */

  int host_nic_id;		/* nic ID on host */
  uint32_t partition;		/* partition membership */
  lf_mac_addr_t mac_addr;

  char *serial_no;
  char *product_id;

  int num_ports;		/* connections and remote ports */
  struct lf_node **phys_ports;
  int *phys_rports;
  struct lf_node **topo_ports;
  int *topo_rports;

  struct lf_host *host;		/* parent host */
  
  struct lf_nic_def *def;	/* definition of this NIC */
  struct lf_nic_data **data;	/* data about this NIC */

  void *user;			/* user defined pointer */
};
typedef struct lf_nic lf_nic_t;
#define LF_NIC(N) ((lf_nic_t *)(N))

struct lf_xbar {		/* an xbar - LF_NODE_XBAR */
  struct lf_node n;

  int xbar_id;			/* xbar ID */

  int num_ports;		/* physical connections */
  struct lf_node **phys_ports;
  int *phys_rports;
  struct lf_node **topo_ports;
  int *topo_rports;

  struct lf_linecard *linecard;	/* parent linecard */
  int xbar_no;			/* which xbar on linecard */

  int clos_level;			/* level in clos, 0 closest to NICs */
  int min_nic_dist;			/* distance to closest NIC */
  int quadrant_disable;			/* TRUE if quad-disable enabled */

  struct lf_xbarport_data *data; /* array of per-port data from the switch */

  void *user;			/* user defined pointer */
};
typedef struct lf_xbar lf_xbar_t;
#define LF_XBAR(N) ((lf_xbar_t *)(N))

/*
 * A switch enclosure
 */
struct lf_enclosure {
  char *name;
  char *product_id;

  int num_lc_slots;
  int lc_slotbase;
  int num_bp_slots;
  int bp_slotbase;

  int num_slots;
  struct lf_linecard **slots;

  struct lf_enclosure_def *def;	/* definition for this enclosure */
  struct lf_enclosure_data *data;	/* data from the switch */

  void *user;			/* user defined pointer */
};
typedef struct lf_enclosure lf_enclosure_t;

/*
 * A line card
 */
struct lf_linecard {
  char *serial_no;
  char *product_id;

  int num_xcvrs;		/* transceivers on this card */
  struct lf_node **xcvrs;
  int *xcvr_labels;		/* what the silkscreen says */

  int num_xbars;		/* xbars on this card */
  struct lf_node **xbars;

  struct lf_enclosure *enclosure;	/* parent */
  int slot;
  int backplane;

  struct lf_linecard_def *def;		/* linecard definition */
  struct lf_linecard_data *data;	/* data from the switch */

  void *user;			/* user defined pointer */
};
typedef struct lf_linecard lf_linecard_t;


/*
 * A host
 */
struct lf_host {
  char *hostname;

  int num_nics;
  struct lf_nic **nics;

  void *user;			/* user-defined pointer */
};
typedef struct lf_host lf_host_t;

/*
 * prototypes
 */
lf_fabric_t *lf_simple_load_fabric(void);
void lf_simple_set_fabric_dirs(char *home, char *db_name);
void lf_free_fabric(lf_fabric_t *);
void lf_free_host(lf_host_t *);
void lf_free_node(lf_node_t *);
struct lf_nic *lf_alloc_nic(int num_ports);
struct lf_nic *lf_alloc_nic_by_product_id(char *product_id);
struct lf_nic *lf_alloc_generic_nic(int num_ports);
void lf_free_nic(lf_nic_t *);
void lf_free_xcvr(lf_xcvr_t *);
struct lf_xbar *lf_alloc_xbar(int num_ports);
void lf_free_xbar(lf_xbar_t *);
void lf_free_enclosure(lf_enclosure_t *);
void lf_free_linecard(lf_linecard_t *);
int lf_fill_enclosure(lf_enclosure_t *ep);
void lf_create_implicit_links(struct lf_enclosure *ep);
void lf_make_linecard_internal_links(lf_linecard_t *lp);
lf_node_t *lf_find_phys_node(lf_fabric_t *fp, char *name, int slot, int element);
lf_node_t *lf_get_topo_node(lf_node_t *np, int port, int *tport);
void lf_make_phys_link(lf_node_t *from, int fport, lf_node_t *to, int tport);
void lf_make_topo_link(lf_node_t *from, int fport, lf_node_t *to, int tport);
struct lf_node *lf_get_topo_link(lf_node_t *np, int port, int *oport);
struct lf_host *lf_find_host_by_name(struct lf_fabric *fp, char *hostname);
struct lf_enclosure *lf_find_enclosure_by_name(struct lf_fabric *fp,
                                               char *name);
struct lf_xbar *lf_find_xbar_by_id(struct lf_fabric *fp, int xbar_id);
struct lf_xbar *lf_find_another_xbar_by_id(struct lf_fabric *fp,
					   struct lf_xbar *orig_xp);
struct lf_nic *lf_find_nic_by_mac(struct lf_fabric *fp, lf_mac_addr_t mac);
struct lf_nic *lf_find_nic_by_mac_and_host(struct lf_fabric *fp,
                         lf_mac_addr_t mac, struct lf_host *hp);
int lf_slot_display_no(lf_linecard_t *lp);
lf_enclosure_t *lf_allocate_enclosure(char *product_id, char *name);
lf_linecard_t *lf_allocate_linecard(lf_enclosure_t *ep, int slot,
                                    char *product_id, char *serial_no);
lf_linecard_t *lf_allocate_generic_linecard(lf_enclosure_t *ep, int slot,
                                    int num_ports, char *serial_no);
int lf_build_xbar_array(struct lf_fabric *fp);
struct lf_nic *lf_find_nic_by_id(struct lf_host *myhp, int nic_id);
struct lf_fabric *lf_clone_fabric(struct lf_fabric *ofp);
struct lf_enclosure *lf_clone_enclosure(struct lf_enclosure *ep);
struct lf_host *lf_clone_host(struct lf_host *ohp);
int lf_clone_host_links(struct lf_fabric *ofp, struct lf_host *ohp,
			struct lf_fabric *nfp, struct lf_host *nhp);
int lf_clone_nic_links(struct lf_fabric *ofp, struct lf_nic *onicp,
			struct lf_fabric *nfp, struct lf_nic *nnicp);
struct lf_linecard *lf_clone_linecard(struct lf_enclosure *ep,
                                      struct lf_linecard *olp);
struct lf_node *lf_find_clone_counterpart(struct lf_node *onp,
                                          struct lf_fabric *fp);
struct lf_host *lf_add_host_to_fabric(struct lf_fabric *fp, char *hostname);
int lf_add_existing_host_to_fabric(struct lf_fabric *fp, struct lf_host *hp);
int lf_add_existing_enclosure_to_fabric(struct lf_fabric *fp,
                                        struct lf_enclosure *ep);
struct lf_nic *lf_add_nic_by_product_id(struct lf_host *hp, char *product_id);
void lf_remove_host_from_fabric(struct lf_fabric *fp, struct lf_host *hp);
void lf_remove_enclosure_from_fabric(struct lf_fabric *fp,
                                     struct lf_enclosure *ep);
void lf_remove_nic_from_host(struct lf_nic *nicp);
void lf_remove_link_by_topo(struct lf_node *np, int port);
void lf_remove_enclosure_from_fabric(struct lf_fabric *fp,
                                     struct lf_enclosure *ep);
void lf_reverse_route(unsigned char *dst, unsigned char *src, int len);
struct lf_node *lf_follow_topo_link(struct lf_node *np, int port, int *oportp);
struct lf_node *lf_phys_for_topo(struct lf_node *tnp, int tport, int *pportp);
void lf_connect_topo_nodes(struct lf_node *np1, int port1,
                           struct lf_node *np2, int port2);
void lf_node_string(char *s, struct lf_node *np, int port);
int lf_add_existing_nic_to_host(struct lf_nic *nicp, struct lf_host *hp);
int lf_get_xbar_id(struct lf_linecard *lp, int xbar_no);
int lf_get_topo_link_db_info(struct lf_node *np, int port, char **name,
    int *slot, int *dbport, int *subport);
int lf_get_enclosure_index(struct lf_fabric *fp, struct lf_enclosure *ep);
int lf_clos_label_xbars(struct lf_fabric *fp);
